%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.
%%%
%%% @author Viktoria Fordos <f-viktoria@elte.hu>
%%% @doc
%%% <b>RefactorErl API.</b>
%%%
%%% @todo doc!

-module(referl_api).

-behaviour(gen_server).
-include("referl.hrl"). %

% Interface
% requests
-export([low_level_request/2, urequest/3, request/1]).
% start/stop
-export([start/1, stop/0]).

% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

-record(state, {referl_node='refactorerl@localhost'}).
%%% ============================================================================
%%% Interface

% Starts RefactorErl API as a global server
start(ReferlHost)->
    start_server(ReferlHost).

% Stops both of RefactorErl's node and this node.
stop()->
    gen_server:call(?ServerName, {stop}, infinity),
    init:stop().

low_level_request(synch, Request)->
    ?SynchCall(gen_server:call(?ServerName, {synch_req, Request}, infinity));

low_level_request(asynch, Request)->
    gen_server:call(?ServerName, {synch_req, Request}, infinity).

urequest(Mod, Fun, Args)->
    ?SynchCall(gen_server:call(?ServerName, {unsafe_req, {Mod, Fun, Args}}, infinity)).

request(Request) when is_tuple(Request), is_atom(element(1, Request))->
    ?SynchCall(gen_server:call(?ServerName, {safe_req, Request}, infinity)).

%%% ============================================================================
%%% Callback fun.s

init([Referl])->
    %ReferlHost = list_to_atom(atom_to_list(Referl) ++"@"++ net_adm:localhost()),
    {ok, #state{referl_node=Referl}}.

handle_call({stop}, _From, State=#state{referl_node=Referl})->
    rpc:call(Referl, init, stop, []),
    {reply, ok, State};

handle_call({safe_req, Req}, _From, State=#state{referl_node=ReferlNode})->
    {reply, fun()-> safe_request0(ReferlNode, Req) end, State};

handle_call({unsafe_req, Req}, _From, State=#state{referl_node=ReferlNode})->
    {reply, fun()-> unsafe_request0(ReferlNode, Req) end, State};

handle_call({synch_req, Request}, _From, State=#state{referl_node=ReferlNode})->
    {reply, fun()-> request0(ReferlNode, Request) end, State}.

handle_cast(_,State)->
    {noreply, State}.

handle_info(_, State)->
    {noreply, State}.

terminate(_, _) ->
    ok.

code_change(_,_,_)->
    {error, no_update_mode}.

%%% ============================================================================
%%% Helpers

start_server(ReferlHost)->
    gen_server:start(?ServerName, ?MODULE, ReferlHost, []),
    ok.

 try_referl(ReferlHost)->
     case net_adm:ping(ReferlHost) of
         pang ->
             throw("RefactorErl hasn't been started, yet. Please, start it!");
         _ ->
             ok
     end.


request_loop(ReqID)->
    receive
        {ReqID, reply, R} ->
            case R of
                {error, X} ->
                    error;
                {ok, Result} ->
                    Result;
                Err ->
                    throw({unexcepted, Err})
            end;
        _Else ->
            request_loop(ReqID)
    end.

request0(ReferlNode, Request)->
    ReqID = rpc:call(ReferlNode, reflib_ui_router, getid, []),
    Res = rpc:call(ReferlNode, reflib_ui_router, request,
                    [{self(), node()}, ReqID, Request]),

    case Res of
        ok ->   request_loop(ReqID);
        deny -> throw(deny)
    end.

unsafe_request0(ReferlNode, {Mod, Fun, Args})->
    try_referl(ReferlNode),
    rpc:call(ReferlNode, Mod, Fun, Args, infinity).

safe_request0(ReferlNode, Req)->
    case is_request_allowed(Req) of
        true ->
            request0(ReferlNode, generate_request(Req));
        false ->
            deny
    end.


is_request_allowed({sem_query, _})->
    true;
is_request_allowed({add, _})->
    true;
is_request_allowed({drop, _})->
    true;
is_request_allowed({reset})->
    true;
is_request_allowed({db_sync})->
    true;
is_request_allowed({add_env, _, _})->
    true;
is_request_allowed({del_env, _})->
    true;
is_request_allowed({search_and_show_candidates})->
    true;
is_request_allowed({show_candidates})->
    true;
is_request_allowed({has_prev_result})->
    true;
% for general purpose
is_request_allowed({call, MFA})->
    is_allowed_to_call(MFA);
% whitelist-style
is_request_allowed(_Req)->
    false.


is_allowed_to_call({ris, Function,_})->
    AllowedRisFuns = [get_running_queries, kill_query, q, current_posmode,
                      system_info, print_dependencies],
    lists:member(Function, AllowedRisFuns);
is_allowed_to_call({refanal_skeleton, _,_})->
    true;
is_allowed_to_call(_)->
    false.

generate_request({sem_query, QueryString=[C|_]}) when is_integer(C)->
    {transform,semantic_query,
    [{ask_missing,false},
    {send_back_query_id, false},
    {querystr,QueryString},
    {display_opt,[{positions,both}, {output,other}]},
    {start_opt,[]}]};

generate_request({add, What})->
    {add_dir, What};

generate_request({drop, What})->
    {drop_dir, What};

generate_request({reset})->
    {reset};

generate_request({db_sync})->
    {synchronize};

generate_request({add_env, EnvName, EnvVal})->
    {add_env, EnvName, EnvVal};

generate_request({del_env, EnvName})->
    {del_env, EnvName};

generate_request({search_and_show_candidates})->
    {wrangler_general_req, {refpp_api, run, []}};

% @todo Revert this func. clause to the following after EUC.
% generate_request({show_candidates})->
%     {wrangler_general_req, {refpp_output, display, []}};
generate_request({show_candidates})->
    {wrangler_general_req, {refpp_output, display, [asynch]}};

generate_request({has_prev_result})->
    {wrangler_general_req, {refpp_output, has_stored_results, []}};

generate_request({call, MFA={_,_,_}})->
    {wrangler_general_req, MFA}.
