%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

%%% @doc Function side effect analyser. Caches function side effect
%%% information for more efficient recalculation in case of change.
%%%
%%% @author Laszlo Lovei <lovei@inf.elte.hu>
%%% @author Daniel Horpacsi <daniel_h@inf.elte.hu>

%%% TODO:
%%%  - rename "dirty" to "purity"

-module(refcore_funprop).
-vsn("$Rev: 12987 $ ").
-behaviour(gen_server).

-compile(export_all).
-export([update/1, update/2, remove/1, reset/0]).

%% gen_server exports
-export([start_link/0]).
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).

-include("core.hrl").
-include_lib("eunit/include/eunit.hrl").

-define(FUNPROP_TIMEOUT, infinity).

-define(unknown_purity, black).

%% @spec update(gnode()) -> ok
%% @doc Updates side effect and call-graph information. `Fun' is a
%% `#func{}' node, necessary information is queried from the graph.
update(Fun) ->
    gen_server:call(?FUNPROP_SERVER, {update, Fun}, ?FUNPROP_TIMEOUT).

%% @doc Currently, update/2 is only used by unit tests.
update(Fun, Info) ->
    gen_server:call(?FUNPROP_SERVER, {update, Fun, Info}, ?FUNPROP_TIMEOUT).

%% @spec remove(gnode()) -> ok
%% @doc Removes a function node from the cache (must be called when a function
%% node is deleted).
remove(Fun) ->
    gen_server:call(?FUNPROP_SERVER, {remove, Fun}, ?FUNPROP_TIMEOUT).

%% @spec reset() -> ok
%% @doc This function has to be called before the graph is reset.
%%      The function empties the internal funprop representation.
reset() ->
    gen_server:call(?FUNPROP_SERVER, {reset}, ?FUNPROP_TIMEOUT).


%% @spec start_link() -> {ok, Pid} | {error, Error}
%% @doc Starts the server.
start_link() ->
    gen_server:start_link({local, ?FUNPROP_SERVER}, ?MODULE, [], []).

%% @private
init(_) ->
    {ok, init_state()}.

%% @private
handle_call({update, Fun}, _, State) ->
    {reply, ok, handle_update(Fun, State)};
handle_call({update, Fun, Info}, _, State) ->
    {reply, ok, handle_update(Fun, Info, State)};
handle_call({remove, Fun}, _, State) ->
    {reply, ok, handle_remove(Fun, State)};
handle_call({reset}, _, _State) ->
    NewState = reset_call_graph(),
    {reply, ok, NewState}.

%% @private
handle_cast(_, State) ->
    {noreply, State}.

%% @private
handle_info(_Info, State) ->
    {noreply, State}.

%% @private
terminate(_Reason, _State) ->
    ok.

%% @private
code_change(_Old, State, _Extra) ->
    {ok, State}.

%%% ============================================================================
%%% Server implementation

init_state() ->
    ets:new(call_graph, [named_table]),
    ets:new(side_effects, [public, named_table]).

handle_remove(Fun, CG) ->
    {Purity, Refs, Deps} = read_info(Fun, CG),
    [del_ref(From, Fun, CG) || From <- Refs],
    [del_call(Fun, Dep, CG) || Dep <- Deps],
    case Purity of
        {white, _} -> ok; %% should this be refined?
        _ -> [update_external_purity(Ref, CG) || {_, Ref} <- Refs]
    end,
    ets:delete(CG, Fun),
    CG.

handle_update(Fun, CG) ->
    handle_update(Fun, query_info(Fun), CG).

handle_update(Fun, {IntPurity, Calls}, CG) ->
    {OldSEInfo, Refs, _} = read_info(Fun, CG),
    store_info(Fun, {IntPurity, undefined}, Calls, CG),
    update_external_purity(Fun, CG),
    {NewSEInfo, _, _} = read_info(Fun, CG),
    case influential_change(OldSEInfo, NewSEInfo) of
        false -> ok;
        true -> [update_external_purity(F, CG) || F <- closure(Refs, back, CG)]
    end,
    CG.

influential_change({I1, E1}, {I2, E2}) ->
    I1 /= I2 orelse E1 /= E2.

%% darker(undefined, _) -> true;
%% darker(X, Y) -> X > Y. %% gray > black

update_external_purity(Fun, CG) ->
    {{OldInt, OldExt}, _, Calls} = read_info(Fun, CG),
    case closure(Calls, fwd, CG) -- [Fun] of
        [] ->
            case OldExt of
                undefined -> ok;
                _ -> store_info(Fun, {OldInt, white}, CG)
            end;
        Closure ->
            NewExt = lists:foldl(fun(F, Acc) ->
                                         case purity(F, CG) of
                                             {X, _} -> min(X, Acc);
                                             _ -> Acc
                                         end
                                 end, white, Closure),
            OldExt /= NewExt andalso
                store_info(Fun, {OldInt, NewExt}, CG)
    end.

reset_call_graph() ->
    case lists:member(call_graph, ets:all()) of
        true  ->
            ets:delete_all_objects(call_graph),
            call_graph;
        false ->
            ets:new(call_graph, [named_table])
    end.

%%% ============================================================================
%%% Cache-related functionality

purity(Fun, CG) ->
    [{Fun, Purity, _, _}] = ets:lookup(CG, Fun),
    Purity.

store_info(Fun, Dirty, CG) ->
    D = ?Graph:data(Fun),
    ?Graph:update(Fun, D#func{dirty=Dirty}),
    {_, Refs, Deps} = read_info(Fun, CG),
    ets:insert(CG, {Fun, Dirty, Refs, Deps}).

store_info(Fun, Dirty, Calls, CG) ->
    {_, Refs, Deps} = read_info(Fun, CG),
    D = ?Graph:data(Fun),
    ?Graph:update(Fun, D#func{dirty=Dirty}),
    [add_call(Fun, Dep, CG) || Dep <- Calls -- Deps],
    [del_call(Fun, Dep, CG) || Dep <- Deps -- Calls],
    ets:insert(CG, {Fun, Dirty, Refs, Calls}).

del_ref({Type, From}, Fun, CG) ->
    %% We must update the cache entry of `From' in order to be able to
    %% note the deletion of the call to `Fun'; if we didn't update and
    %% `From' was not present in the cache, we would simply skip
    %% noting the new state of the stored call graph (?Graph:delete/1
    %% on `Fun' deletes all call edges without notifying this cache,
    %% thus changing the stored call graph).
    {Dirty, Refs, Deps} = read_info(From, CG),
    ets:insert(CG, {From, Dirty, Refs, Deps -- [{Type, Fun}]}).

add_call(From, {Type, To}, CG) ->
    ?Graph:mklink(From, Type, To),
    case ets:lookup(CG, To) of
        [] -> ok;
        [{To, Dirty, Refs, Deps}] ->
            ets:insert(CG, {To, Dirty, [{Type, From} | Refs], Deps})
    end.

del_call(From, {Type, To}, CG) ->
    ?Graph:rmlink(From, Type, To),
    case ets:lookup(CG, To) of
        [] -> ok;
        [{To, Dirty, Refs, Deps}] ->
            ets:insert(CG, {To, Dirty, Refs -- [{Type, From}], Deps})
    end.

read_info(Fun, CG) ->
    case ets:lookup(CG, Fun) of
        [{Fun, Dirty, Refs, Deps}] -> {Dirty, Refs, Deps};
        [] ->
            #func{dirty=Dirty} = ?Graph:data(Fun),
            Refs = refs(Fun, funcall) ++ refs(Fun, dyncall) ++ refs(Fun, ambcall),
            Deps = deps(Fun, funcall) ++ deps(Fun, dyncall) ++ deps(Fun, ambcall),
            ets:insert(CG, {Fun, Dirty, Refs, Deps}),
            {Dirty, Refs, Deps}
    end.

refs(Fun, Label) ->
    List = ?Graph:path(Fun, [{Label, back}]),
    lists:zip(lists:duplicate(length(List), Label), List).
deps(Fun, Label) ->
    List = ?Graph:path(Fun, [Label]),
    lists:zip(lists:duplicate(length(List), Label), List).

-type calldesc() :: {funcall, node()}
                  | {dyncall, node()}
                  | {ambcall, node()}.
-spec closure([calldesc()], back | fwd, ets:tab()) -> [node()].
closure(Calls, Dir, CG) ->
    Funs = [Fun || {_Label,Fun} <- Calls],
    Tab = ets:new(closure, []),
    ets:insert(Tab, [{Fun} || Fun <- Funs]),
    closure(Funs, Tab, Dir, CG).

-spec closure([node()], ets:tab(), back | fwd, ets:tab()) -> [node()].
closure([], Tab, _, _) ->
    Result = ets:select(Tab, [{{'$1'},[],['$1']}]),
    ets:delete(Tab),
    Result;

closure([Fun | Rest], Tab, Dir, CG) ->
    Next =
        case {Dir, read_info(Fun, CG)} of
            {back, {_, Refs, _}} -> Refs;
            {fwd,  {_, _, Deps}} -> Deps
        end,
    New = [F || {_,F} <- Next, ets:insert_new(Tab, {F})],
    closure(New ++ Rest, Tab, Dir, CG).

%%% ============================================================================
%%% Database-related functionality

-type gnode() :: tuple().
-type purity() :: white | gray | black.
-spec query_info(gnode()) -> {purity() | list(gnode())}.
%% @doc Internal purity and calls (as stored in the graph and the
%% predefinit records).
query_info(Fun) ->
    {M, F, A} = ?Fun:mfa(Fun), %% hopefully, any semantic function has a containing module
    case predef_purity(M, F, A) of
        no_idea ->
            case ?Graph:path(Fun, [{fundef, back}]) of
                [] ->
                    {?unknown_purity, []};
                [DefForm] ->
                    %% TODO: should not include the body of fun_exprs
                    %% QBody = [?Form:clauses(), ?Clause:body(), ?Expr:deep_sub()],
                    %% Nodes = try
                    %%             ?Query:exec(DefForm, ?Query:seq(QBody))
                    %%         catch
                    %%             error:semantic_link ->
                    %%                 children(DefForm)
                    %%         end,
                    Nodes = children(DefForm),
                    Colorize = fun({false, X}) -> {black, X};
                                  ({true, X}) -> {white, X}
                               end,
                    Colorize(lists:foldl(fun node_funprop/2, node_funprop(DefForm, {true, []}), Nodes));
                _ ->
                    %% shit happened here (some fun got two different definitions...)
                    {?unknown_purity, []}
            end;
        Purity ->
            {Purity, []}
    end.

predef_purity(M, F, A) ->
    %% TODO migrate purity_bifs into the ETS tab
    case purity_bifs:is_pure(M, F, A) of
        unknown ->
            case ets:lookup(side_effects, {M,F,A}) of
                []       -> no_idea;
                [{_, P}] -> P
            end;
        P ->
            decode_purity(P)
    end.

children(Node) -> children(Node, []).
children(Nodes, Acc) when is_list(Nodes) -> lists:foldl(fun children/2, Acc, Nodes);
children(Node, Acc) ->
    case ?Graph:class(Node) of
        lex -> Acc;
        _ -> children([Child || {_, Child} <- ?Syn:children(Node)], [Node|Acc])
    end.

node_funprop(Node, Acc) ->
    node_funprop(Node, ?Graph:data(Node), Acc).
node_funprop(ExprNode, Data = #expr{type=Type}, {Pure, Calls}) ->
    {Pure andalso pure(ExprNode, Data),
     if Type == application -> calls(ExprNode, Calls);
        true -> Calls
     end};
node_funprop(_Node, _Data, Acc) -> Acc.

pure(_Node, #expr{type=Type}) -> Type /= send_expr andalso Type /= receive_expr;
pure(_, _) -> true.

calls(Node, Calls) ->
    FunCalls = [{funcall, F} || {T, F} <- ?Graph:links(Node), lists:member(T, [funeref, funlref])],
    DynCalls = [{dyncall, F} || {T, F} <- ?Graph:links(Node), lists:member(T, [dynfuneref, dynfunlref])],
    AmbCalls = [{ambcall, F} || {T, F1} <- ?Graph:links(Node), lists:member(T, [ambfuneref, ambfunlref]),
                                F <- ?Graph:path(F1, [may_be])],
    ordsets:union(Calls, lists:usort(FunCalls++DynCalls++AmbCalls)).

%%% ============================================================================
%%% Helpers

%% @doc Translates Purity's side-effect codes into our
%% terminology.
decode_purity({p, _}) -> white;
decode_purity({e, _}) -> white;
decode_purity({d, _}) -> gray;
decode_purity({s, _}) -> black;
decode_purity({{at_least, P}, L}) -> decode_purity({P,L}); %% should be regarded as black?
decode_purity(unknown) -> ?unknown_purity.

%% @doc Merges internal and external purity. If the function is
%% completely pure (and does not call impure functions), its unified
%% purity is white. Impure functions' purity property is annototated
%% with the source of impurity, where if the external purity is
%% "darker", the annotation of the side-effect is `ext', otherwise it
%% is `int'.
-spec merge_purity(purity(), purity()) -> white | {gray | black, int | ext}.
merge_purity(white, white) -> white;
merge_purity(white, undefined) -> white; %% TODO review this
merge_purity(white, Ext)   -> {Ext, ext};
merge_purity(gray, black)  -> {black, ext};
merge_purity(Int, _Ext)    -> {Int, int}.

merge_purity({A,B}) -> merge_purity(A,B).

%%% ============================================================================
%%% Unit tests

-define(func(F, N, A), F = ?Graph:create(#func{name=N, arity=A})).
-define(update(F, S, C), ?FunProp:update(F, {S, C})).
-define(purity(F), merge_purity((?Graph:data(F))#func.dirty)).
-define(assertWhite(F), ?assertMatch(white, ?purity(F))).
-define(assertIntGray(F), ?assertMatch({gray,int}, ?purity(F))).
-define(assertExtGray(F), ?assertMatch({gray,ext}, ?purity(F))).
-define(assertIntBlack(F), ?assertMatch({black,int}, ?purity(F))).
-define(assertExtBlack(F), ?assertMatch({black,ext}, ?purity(F))).

update_test() ->
    ?func(M, m, 0),
    ?update(M, white, []),
    ?assertWhite(M),
    ?func(S1, s, 1),
    ?update(S1, white, [{funcall, M}]),
    ?assertWhite(S1),
    ?func(S2, s, 2),
    ?update(S2, gray, [{funcall, M}]),
    ?assertIntGray(S2),
    ?func(T1, t, 1),
    ?update(T1, black, []),
    ?assertIntBlack(T1),
    ?update(M, white, [{funcall, T1}]),
    ?assertExtBlack(M),
    ?assertExtBlack(S1),
    ?assertExtBlack(S2),
    ?update(T1, white, []),
    ?assertWhite(T1),
    ?assertWhite(M),
    ?assertWhite(S1),
    ?assertIntGray(S2),
    ?update(S2, gray, [{funcall, M}]),
    ?func(T2, t, 2),
    ?update(T2, white, [{funcall, S2}]),
    ?assertIntGray(S2),
    ?assertExtGray(T2),
    ?update(M, white, [{funcall, T1}, {funcall, T2}]),
    ?assertExtGray(M),
    ?assertExtGray(S1),
    ?assertIntGray(S2),
    ?assertWhite(T1).

update_delete_test() ->
    ?func(A, a, 0),
    ?func(B, b, 0),
    ?func(C, c, 0),
    ?update(C, white, []),
    ?update(B, white, [{funcall, C}]),
    ?update(A, white, [{funcall, B}]),
    ?assertWhite(A),
    ?assertWhite(B),
    ?assertWhite(C),
    ?FunProp:remove(A),
    ?Graph:delete(A),
    ?update(C, black, []),
    ?assertIntBlack(C),
    ?assertExtBlack(B).

update_delete2_test() ->
    ?func(A, a, 0),
    ?func(B, b, 0),
    ?update(A, gray, []),
    ?update(B, white, [{funcall, A}]),
    ?assertIntGray(A),
    ?assertExtGray(B),
    ?FunProp:remove(A),
    ?Graph:delete(A),
    ?assertWhite(B),
    ?FunProp:remove(B),
    ?Graph:delete(B).

update_delete3_test() ->
    ?func(A, a, 0),
    ?func(B, b, 0),
    ?update(A, white, [{funcall, B}]),
    ?assertExtBlack(A),
    ?assertIntBlack(B),
    ?FunProp:remove(A),
    ?Graph:delete(A),
    ?assertIntBlack(B),
    ?FunProp:remove(B),
    ?Graph:delete(B).

%%% ============================================================================
%%% Funprop export/import functionality

load_otp() ->
    %% code:lib_dir(),
    load_otp("/home/daniel/R16_preprocessed/lib").
load_otp(D) ->
    Apps = string:tokens(os:cmd("ls " ++ D ++ " | cut -d '-' -f 1"), "\n"),
    [ri:add('R16', list_to_atom(L)) || L <- Apps].

update_all() ->
    lists:map(fun(Fun) ->
                      io:format("Updating ~p:~p/~p...~n", tuple_to_list(?Fun:mfa(Fun))),
                      update(Fun)
              end,
              ?Query:exec([module, func])).

-define(path, "side_effects.dat").

%% TODO export from the cache? (read_info queries the graph otherwise)
%% -- then this should be turned into a server call
export() ->
    Data = [export_data({mod, M}) || M <- ?Query:exec([module])],
    file:write_file(?path, io_lib:fwrite("~p.~n", [Data])).

export_data({mod, Mod}) ->
    {?Mod:name(Mod), [export_data({func, F}) || F <- ?Query:exec(Mod, [func])]};
export_data({func, Fun}) ->
    #func{name=N, arity=A, dirty=P} = ?Graph:data(Fun),
    {N, A, export_purity(P)}.

export_purity({I,undefined}) -> I;
export_purity({I,E}) -> min(I,E). % black < gray

import() ->
    {ok, [Data]} = file:consult(?path),
    [import_data(D) || D <- Data].

import_data({M, Funs}) ->
    [ets:insert(side_effects, {{M, N, A}, P}) || {N, A, P} <- Funs].
    

%%% ============================================================================
%%% Component purity (sanity)

-define(unknown_resource, '$unknown_resource').

sanity(Components) when is_list(Components) ->
    Info = [inspect_component(C) || C <- Components],
    lists:all(fun({C, Rest}) -> sanity(C, [R || C2 <- Rest, {_, R} <- C2]) end, lists:zip(Info, sublists(Info))).

%% n-1 length sublists
sublists([]) -> [];
sublists([H | T]) -> [T | [[H | X] || X <- sublists(T)]].

inspect_component(Component) ->
    %% Component = list of its expressions
    [{P, R} || E <- children(Component), {P, R} <- [expr_purity(E)]]. % we're omitting white expressions

sanity(Component, Resources) ->
    case [R || {black, R} <- Component] of
        [] -> true;
        Blacks ->
            not ((Resources /= [] andalso lists:member(?unknown_resource, Blacks)) orelse
                 lists:any(fun(R) -> lists:member(R, Resources) end, Blacks))
    end.

expr_purity(Expr) ->
    {Purity, Calls} = node_funprop(Expr, {true, []}),
    IntPurity = case Purity of
                    true -> white;
                    false -> black
                end,
    ExtPurity = lists:foldl(fun(F, Acc) ->
                                    case merge_purity(?Fun:dirtiness(F)) of
                                        white -> Acc;
                                        {X, _} -> min(X, Acc) %% gray > black
                                    end
                            end, white, [Fun || {_, Fun} <- Calls]),
    case merge_purity(IntPurity, ExtPurity) of
        white -> white;
        {X, _} -> %% gray or black
            {X, case sideeffect_resource(Expr) of
                    %% TODO multiple resources referred to?
                    [Resource] -> Resource;
                    [] -> ?unknown_resource
                end}
    end.
    
sideeffect_resource(Expr) ->
    %% Currently we only support ETS operations
    ?Query:exec(Expr, ?Query:any([ets_def], [ets_ref])).

%%% ============================================================================
%%% Tests

get_components(Funnames) -> lists:map(fun get_component/1, Funnames).
get_component(Funname) -> ?Query:exec([module, {func, {name, '==', Funname}}, {fundef, back}, funcl, body]).

-define(pure_comps(L), begin io:format("~p~n", [L]), sanity(get_components(L)) end).

comp_test() ->
    true = ?pure_comps([rx1, rx2]),
    true = ?pure_comps([rx1, rx2, wy1]),
    true = ?pure_comps([wx1, wy1]),
    %% unknown_resource
    true = ?pure_comps([rx1, rx2, rz]),
    true = ?pure_comps([wz]),

    false = ?pure_comps([rx1, wx1]),
    false = ?pure_comps([rx2, wx1]),
    false = ?pure_comps([rx1, rx2, wx1]),
    false = ?pure_comps([wx2, wx1]),
    %% unknown_resource
    false = ?pure_comps([wx1, wz]),
    false = ?pure_comps([wz, wx2]),
    false = ?pure_comps([wy1, wz]),
    false = ?pure_comps([rx1, wz]),
    false = ?pure_comps([rx1, wz, rx2]),
    false = ?pure_comps([rz, wz]).
