-module(costest).
-include("referl_paraphrase.hrl").
-export([run/1, run_skel/1]).

-callback init() -> ok.
-callback files() -> list().
-callback components() -> any().
-callback inputs() -> list().
-callback skeleton(list()) -> any().
-callback candidate(list()) -> any().

-define(CORES, 24).
-define(SCHEDULERS, ?CORES).

-define(debug(S), io:format("~s~n", [S])).
-define(debug(S, D), io:format("~s: ~w~n", [S, D]), _ = D).
-define(debug(S, X, D), io:format("~s-~w: ~w~n", [S, X, D]), _ = D).

%% TODO: handling of worker numbers should be generalised

run_skel(TestMod) ->
    TestMod:init(),
    ?debug("Benchmarking with skel..."),
    {S, P} = benchmark(TestMod, ?CORES),
    ?debug("Realised  speedup   ", S/P),
    ?debug("Done.").

run(TestMod) ->
    TestMod:init(),
    ?debug("Benchmarking with referl..."),
    {ES, EP, W} = referl(TestMod),
    ?debug("Benchmarking with skel..."),
    {S, P} = benchmark(TestMod, W),
    ?debug("Estimated speedup   ", ES/EP),
    ?debug("Corrected estimation", (ES/EP)*0.9),
    ?debug("Realised  speedup   ", S/P),
    ?debug("Accuracy %          ", (S/P)/((ES/EP)*0.9)*100),
    ?debug("Done.").

%% -----------------------------------------------------------------------------
%% RefactorErl

referl(TestMod) ->
    ris:reset(),
    [ris:add({F}) || F <- TestMod:files()],
    Exprs = reflib_query:exec([module, {func, {name, '==', components}}, {fundef, back}, funcl, body, esub, esub, {esub, 2}, exprcl, body]),
    ?silent(R = refpp_costs:calc(TestMod:candidate(Exprs), ?CORES, length(TestMod:inputs()))),
    %R = refpp_costs:calc(TestMod:candidate(Exprs), ?CORES, length(TestMod:inputs())),
    % refcore_esg:finalize(),
    [#tr_seq{attribs=#attrib{tseqcpu=S, tparcpu=P},comps=_}] = R,
    W = tl(hd(workers(R))),
    ?debug("   Workers--", W),
    ?debug("Sequential--", S),
    ?debug("  Parallel--", P),
    {S, P, W}.

workers(Xs) when is_list(Xs) ->
    lists:map(fun workers/1, Xs);
workers(#tr_seq{attribs=#attrib{num_of_workers=W},comps=Comps}) ->
    [W|workers(Comps)];
workers(#comp{attribs=#attrib{num_of_workers=W}}) ->
    W.
          
    

%% -----------------------------------------------------------------------------
%% Skel

benchmark(TestMod, W) ->
    Input = TestMod:inputs(),
    ?debug("Workers to be started", W),
    {Comps, _} = lists:unzip(TestMod:components()),
    Skel = TestMod:skeleton({Comps, W}),
    {sk_sequential(Skel, Input), sk_parallel(Skel, Input, ?SCHEDULERS)}.

sk_sequential(Skeleton, Input) ->
    R = sk_profile:benchmark(
          fun() -> sk_assembler_sequential:run(Skeleton, Input),
                   receive
                       {sink_results, Results} -> Results
                   end
          end, [], 1),
    ?debug("Sequential--", proplists:get_value(mean, R)).

sk_parallel(Skeleton, Input, Schedulers) ->
    erlang:system_flag(schedulers_online, Schedulers),
    ?debug("  Parallel", Schedulers, proplists:get_value(mean, sk_parallel(Skeleton, Input))).

sk_parallel(Skeleton, Inputs) ->
    sk_profile:benchmark(
      fun() -> sk_assembler:run(Skeleton, Inputs),
               receive
                   {sink_results, Results} -> Results
               end
      end, [], 1).
