%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

-module(refpp_benchmark).
-compile([export_all]).
-export([test/0, typetest/0]).
-export([time/1, time/4]).
-export([timertc/3]).
-export([closure/1]).

-include("referl_paraphrase.hrl").
-define(portedname, packegedfun).
-define(numtests, 3).
-define(qc_gen_size, 100).

-define(Warning(S, A), io:format("[WARNING] " ++ S ++ "~n", A)).

typetest() ->
    typetest(random_expr_seq()).

random_expr_seq() ->
    ?out("Looking for a random expression sequence..."),
    FunClauses   = ?Query:exec(?Query:seq([?Mod:all(), ?Mod:locals(), ?Fun:definition(), ?Form:clauses()])),
    Exprs        = ?Query:exec(FunClauses, ?Query:seq(?Clause:body(), ?Expr:deep_sub())),
    Blocks       = lists:filter(fun(E) -> lists:member(?Expr:type(E), [fun_expr, block_expr]) end, Exprs),
    ExprClauses  = ?Query:exec(Blocks, [exprcl]),
    Clauses      = FunClauses ++ ExprClauses,
    ?out("\t(found ~p clauses in the database)", [length(Clauses)]),
    random:seed(now()),
    RandomClause = lists:nth(random:uniform(length(Clauses)), Clauses),
    ?Query:exec(RandomClause, [body]).

test() ->
    ?MODULE:time(?Query:exec(
                    ?Query:seq(
                       [?Mod:find(fact),
                        [{func, {name, '==', f}}],
                        ?Fun:definition(),
                        [{funcl, 1}],
                        [{body, {0, last}}]]))).

test2() ->
    ?MODULE:time(?Query:exec(
                    ?Query:seq(
                       [?Mod:find(matrix_ex_paper),
                        [{func, {name, '==', mult_sum}}],
                        ?Fun:definition(),
                        [{funcl, 1}],
                        [{body, {0, last}}]]))).

time({Mod, Fun, Arity}, C, N, M) ->
    ?MODULE:time(?Query:exec(
                    ?Query:seq(
                       [?Mod:find(Mod),
                        ?Fun:find(Fun, Arity),
                        ?Fun:definition(),
                        [{funcl, C}],
                        [{body, {N, M+1}}]]))).

time(ExprsOrFunc, Ss) ->
    ets:new(tobechanged, [named_table, public]),
    ets:new(mycounter, [named_table, public]),
    ets:insert(mycounter, {counter, 1}),
    [ets:insert(tobechanged, {S, []}) || S <- Ss],
    try
        time(ExprsOrFunc)
    after
        ets:delete(tobechanged),
        ets:delete(mycounter)
    end.

insert_checkpoints() ->
    try ets:tab2list(tobechanged) of
        [] -> ok;
        [{E,Copies}|_] ->
            ets:delete_object(tobechanged, {E, Copies}),
            [catch insert_checkpoint(E, E2) || E2 <- Copies],
            insert_checkpoints()
    catch
        _:_ -> ok
    end.

insert_checkpoint({'$gn',X,Y} = _Orig, Expr) ->
    [{counter, I}] = ets:lookup(mycounter, counter),
    ets:insert(mycounter, {counter, I+1}),
    V = "GENERATED" ++ integer_to_list(I),
    E1 = add_prews(?Syn:construct({send_expr, {atom, ts_receiver}, {tuple, [{atom, s}, {tuple, [{atom, X}, {integer, Y}]}]}})),
    E2 = ?Syn:construct({match_expr, {var,V}, copy(Expr)}),
    E3 = ?Syn:construct({send_expr, {atom, ts_receiver}, {tuple, [{atom, e}, {tuple, [{atom, X}, {integer, Y}]}]}}),
    E4 = add_postws(?Syn:construct({var,V})),
    Block = ?Syn:construct({block_expr, [E1,E2,E3,E4]}),
    replace(Expr, Block).

time(ExprsOrFunc) ->
    try
        %% TODO: use data-flow to reduce the number of free vars
        {Texts, Hash, _FreeVars} = encapsulate(ExprsOrFunc),

        %% Loading the synthesised module into the runtime
        ByteCode = text2bytecode(Texts, Hash),
        code:load_binary(Hash, doesnotmatter, ByteCode),

        %% Translating the synthesised module into abstract code and
        %% inferring the type of the synthesised function with typer
        AbstractCode = text2abstractcode(Texts, Hash),
        [[{_RetType, ArgType}]] = typer:start({abstract_code, Hash, AbstractCode}),
        ?d(ArgType),

        %% Creating QC data generators for each formal parameter and
        %% generting random parameters with QC
        ?silent(TestData = egg_types:samples(ArgType, ?numtests, ?qc_gen_size)),
      %  ?d(TestData),

        benchmark(Hash, ?portedname, TestData)
    catch
        cannot_compile ->
            ?Warning("Benchmarking skipped due to an error related to compilation of the synthesised module.", []),
            stat([-1]);
        cannot_encapsulate_types ->
            ?Warning("Benchmarking skipped due to an error related to encapsulation of type specifications.", []),
            stat([-1]);
        type_not_supported ->
            ?Warning("Benchmarking skipped due to an error related to random data generation.", []),
            stat([-1])
    end.

typetest(ExprsOrFunc) ->
    ?out("Encapsulating the expression sequence..."),
    {Texts, Hash, FreeVars} = encapsulate(ExprsOrFunc),
    try
        ?out("Compiling the synthesised module into bytecode..."),
        ByteCode = text2bytecode(Texts, Hash),
        ?out("Loading the bytecode into the VM..."),
        code:load_binary(Hash, doesnotmatter, ByteCode),
        ?out("Compiling the synthesised module into abstract code..."),
        AbstractCode = text2abstractcode(Texts, Hash),
        ?out("Calculating value type..."),
        {_, FunValueType} = proplists:lookup({?portedname, length(FreeVars)}, valuetypes:start({forms, AbstractCode})),
        ValueType = erl_types:t_fun_args(FunValueType),
        ?out("Calculating success type..."),
        [[{_RetType, SuccessType}]] = typer:start({abstract_code, Hash, AbstractCode}),
        ?d(ValueType),
        ?d(SuccessType),
        ?out("Generating random arguments based on the types..."),
        TestData = egg_types:samples(ValueType, ?numtests, 10000),
        TestData2 = egg_types:samples(SuccessType, ?numtests, 10000),
        ?out("Trying to execute the synthesised code..."),
        Ex = lists:flatten([try2run(Hash, ?portedname, D) || D <- TestData]),
        Ex2 = lists:flatten([try2run(Hash, ?portedname, D) || D <- TestData]),
        ?d(Ex),
        ?d(Ex2)
    catch
        cannot_compile ->
            ':('
    end.

benchmark(Mod, Fun, TestData) ->
    Positive = fun(X) -> X > 0 end,
    Ts = [timertc(Mod, Fun, Args) || Args <- TestData],
    Timing = case lists:filter(Positive, Ts) of
                 [] -> Ts;
                 X  -> X
             end,
    stat(Timing).

stat(Timing) ->
    Mean = mean(Timing),
    [{min, lists:min(Timing)},
     {max, lists:max(Timing)},
     {med, median(Timing)},
     {mean, Mean},
     {std_dev, std_dev(Timing, Mean)}].

timertc(Mod, Fun, Args) ->
    timertc(Mod, Fun, Args, 0, 0).
timertc(_Mod, _Fun, _Args, N, Acc) when N > 2, (N > 10000 orelse Acc > 200000) ->
    Acc / N;
timertc(Mod, Fun, Args, N, Acc) ->
    catch ts_receiver ! s,
    Pid = self(),
    TimerFun = fun() -> Res = try timer:tc(Mod, Fun, Args)
                              catch _:_ -> failed
                              end,
                        Pid ! Res
               end,
    Pid2 = spawn(TimerFun),
    receive
        failed -> -1;
        {T, _} ->
            catch ts_receiver ! e,
            timertc(Mod, Fun, Args, N + 1, Acc + T)
    after 5000 ->
            exit(Pid2, kill),
            -1
    end.

try2run(Mod, Fun, Args) ->
    try
        apply(Mod, Fun, Args),
        []
    catch
        _Class : Exception -> [Exception]
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% FROM   skel/sk_profile.erl

median(List) -> lists:nth(round((length(List) / 2)), lists:sort(List)).
mean(List) -> lists:foldl(fun(X, Sum) -> X + Sum end, 0, List) / length(List).
std_dev(List, Mean) -> math:pow(variance(List, Mean), 0.5).
variance(List, _Mean) when length(List) == 1 -> 0.0;
variance(List, Mean) -> lists:foldl(fun(X, Sum) -> math:pow(Mean - X, 2) + Sum end, 0, List) / (length(List) - 1).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% FROM   branches/transform/core/reftrans_fun.erl
%% [NOTE] Extended to handle func nodes
%% [NOTE] Fixed the naming issues with depfuns

-define(include_specs(S), S).
%-define(include_specs(S), begin false andalso S, [] end). %% no unusedfun warnings

encapsulate(Func = {'$gn', func, _}) ->
    ExportForm = ?Syn:construct({{form, export}, [{funlist, [{funref, {atom, ?portedname}, {integer, ?Fun:arity(Func)}}]}]}),
    NewForm    = ported_fun(Func),
    DepFuns    = closure([Func]),
    DepForms   = ?Query:exec(DepFuns, ?Fun:definition()),
    PortedDeps = ?Query:exec(DepForms, [fundef]),
    CDepForms  = [ported_dep(F, PortedDeps) || F <- DepForms],
    Specs      = ?include_specs(specs(DepFuns)),
    CSpecs     = [ported_spec(F) || F <- Specs],
    RecSpec    = copy(lists:usort(records(DepForms))) ++ CSpecs,

    insert_checkpoints(),

    Texts1     = [?Syn:tree_text(Form) || Form <- RecSpec ++ [ExportForm, NewForm] ++ CDepForms],
    Hash       = list_to_atom(integer_to_list(erlang:phash2(Texts1))),
    ModuleForm = ?Syn:construct({{form, module, Hash}, []}),
    {[?Syn:tree_text(ModuleForm) | Texts1], Hash, []};

encapsulate({func, FunExpr}) ->
    FreeVars   = free_vars(FunExpr),
    ExportForm = ?Syn:construct({{form, export}, [{funlist, [{funref, {atom, ?portedname}, {integer, length(FreeVars)+1}}]}]}),
    DepFuns    = closure(?Query:exec(FunExpr, funapps())),
    DepForms   = ?Query:exec(DepFuns, ?Fun:definition()),
    PortedDeps = ?Query:exec(DepForms, [fundef]),
    CDepForms  = [ported_dep(F, PortedDeps) || F <- DepForms],
    CBody      = fix_apps(FunExpr, PortedDeps),
    CFunForm   = ported_fun(FreeVars, {expr, CBody}),
    RecSpec    = rec_spec(FunExpr, DepFuns, DepForms),

    insert_checkpoints(),

    Texts1     = [?Syn:tree_text(Form) || Form <- RecSpec ++ [ExportForm, CFunForm] ++ CDepForms],
    Hash       = list_to_atom(integer_to_list(erlang:phash2(Texts1))),
    ModuleForm = ?Syn:construct({{form, module, Hash}, []}),
    {[?Syn:tree_text(ModuleForm) | Texts1], Hash, []};

encapsulate(ExprOrExprs) ->
    FreeVars   = free_vars(ExprOrExprs),
    ExportForm = ?Syn:construct({{form, export}, [{funlist, [{funref, {atom, ?portedname}, {integer, length(FreeVars)}}]}]}),
    DepFuns    = closure(?Query:exec(ExprOrExprs, funapps())),
    DepForms   = ?Query:exec(DepFuns, ?Fun:definition()),
    PortedDeps = ?Query:exec(DepForms, [fundef]),
    CDepForms  = [ported_dep(F, PortedDeps) || F <- DepForms],
    CBody      = fix_apps(ExprOrExprs, PortedDeps),
    CFunForm   = ported_fun(FreeVars, CBody),
    RecSpec    = rec_spec(ExprOrExprs, DepFuns, DepForms),

    insert_checkpoints(),

    Texts1     = [?Syn:tree_text(Form) || Form <- RecSpec ++ [ExportForm, CFunForm] ++ CDepForms],
    Hash       = list_to_atom(integer_to_list(erlang:phash2(Texts1))),
    ModuleForm = ?Syn:construct({{form, module, Hash}, []}),
    {[?Syn:tree_text(ModuleForm) | Texts1], Hash, FreeVars}.

rec_spec(ExprOrExprs, DepFuns, DepForms) ->
    Recs = lists:usort(records(ExprOrExprs) ++ records(DepForms)),
    CSpecs = [ported_spec(F) || F <- ?include_specs(specs(DepFuns++Recs))],
    copy(Recs) ++ CSpecs.

%% @doc Renames functions by prefixing them with their containing
%% module, in order to avoid name clashes. Accordingly, fix_apps/2
%% adjusts applications to the "ported names" (for example, m:f()
%% becomes m_f()).
ported_dep(OrigForm, PortedDeps) ->
    [Fun] = ?Query:exec(OrigForm, [fundef]),
    NewName = prefixed_name(Fun),
    NewForm = fix_apps(OrigForm, PortedDeps),
    [replace(E, ?Syn:construct({atom, NewName})) || E <- ?Query:exec(NewForm, [funcl, name])],
    NewForm.

%% We should properly handle specifications and references inside
%% them. Here's a very first, quite naive try.
ported_spec(OrigForm) ->
    SpecForm = copy(OrigForm),
    case ?ESG:data(OrigForm) of
        #form{type=spec} ->
            [NameExpr] = ?Query:exec(SpecForm, [{tattr, 1}]),
            case ?ESG:data(NameExpr) of
                #typexp{type=atom, tag=OldName} ->
                    [Module] = ?Query:exec(OrigForm, [{form, back}, moddef]),
                    NewName = list_to_atom(atom_to_list(?Mod:name(Module)) ++ "_" ++ atom_to_list(OldName)),
                    NewNameExpr = ?Syn:create(#typexp{type=atom}, [{tlex, ?Syn:create_lex(atom, io_lib:write(NewName))}]),
                    add_prews(NewNameExpr),
                    replace(NameExpr, NewNameExpr);
                _ ->
                    throw(todo)
            end;
        #form{type=type} ->
            todo;
        _ ->
            todo
    end,
    SpecForm.

fix_apps(Nodes, PortedDeps) when is_list(Nodes) ->
    lists:map(fun(Node) -> fix_apps(Node, PortedDeps) end, Nodes);
fix_apps(Node, PortedDeps) ->
    Filter = fun(N) ->
                     case ?Graph:data(N) of
                         #expr{type=application} ->
                             case ?Query:exec(N, ?Expr:dynfunction()) of
                                 [] -> false; %% dynfun...
                                 [Fun] ->
                                     case lists:member(Fun, PortedDeps) of
                                         true  -> {true, ported_app(N, Fun, PortedDeps)};
                                         false -> false
                                     end;
                                 _ -> false %% dynfuns...
                             end;
                         #expr{type=implicit_fun} ->
                             case ?Query:exec(N, ?Expr:dynfunction()) of
                                 [] -> false; %% dynfun...
                                 [Fun] ->
                                     case lists:member(Fun, PortedDeps) of
                                         true  -> {true, ported_implicitfun(N, Fun)};
                                         false -> false
                                     end;
                                 _ -> false %% dynfuns...
                             end;
                         _ -> false
                     end
             end,
    proplists:get_value(Node, register_copies(?Syn:copy(Node, Filter, true))).

ported_app(App, Fun, PortedDeps) ->
    NewName = prefixed_name(Fun),
    [_Name, Args] = ?Query:exec(App, [esub]),
    ListOfArgs = ?Query:exec(Args, [esub]),
    NewApp = ?Syn:construct({app, {atom, NewName}, fix_apps(ListOfArgs, PortedDeps)}),
    %% The fix_apps call on ListOfArgs is needed: when fix_apps (which
    %% is basically a special deep copy) finds an application, it
    %% fixes the name of the applied function according to the
    %% dependency changes; however, arguments of the application might
    %% contain applications as well... ListOfArgs contains the old
    %% (and probably bad) references to the moved functions, so that
    %% even though the deep copy creates a good (reshaped) copy of the
    %% embedded applications, they won't be used... what we do instead
    %% is regardsless of those unused copies, redo the work and create
    %% a good (reshaped) copy of the arguments here.
    add_prews(NewApp).

ported_implicitfun(App, Fun) ->
    NewName = prefixed_name(Fun),
    [_Name, Arity] = ?Query:exec(App, [esub]),
    NewNameExpr = add_prews(?Syn:construct({atom, NewName})),
    NewApp = ?Syn:construct({implicit_fun, NewNameExpr, Arity}),
    add_prews(NewApp).

add_prews(Node) ->
    [L|_] = ?Syn:leaves(Node),
    LD = #lex{data=TD=#token{}} = ?ESG:data(L),
    ?ESG:update(L, LD#lex{data=TD#token{prews=" "}}),
    Node.

add_postws(Node) ->
    [L|_] = lists:reverse(?Syn:leaves(Node)),
    LD = #lex{data=TD=#token{}} = ?ESG:data(L),
    ?ESG:update(L, LD#lex{data=TD#token{postws=" "}}),
    Node.

prefixed_name(Fun) ->
    {M, N, _} = ?Fun:mfa(Fun),
    list_to_atom(atom_to_list(M) ++ "_" ++ atom_to_list(N)).

ported_fun(Func) ->
    Pattern = [{var, "V" ++ integer_to_list(I)} || I <- lists:seq(1, ?Fun:arity(Func))],
    ?Syn:construct({func, [{fun_clause, {atom, ?portedname}, Pattern, [], [{app, {atom, prefixed_name(Func)}, Pattern}]}]}).

ported_fun(FreeVars, {expr, Expr}) ->
    Pattern = [{var, "X"} | [{var, ?Var:name(V)} || V <- FreeVars]],
    ?Syn:construct({func, [{fun_clause, {atom, ?portedname}, Pattern, [], [{app, Expr, [{var, "X"}]}]}]});
ported_fun(FreeVars, Exprs) when is_list(Exprs) ->
    Pattern = [{var, ?Var:name(V)} || V <- FreeVars],
    ?Syn:construct({func, [{fun_clause, {atom, ?portedname}, Pattern, [], Exprs}]});
ported_fun(FreeVars, FunExpr) ->
    Pattern = [{var, ?Var:name(V)} || V <- FreeVars],
    ?Syn:construct({func, [{fun_clause, {atom, ?portedname}, Pattern, [], [copy(FunExpr)]}]}).

register_copies(Copied) ->
    lists:foreach(
      fun({N, N2}) ->
              L = ets:tab2list(tobechanged),
              lists:foreach(
                fun({Orig, Copies}) ->
                        if N == Orig -> ets:insert(tobechanged, {N,[N2|Copies]});
                           true -> ok
                        end,
                        case lists:member(N, Copies) of
                            true -> ets:insert(tobechanged, {Orig,[N2|Copies]});
                            false -> ok
                        end
                end,
                L)
      end, Copied),
    Copied.

copy(Nodes) when is_list(Nodes) ->
    [copy(Node) || Node <- Nodes];
copy(Node) ->
    proplists:get_value(Node, register_copies(?Syn:copy(Node, fun(_) -> false end, true))).
replace(Old, New) when not is_list(New) ->
    replace(Old, [New]);
replace(Old, New) ->
    [{_, Parent}|_] = ?Syn:parent(Old),
    ?Syn:replace(Parent, {node, Old}, New).

closure(Funs) ->
    Tab = ets:new(closure, []),
    ets:insert(Tab, [{Fun} || Fun <- Funs]),
    closure(Funs, Tab).
closure([], Tab) ->
    Result = ets:select(Tab, [{{'$1'},[],['$1']}]),
    ets:delete(Tab),
    Result;
closure([Fun | Rest], Tab) ->
    Next = ?Query:exec(Fun, [funcall]) ++ ?Query:exec(Fun, [dyncall]),
    New = [F || F <- Next, ets:insert_new(Tab, {F})],
    closure(New ++ Rest, Tab).

funapps()->
    fun(Expr) ->
            Apps = [App ||App <- ?Query:exec(Expr, ?Expr:deep_sub()),
                          ?Expr:type(App) == 'application' orelse ?Expr:type(App) == 'implicit_fun'],
            ?Query:exec(Apps, ?Expr:dynfunction())
    end.

free_vars(Exprs) when is_list(Exprs) ->
    X = [begin
             AllSubs = ?Query:exec(Expr, ?Expr:deep_sub()),
             {used_vars(AllSubs), bound_vars(AllSubs)}
         end ||
            Expr <- Exprs],
    {UsedVars, BoundVars} = lists:unzip(X),
    lists:usort(lists:flatten(UsedVars)) -- lists:flatten(BoundVars);
free_vars(Expr) ->
    AllSubs = ?Query:exec(Expr, ?Expr:deep_sub()),
    UsedVars = used_vars(AllSubs),
    BoundVars = bound_vars(AllSubs),
    lists:usort(UsedVars) -- BoundVars.

used_vars(AllSubs) -> ?Query:exec(AllSubs, [varref]).
bound_vars(AllSubs) -> ?Query:exec(AllSubs, [{varintro, back}]). %% should this use `varbind' perhaps?

records(Nodes) when is_list(Nodes) ->
    lists:flatten(lists:map(fun records/1, Nodes));
records(Node) ->
    records(Node, ?ESG:data(Node)).

records(Node, #expr{}) ->
    AllSubs  = ?Query:exec(Node, ?Expr:deep_sub()),
    ?Query:exec(AllSubs, [recref, {recdef, back}]);
records(Node, #form{}) ->
    AllSubs  = ?Query:exec(Node, ?Form:deep_exprs()),
    ?Query:exec(AllSubs, [recref, {recdef, back}]).

specs(Funs) when is_list(Funs) ->
    lists:flatten(lists:map(fun specs/1, Funs));
specs(Form = {'$gn', form, _}) ->
    TypeDefs = typedefs(),
    case ?Query:exec(Form, [{form, back}, moddef]) of
        [Mod] ->
            ModName = ?Mod:name(Mod),
            lists:usort(lists:flatten(closure_typeref([Form], {ModName, TypeDefs})));
        _ ->
            throw(cannot_encapsulate_types)
    end;
specs(Fun = {'$gn', func, _}) ->
    [Mod] = ?Query:exec(Fun, [{func, back}]),
    ModName = ?Mod:name(Mod),
    FunName = ?Fun:name(Fun),
    Arity = ?Fun:arity(Fun),
    case [Form || {Form, {MN, FN, Ar}} <- specs(),
                  MN == ModName,
                  FN == FunName,
                  Ar == Arity] of
        [SpecForm] ->
            TypeDefs = typedefs(),
            lists:usort(lists:flatten(closure_typeref([SpecForm], {ModName, TypeDefs})));
        [] ->
            []
    end.

typedefs() ->
    TypForms = ?Query:exec([file, {form, {{type, '==', type}, 'and', {tag,  '==', type}}}]), %% todo tag opaque
    [{Form, type_info(Form)} || Form <- TypForms].

type_info(Form) ->
    [TAtom, TFunArgs, _TypSpec] = ?Query:exec(Form, [tattr]),
    Arity = length(?Query:exec(TFunArgs, [tsub])),
    #typexp{type=atom, tag=TypName} = ?ESG:data(TAtom),
    [Mod] = ?Query:exec(Form, [{form, back}, moddef]),
    {?Mod:name(Mod), TypName, Arity}.

type_refs(Nodes) when is_list(Nodes) ->
    lists:flatten(lists:map(fun type_refs/1, Nodes));
type_refs(Node) ->
    case ?ESG:data(Node) of
        #form{type = spec} ->
            [_TName, TSig] = ?Query:exec(Node, [tattr]),
            type_refs(TSig);
        #form{type = type} ->
            [_TAtom, _TFunArgs, TypSpec] = ?Query:exec(Node, [tattr]),
            type_refs(TypSpec);
        #form{type = record} ->
            type_refs(?Query:exec(Node, [tattr]));
        #typexp{type=call} ->
            [TypName, TArgList] = ?Query:exec(Node, [tsub]),
            Arity = length(?Query:exec(TArgList, [tsub])),
            [{spec_name(TypName), Arity}] ++ type_refs(TArgList);
        #typexp{} ->
            lists:map(fun type_refs/1, ?Query:exec(Node, [tsub]));
        _ ->
            []
    end.

closure_typeref(Forms, Context) ->
    Tab = ets:new(closure, []),
    ets:insert(Tab, [{Form} || Form <- Forms]),
    closure_typeref(Forms, Context, Tab).
closure_typeref([], _, Tab) ->
    Result = ets:select(Tab, [{{'$1'},[],['$1']}]),
    ets:delete(Tab),
    Result;
closure_typeref([Form | Rest], {ModName, Defs}, Tab) ->
    TypeRefs = lists:usort(lists:flatten(type_refs(Form))),
    TypeRefs2 = [{{ModName, TypName}, TypArity} || {TypName, TypArity} <- TypeRefs, is_atom(TypName)],
    Next = [F || {{ModName2, TypName}, TypArity} <- TypeRefs2,
                 {F, {MN, TN, Ar}} <- Defs,
                 MN == ModName2,
                 TN == TypName,
                 Ar == TypArity],
    New = [F || F <- Next, ets:insert_new(Tab, {F})],
    closure_typeref(New ++ Rest, {ModName, Defs}, Tab).

specs() ->
    SpecForms = ?Query:exec([file, {form, {type, '==', spec}}]),
    [{Form, spec_info(Form)} || Form <- SpecForms].

spec_info(Form) ->
    [TName, TSig] = ?Query:exec(Form, [tattr]),
    case {spec_name(TName), spec_arity(TSig)} of
        {FunName, Arity} when is_atom(FunName), is_integer(Arity) ->
            [Mod] = ?Query:exec(Form, [{form, back}, moddef]),
            {?Mod:name(Mod), FunName, Arity};
        {{ModName, FunName}, Arity} when is_atom(ModName), is_atom(FunName), is_integer(Arity) ->
            {ModName, FunName, Arity};
        {{FunName, Arity}, _Arity} when is_atom(FunName), is_integer(Arity) ->
            [Mod] = ?Query:exec(Form, [{form, back}, moddef]),
            {?Mod:name(Mod), FunName, Arity};
        _ ->
            undefined
    end.

spec_name(TExpr) ->
    case ?ESG:data(TExpr) of
        #typexp{type=atom, tag=Atom} ->
            Atom;
        #typexp{type=integer, tag=Int} ->
            Int;
        #typexp{type=module_qualifier} ->
            [TAtom, TAtom2] = ?Query:exec(TExpr, [tsub]),
            {spec_name(TAtom), spec_name(TAtom2)};
        #typexp{type=arity_qualifier} ->
            [TypName, TInt]  =  ?Query:exec(TExpr, [tsub]),
            {spec_name(TypName), spec_name(TInt)};
        _ ->
            throw(name_not_present)
    end.

spec_arity(TExpr) ->
    case ?ESG:data(TExpr) of
        #typexp{type=spec_union} ->
            [TypFunSpec | _TypFunSpecs] = ?Query:exec(TExpr, [tsub]), % omitting all but the first
            spec_arity(TypFunSpec);
        #typexp{type=spec_guard} ->
            [TFunSig, _TypGuard] = ?Query:exec(TExpr, [tsub]),
            spec_arity(TFunSig);
        #typexp{type=fun_sig} ->
            [TArgList, _TypSpec] = ?Query:exec(TExpr, [tsub]),
            spec_arity(TArgList);
        #typexp{type=poly_sig} ->
            undefined;
        #typexp{type=arglist} ->
            length(?Query:exec(TExpr, [tsub]));
        _ ->
            undefined
    end.

-define(report_compilation_error(T, E), % ok).
        begin
            ?out("The module synthesised from the anonymous function cannot be"
                 " ~p. Reason:", [T]),
            ?out("~p", [E])
        end).

subst_module_macro(Texts, ModName) ->
    [re:replace(lists:flatten(Text), "\\?MODULE", atom_to_list(ModName), [global, {return, list}]) || Text <- Texts].

text2bytecode(Texts, Hash) ->
    syntree2bytecode(text2syntree(subst_module_macro(Texts, Hash))).

text2abstractcode(Texts, Hash) ->
    syntree2abstractcode(text2syntree(subst_module_macro(Texts, Hash))).

text2syntree(Texts) -> [parse_form(Text) || Text <- Texts].

parse_form(Chars) ->
    ?out("~s", [Chars]),
    case erl_scan:string(Chars) of
        {ok, Ts, _} ->
            case erl_parse:parse_form(Ts) of
                {ok, F} ->
                    F;
                {error, Reason} ->
                    ?out("~s", [Chars]),
                    ?report_compilation_error(parsed, Reason),
                    throw(cannot_compile)
            end;
        {error, Reason, _} ->
            ?report_compilation_error(tokenized, Reason),
            throw(cannot_compile)
    end.

syntree2abstractcode(Forms) ->
    case compile:forms(Forms, [binary, to_pp]) of
        {ok, _Module, AbstrCode} ->
            AbstrCode;
        error ->
            ?report_compilation_error(compiled, no_info),
            throw(cannot_compile);
        {error, Errors, _Warnings} ->
            ?report_compilation_error(compiled, Errors),
            throw(cannot_compile)
    end.

syntree2bytecode(Forms) ->
    case compile:forms(Forms, [binary, return]) of
        {ok, _Module, Binary, _List} ->
            Binary;
        error ->
            ?report_compilation_error(compiled, no_info),
            throw(cannot_compile);
        {error, Errors, _Warnings} ->
            ?report_compilation_error(compiled, Errors),
            throw(cannot_compile)
    end.
