%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

-module(refpp_cost_model).

-import(lists, [max/1, sum/1]).
-export([calibrate/0, receiver/1, number_of_processors/0]).
-export([timecost_pipe/2,
	 timecost_farm/3, timecost_farm/4]).

-include("referl_paraphrase.hrl").

-define(DTAB, filename:join([?MISC:data_dir(),
                             "benchmark_table"++ ?MISC:graph_based_postfix()])).
-define(wordsize, erlang:system_info(wordsize)).

number_of_processors() ->
    erlang:system_info(logical_processors_available).

%% Assuming that we have at least as many cores as stages...
timecost_pipe(L, T_stages) ->
    M = length(T_stages),
    %% The cost of processing all the elements in parallel
    sum(T_stages) + (L - 1) * max(T_stages)
    %% Setting up M workers and pushing every item through the whole pipeline
        + (timecost_spawn() + timecost_setup() +  timecost_copy(L)) * M.

timecost_farm(N_w, L, T_work) ->
    timecost_farm(number_of_processors(), N_w, L, T_work).

timecost_farm(N_p, N_w, L, T_work) ->
    %% The cost of processing all the elements in parallel
    T_work * ceiling(L / min(N_p, N_w))
    %% Setting up N_w workers and an emitter + distributing the data
        + timecost_emitter(N_w, L)
    %% Setting up a collector and collecting the data
        + timecost_collector(N_w, L).

timecost_emitter(N_w, L) ->
    (timecost_spawn() + timecost_setup()) * (N_w + 1) %% + sk_source spawn time?
        + timecost_copy(L) * 3. %% sk_source -> emitter -> worker

timecost_collector(_N_w, L) ->
    %% 1) the collector receives the data from the worker and 2) sends
    %% towards the final pid
    timecost_spawn() + timecost_setup() + timecost_copy(L) * 2.

timecost_spawn() ->
    case get(t_spawn) of
        undefined -> 0;
        X -> X
    end.

timecost_setup() ->
    0.

timecost_copy(L) ->
    case get(t_copy) of
        undefined -> L;
        X -> (L * 2 * ?wordsize) * X
    end.

ceiling(X) when X < 0 ->
    trunc(X);
ceiling(X) ->
    T = trunc(X),
    case X - T == 0 of
        true -> T;
        false -> T + 1
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

calibrate() ->
    calibrate(t_copy),
    calibrate(t_spawn),
    ok.

calibrate(Tag) ->
    case get(Tag) of
        undefined ->
            dets:open_file(?DTAB, []),
            case dets:lookup(?DTAB, Tag) of
                [{Tag, T}] -> ok;
                [] -> T = calibrate_timecost(Tag),
                      dets:insert(?DTAB, {Tag, T})
            end,
            dets:close(?DTAB),
            put(Tag, T);
        _ ->
            ok
    end.

calibrate_timecost(t_copy)  -> calibrate_timecost_copy();
calibrate_timecost(t_spawn) -> calibrate_timecost_spawn().

calibrate_timecost_copy() ->
    ?out("Testing your heap copy speed..."),
    Indexes = lists:seq(16,25),
    T = lists:sum([time_copy(trunc(math:pow(2, X))) || X <- Indexes])/length(Indexes),
    ?out("Estimated copy speed: ~.2fMB/s", [1/T*1000000/1024/1024]),
    T.

time_copy(Bytes) when Bytes rem 8 == 0, Bytes >= 2 * 8 ->
    HeapWords = Bytes div ?wordsize,
    %% the number 42 takes 2 heap words (at least in R16)
    Data = lists:duplicate(HeapWords div 2, 42),
    Self = self(),
    P = spawn(?MODULE, receiver, [Self]),
    %%?out("Sending ~p megabytes...", [Bytes/1024/1024]),
    Times = [send_time(P, Data ++ [X]) || X <- lists:seq(1,10)],
    %% io:format("~p~n", [proplists:get_value(heap_size, erlang:process_info(P))]),
    exit(P, kill),
    lists:sum(Times)/length(Times)/Bytes.

receiver(Parent) ->
    receive _Data -> Parent ! done end,
    receiver(Parent).

send_time(Child, Data) ->
    T1 = os:timestamp(),
    Child ! Data,
    receive
        done ->
            T2 = os:timestamp(),
            timer:now_diff(T2, T1)
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

calibrate_timecost_spawn() ->
    ?out("Testing your spawning speed..."),
    Limit = erlang:system_info(process_limit),
    Current = erlang:system_info(process_count),
    Max = (Limit-Current), % div 1000,
    Ns = [Max div 8, Max div 4, Max div 2],
    R = [time_spawn_n(N) || N <- Ns, _ <- lists:seq(1, 10)],
    T = lists:sum(R) / length(R),
    ?out("Estimated spawning speed: ~p proc/s", [round(1/T*1000000)]),
    T.

time_spawn_n(N) ->
    %%?out("Starting ~p processes...", [N]),
    T1 = os:timestamp(),
    Pids = spawn_n(N, []),
    T2 = os:timestamp(),
    N = length(Pids), % check
    [exit(Pid, kill) || Pid <- Pids],
    timer:now_diff(T2, T1) / N.

spawn_n(N, Pids) when N < 1 -> Pids;
spawn_n(N, Pids) ->
    Pid = spawn(fun() -> receive _ -> ok end end),
    spawn_n(N - 1, [Pid | Pids]).
