%%% -*- coding: utf-8 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Eötvös Loránd University.
%%% Portions created  by Eötvös Loránd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Eötvös Loránd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.
%%%
%%% @author Judit Koszegi <koszegijudit@elte.hu>

-module(refpp_costs).

-vsn("$Rev: 13345 $").

-export([calc/2]).
-export([testcalc/0, calc_metric/1, syn_metric/1]).

-include("referl_paraphrase.hrl").

-define(FARM_LABEL, "Δ").
-define(PIPE_LABEL, "||").
-define(DNC_LABEL, "ꕺ").
-define(POOL_LABEL, "o").

-define(DEFAULT_LENGTH_OF_STREAM, 240).
-define(METRIC(Node), calc_metric(Node)).

testcalc() ->
    calc(refpp_skeleton:get_candidate_chain_combinations(), []).

%% Input comes from refanal_skeleton:get_candidate_chain_combinations().
-spec calc(CandidateChainCombinations :: [candidate()],
           PropList :: search_options()) 
          -> [#tr_seq{}].
calc(CandidateChainCombinations, PropList) ->
    DefaultLength = proplists:get_value(default_length_of_stream, PropList, 
					?DEFAULT_LENGTH_OF_STREAM),
    put(default_length, DefaultLength),
    Filter = proplists:get_value(filter_benchmark_errors, PropList, 
					false),
    put(filter_benchmark_errors, Filter),
    refpp_cost_model:calibrate(),
    ets:new(tseqcpu_table, [set, named_table]),
    NumOfCores = proplists:get_value(num_of_cores, PropList, 
				      refpp_cost_model:number_of_processors()),
    AnnotatedTrees = annotate_trees(CandidateChainCombinations, NumOfCores),
    ets:delete(tseqcpu_table),
    lists:map(fun convert_to_tr_seq/1, 
	      lists:filter(fun(Tree) -> Tree /= error end, AnnotatedTrees)).

annotate_trees(CandidateChainCombinations, NumOfCores)->
    [annotate_tree(Tree, NumOfCores) || Tree <- CandidateChainCombinations].

annotate_tree(Candidate, NumOfCores)->
    TreeWithTseq = add_label_and_tseq(Candidate),
    case validTseq(TreeWithTseq) of
	true -> add_cost(TreeWithTseq, NumOfCores, -1);
	false -> 
	    case get(filter_benchmark_errors) of
		false -> TreeWithTseq;
		true -> error
	    end
    end.
   
%% Stage1
%% Annotates the tree with labels (for the ui module) and sequential
%% runtime which will be measured in case of primitive components (jobs/stages)
%% and will be calculated in case of compound ones.

add_label_and_tseq(#seq{node = Node, parent_type = PType}) ->
    {Tseqcpu, Label, ListLength} = stage1_seq(Node, PType),
    Attribs = #attrib{tseqcpu = Tseqcpu, label = Label, list_length = ListLength},
    #seq{node = Node, parent_type = PType, attribs = Attribs};
add_label_and_tseq(Tree) ->
    case Tree of
        #farm{} -> 
	    {Tseqcpu, Tseqother, Label, ListLength, NewJob} = stage1_farm(Tree),
	    Attribs = #attrib{tseqcpu = Tseqcpu,
			      tseqother = Tseqother,
			      label = Label, 
			      list_length = ListLength},
	    Tree#farm{job = NewJob, attribs = Attribs};
        #pipe{} ->
	    {Tseqcpu, Label, ListLength, NewStages} = stage1_pipe(Tree),
	    Attribs = #attrib{tseqcpu = Tseqcpu,
			      label = Label,
			      list_length = ListLength},
	    Tree#pipe{stages = NewStages, attribs = Attribs};
	#dnc{} ->
	    {Label, NewEmbeds} = stage1_dnc(Tree),
	    Attribs = #attrib{label = Label},
	    Tree#dnc{embeds = NewEmbeds, attribs = Attribs};
	#pool{} ->
	    {Label, NewEmbeds} = stage1_pool(Tree),
	    Attribs = #attrib{label = Label},
	    Tree#pool{embeds = NewEmbeds, attribs = Attribs}
    end.

add_func_label(Node, PType) ->
    case {Node, PType} of
        {#expr{}, mapapp} -> 
            NodeType = ?Expr:type(Node),
            case NodeType == fun_expr orelse NodeType == implicit_fun of
                true -> {func,Node};
                _    ->  Node
            end;
        _ -> Node
    end.

stage1_seq(Node, PType) ->
    Tseqcpu = measure_tseqcpu(add_func_label(Node, PType)),
    Label = generate_new_label(Node),
    ListLength = 1,
    {Tseqcpu, Label, ListLength}.

stage1_farm(Tree) ->
    #farm{node = Node, job = Job, type = Type} = Tree,
    L = length_of_stream(Node, Type),
    {JobNode, Subtrees} = Job,
    SubtreeNodes = lists:map(fun get_node/1, Subtrees),

    Tseqother = 
        case SubtreeNodes of
            [] -> 
                0;
            _ -> 
                NodesWithTseqs = 
                    refpp_timestamp:measure(embedded, 
                                            add_func_label(JobNode, Type), 
                                            SubtreeNodes),
                %?d(NodesWithTseqs),
                Tseqs = [Tseq || {_Node, Tseq} <- NodesWithTseqs],
		case lists:member(undefined, Tseqs) of
		    false -> lists:sum(Tseqs);
		    true -> 0
		end
        end,

    NewSubtrees = 
        if 
            Type == func andalso Subtrees == [] ->
                RecNodes = rec_call_nodes(Node),
                [{_, Tseq} | _ ]= refpp_timestamp:measure(maplike, 
                                                     add_func_label(Node, Type), 
                                                     RecNodes),
                [create_seq_with_tseq(JobNode, Type, Tseq)];
            true -> 
                add_label_and_tseq_to_subtrees(Subtrees, JobNode, Type)
        end,
    Label = "("++ ?FARM_LABEL ++ concat_labels_of_trees(NewSubtrees) ++ ")",
    NewJob = {JobNode, NewSubtrees},
    
    Tseqcpu = 
	case tseq_sum_of_trees(NewSubtrees) of
	    undefined -> undefined; 
	    JobTseq ->
                case is_rewrite_embedding(Node, NewSubtrees) of
                    true ->  JobTseq;
                    false -> (JobTseq + Tseqother) * L
                end
	end,
    {Tseqcpu, Tseqother, Label, L, NewJob}.

stage1_pipe(Tree) ->
    #pipe{nodes = Nodes, stages = Stages, type = Type} = Tree,
    L = 
        case Nodes of
            [Node] -> length_of_stream(Node, Type);
            _ ->  length_of_stream(hd(Nodes), func) %hack!
        end,

    StNodes = [Node || {Node,_Subtrees} <- Stages],
    NodesWithTseqs = refpp_timestamp:measure(pipe, 
                                             add_func_label(hd(Nodes), Type), 
                                             StNodes),
    %?d(NodesWithTseqs),
    Tseqs = [Tseq || {_Node, Tseq} <- NodesWithTseqs],
    ListOfNewSubtrees = add_label_and_tseq_to_stages(Stages, Type, Tseqs),

    Labels = [concat_labels_of_trees(NewSubtrees) || 
		 NewSubtrees <- ListOfNewSubtrees],
    Label = "(" ++ concat_labels(Labels,?PIPE_LABEL) ++ ")",
    NewStages = lists:zip(StNodes, ListOfNewSubtrees),
    Times = lists:map(fun tseq_sum_of_trees/1, ListOfNewSubtrees),
         
    Tseqcpu =
	case lists:member(undefined, Times) of
	    true -> undefined;
	    _ ->
		%% modify in case of "rewriting rule" embedding!
		lists:sum(Times) * L
	end, 
    {Tseqcpu, Label, L, NewStages}.

stage1_dnc(Tree) ->
    #dnc{node = Node, embeds = Embeds, type = Type} = Tree,
    NewEmbeds = add_label_and_tseq_to_subtrees(Embeds, Node, Type),
    Label = "("++ ?DNC_LABEL ++ concat_labels_of_trees(NewEmbeds) ++ ")",
    {Label, NewEmbeds}.

stage1_pool(Tree) ->
    #pool{node = Node, embeds = Embeds, type = Type, job = {JobNode, _}} = Tree,
                                                     % job never embeds other candidates
    {NewEmbeds, ExtraJobLabel} =
	case Embeds of
	    [] -> {add_label_and_tseq_to_subtrees([], JobNode, Type),
		   []};
	    _  -> {add_label_and_tseq_to_subtrees(Embeds, Node, Type),
		   generate_new_label(JobNode)}
	end,
    Label = "("++ ?POOL_LABEL ++ ExtraJobLabel ++ concat_labels_of_trees(NewEmbeds) ++ ")",
    {Label, NewEmbeds}.

measure_tseqcpu(NodeOrNodeWithLabel) ->
    Tseqcpu =
        case ets:lookup(tseqcpu_table,NodeOrNodeWithLabel) of
            [] ->
                %?d(NodeOrNodeWithLabel),
                T = proplists:get_value(mean, refpp_benchmark:time(NodeOrNodeWithLabel,[])),
                ets:insert(tseqcpu_table,{NodeOrNodeWithLabel,T}),
                T;
            [{_ , T}] ->
                T
        end,
    case Tseqcpu =< 0 of
        true -> undefined;
        _ -> Tseqcpu
    end.
        
%% Stage2
%% Annotate the tree with the calculated paralell runtime.
%% In case of farms it tries every possible combinations (from 1 to
%% 2*number_of_cores) in every level of the tree and choose the best.

add_cost(Tree, Np, Length) ->
    case Tree of
        #seq{} ->
            {Nw, NwSum, Tparcpu} = stage2_seq(Tree),
            Attribs = Tree#seq.attribs,
            NewAttribs = Attribs#attrib{num_of_workers = Nw, nw_sum = NwSum, tparcpu = Tparcpu},
            Tree#seq{attribs = NewAttribs};

        #farm{} ->
            {Nw, NwSum, Tparcpu, NewJob} = stage2_farm(Tree, Np, Length),
            Attribs = Tree#farm.attribs,
            NewAttribs = Attribs#attrib{num_of_workers = Nw, nw_sum = NwSum, tparcpu = Tparcpu},
            Tree#farm{job = NewJob, attribs = NewAttribs};       

        #pipe{} -> 
            {Nw, NwSum, Tparcpu, NewStages} = stage2_pipe(Tree, Np, Length),
            Attribs = Tree#pipe.attribs,
            NewAttribs = Attribs#attrib{num_of_workers = Nw, nw_sum = NwSum, tparcpu = Tparcpu},
            Tree#pipe{stages = NewStages, attribs = NewAttribs}
    end.

stage2_seq(Tree) ->
    #seq{attribs = Attribs} = Tree,
    Nw = 1, NwSum = 1, Tparcpu = Attribs#attrib.tseqcpu,
    {Nw, NwSum, Tparcpu}.

stage2_farm(Tree, Np, Length) ->
    #farm{node = Node, job = Job, type = Type, attribs = Attribs} = Tree,
    Tseqother = Attribs#attrib.tseqother,
    L = case Length of
            -1 -> length_of_stream(Node, Type);
            _ -> Length
        end,
    {JobNode, Subtrees} = Job,
    IsEmbedd = is_rewrite_embedding(Node, Subtrees),
    {Nw, Tparcpu0, NewSubtrees} = best_nw_data(Np, L, Tseqother, Subtrees, IsEmbedd),
    NwSum = 
        case max_nw_sum(NewSubtrees) of
            1 -> Nw;
            SubNw -> Nw + Nw *  SubNw
        end,            
    NewJob = {JobNode, NewSubtrees},
    Tparcpu = if Nw == 1 -> Attribs#attrib.tseqcpu; %% not worth to introduce farm
                 true -> Tparcpu0
              end,
    {Nw, NwSum, Tparcpu, NewJob}.    

stage2_pipe(Tree, Np, Length) ->
    #pipe{nodes = Nodes, stages = Stages, type = Type, attribs = Attribs} = Tree,
    L = case Length of
            -1 ->  case Nodes of
                       [Node] -> length_of_stream(Node, Type);
                       _ ->  length_of_stream(hd(Nodes), func) %hack!
                   end;
            _ -> Length
        end,
    ListOfNewSubtrees = add_cost_to_stages(Stages, Np),
    StNodes = [Node || {Node,_Subtrees} <- Stages],
    NewStages = lists:zip(StNodes, ListOfNewSubtrees),
    ParTimes = lists:map(fun tpar_sum_of_trees/1, ListOfNewSubtrees),
    {Tparcpu, NwSum, Nw} =
        if Np < length(Stages) -> {Attribs#attrib.tseqcpu, Np, Np}; %% not worth to introduce pipe
           true -> {cost_pipe(Np, L, ParTimes),
                   lists:sum([max_nw_sum(NewSubtrees) || NewSubtrees <- ListOfNewSubtrees]),
                   length(Stages)}
        end,
    {Nw, NwSum, Tparcpu, NewStages}. 
                    
best_nw_data(Np, L, Tseqother, Subtrees, IsDivided) ->
    NewL = if IsDivided -> L; true -> -1 end,
    NewSubtrees = [add_cost(Tree, Np, NewL) || Tree <- Subtrees],
    Tw = tpar_sum_of_trees(NewSubtrees) + Tseqother,
    best_nw_data(2, Np, L, Tseqother, Subtrees, IsDivided, 
                 {1, cost_farm(1, Np, L, Tw), NewSubtrees}).

best_nw_data(Nw, Np, _L, _Tseqother, _Subtrees, _IsDivided, ResultTuple) when Nw >= Np + 1 ->
    ResultTuple;

best_nw_data(Nw, Np, L, Tseqother, Subtrees, IsDivided, {BestNw, BestTime, BestSubtrees}) ->
    NewNp =
        case floor((Np - Nw)/Nw) of
            0 -> 1;
           -1 -> 1;
            X -> X
        end,
    NewL =  if IsDivided -> erlang:round(L/Nw); true -> -1 end,
    case NewNp of
        0 -> {BestNw, BestTime, BestSubtrees};
        _ ->
            NewSubtrees = [add_cost(Tree, NewNp, NewL) || Tree <- Subtrees],
            Tw = tpar_sum_of_trees(NewSubtrees) + Tseqother,

            Cost = if IsDivided -> cost_farm(Nw, Np, Nw, Tw); 
                      true -> cost_farm(Nw, Np, L, Tw) end,
            case Cost < BestTime of
                true  -> best_nw_data(Nw+1, Np, L, Tseqother, Subtrees, IsDivided, 
                                      {Nw, Cost, NewSubtrees});
                false -> best_nw_data(Nw+1, Np, L, Tseqother, Subtrees, IsDivided, 
                                      {BestNw, BestTime, BestSubtrees})
            end
    end.

%% In case of pipe we calculate with 1 core in case of a "primitive" stages
%% and distribute equally the remainder cores among the embedding stages.
%% TODO: find the optimal distribution (?)
add_cost_to_stages(Stages, Np) ->
    SeqStages =
        lists:filter(fun ({_StNode, [#seq{}]}) -> true;
                         (_) -> false
                     end, Stages),
    NumOfSeqStages = length(SeqStages),
    NpEmbedd0 = 
        case (length(Stages) == NumOfSeqStages) of
            true -> 1;
            false -> 
                floor((Np - NumOfSeqStages) / (length(Stages) - NumOfSeqStages))
        end,
    NpEmbedd = case NpEmbedd0 of 0 -> 1; N -> N end,
    lists:map(fun ({StNode, Trees}) ->
                      StNp = case lists:member({StNode, Trees}, SeqStages) of
                                 true  -> 1; 
                                 false -> NpEmbedd
                             end,
                      [add_cost(Tree, StNp, -1) || Tree <- Trees]
              end, Stages).

cost_farm(Nw, Np, L, Tw) ->
    refpp_cost_model:timecost_farm(Np, Nw, L, Tw).

cost_pipe(_Np, L, Ts) ->
    %% later we can use Np in the cost model
    refpp_cost_model:timecost_pipe(L, Ts).


%% Stage3
%% Converts the tree to a record that is needed for the ui.
%% Calculate and add metric values to the components.

convert_to_tr_seq(Tree) ->
    Comps = create_components(Tree),
    create_tr_seq_record(Comps).

create_components(#seq{node = Node, attribs = Attribs}) ->
    Label = Attribs#attrib.label,
    MetricValue = ?METRIC(Node),
    NewAttribs = Attribs#attrib{metric = MetricValue}, 
    [#comp{node_id = Node, label = Label, attribs = NewAttribs, type = seq}];
create_components(#farm{node = Node, attribs = Attribs, job = Job}) ->
    Label = Attribs#attrib.label,
    {JobNode, Subtrees} = Job,
    MetricValue = ?METRIC(JobNode),
    NewAttribs = Attribs#attrib{metric = MetricValue}, 
    Comp = #comp{node_id = Node, label = Label, attribs = NewAttribs, type = farm},
    [Comp | lists:append([create_components(SubTree)  || SubTree <- Subtrees])];
create_components(#pipe{nodes = Nodes, attribs = Attribs, stages = Stages}) ->
    Label = Attribs#attrib.label,
    Subtrees = lists:concat([Subtrees || {_StNode, Subtrees} <- Stages]),
    MetricValue = metrics_of_stages([?METRIC(StNode) || {StNode, _Subtrees} <- Stages]), 
    NewAttribs = Attribs#attrib{metric = MetricValue}, 
    Comp = #comp{node_id = hd(Nodes), label = Label, attribs = NewAttribs, type = pipe},
    %%TODO: handle list of nodes in the UI
    [Comp | lists:append([create_components(SubTree)  || SubTree <- Subtrees])];
create_components(#dnc{node = Node, attribs = Attribs, embeds = Subtrees}) ->
    Label = Attribs#attrib.label,
    MetricValue = ?METRIC(Node),
    NewAttribs = Attribs#attrib{metric = MetricValue}, 
    Comp = #comp{node_id = Node, label = Label, attribs = NewAttribs, type = dnc},
    [Comp | lists:append([create_components(SubTree)  || SubTree <- Subtrees])];
create_components(#pool{node = Node, attribs = Attribs, embeds = Subtrees}) ->
    Label = Attribs#attrib.label,
    MetricValue = ?METRIC(Node),
    NewAttribs = Attribs#attrib{metric = MetricValue}, 
    Comp = #comp{node_id = Node, label = Label, attribs = NewAttribs, type = pool},
    [Comp | lists:append([create_components(SubTree)  || SubTree <- Subtrees])].

create_tr_seq_record(Comps) ->
    Tr_seq = #tr_seq{comps = Comps},
    case Comps of
        [] -> Tr_seq;
        [MostOuterComp | _] ->
            Attribs = MostOuterComp#comp.attribs,
            NwSum = Attribs#attrib.nw_sum,
            NewAttribs = Attribs#attrib{num_of_workers = NwSum},
            Tr_seq#tr_seq{attribs = NewAttribs,
                          node_id = MostOuterComp#comp.node_id,
                          label = MostOuterComp#comp.label}
    end.

%% helper functions

create_seq(Node,ParentType) ->
    #seq{parent_type = ParentType, node = Node}.

create_seq_with_tseq(Node, ParentType, Tseq) -> 
    Label = generate_new_label(Node),
    Attribs = #attrib{tseqcpu = Tseq, label = Label, list_length = 1},
    #seq{parent_type = ParentType, node = Node, attribs = Attribs}.

add_label_and_tseq_to_stages([], _Type, []) -> [];
add_label_and_tseq_to_stages([{Node, Subtrees} | Sts], Type, [Tseq | Tseqs]) ->
    H =
        case Subtrees of
            [] -> [create_seq_with_tseq(Node, Type, Tseq)];
            _ -> add_label_and_tseq_to_subtrees(Subtrees, Node, Type)
        end,
    [H | add_label_and_tseq_to_stages(Sts, Type, Tseqs)].
add_label_and_tseq_to_subtrees(Trees, Node, Type) ->   
    Subtrees = 
        case Trees of
            [] -> [create_seq(Node, Type)];
            T -> T
        end,
    lists:map(fun add_label_and_tseq/1, Subtrees).

concat_labels(Strings, Separator) ->
    concat_labels(Strings, Separator, "").
concat_labels([String], _Separator, Acc) ->
    Acc ++ String; 
concat_labels([String | List], Separator, Acc) ->
    concat_labels(List, Separator, Acc ++ String ++ Separator).

concat_labels_of_trees(Trees) ->    
    concat_labels(lists:map(fun (Tree) -> 
                                    Attribs = get_attrib(Tree), 
                                    Attribs#attrib.label
                            end ,Trees), 
                  ",").

tseq_sum_of_trees(Trees) ->
    Tseqs = lists:map(fun (Tree) ->
			      Attribs = get_attrib(Tree),
			      Attribs#attrib.tseqcpu
		      end, Trees),
    case lists:member(undefined, Tseqs) of
	true -> undefined;
	false -> lists:sum(Tseqs)
    end.
	     
tpar_sum_of_trees(Trees) ->
    lists:sum(lists:map(fun (Tree) -> Attrib = get_attrib(Tree),
                                      Attrib#attrib.tparcpu
                        end, Trees)).

get_attrib(Tree) ->
    case Tree of
        #farm{attribs = Attribs} -> Attribs;
        #pipe{attribs = Attribs} -> Attribs;
	#dnc{attribs = Attribs} -> Attribs;
	#pool{attribs = Attribs} -> Attribs;
        #seq{attribs = Attribs}  -> Attribs
    end.

get_node(Tree) ->
    case Tree of
        #farm{node = Node} -> Node;
        #pipe{nodes = Nodes} -> Nodes;
	#dnc{node = Node} -> Node;
	#pool{node = Node} -> Node;
        #seq{node = Node}  -> Node
    end.

generate_new_label({_ , Type, Id}) ->
    [hd(atom_to_list(Type)) | integer_to_list(Id)].


length_of_stream(Node, Type) ->
    ListNode =
        try
            case Type of
                func -> none;
                listcomp -> extract_listexpr(Node, listcomp);
                mapapp -> extract_listexpr(Node, mapapp)
            end
        catch _:_ ->
                none
        end,
    DefaultLength = get(default_length),
    case ListNode of
        none -> DefaultLength;
        _ ->
            case refcore_callanal:listcons_length(ListNode) of
                incalculable -> DefaultLength;
                Is -> Length = lists:max([-1*I-1 || {I} <- Is] ++
                                            [I || I <- Is, is_integer(I)]),
                      case Length of
                          0 -> DefaultLength;
                          _ -> Length
                      end
            end
    end.

extract_listexpr(Node, listcomp) ->
    ListGens =
    [Expr ||
        Expr <- ?Query:exec(Node, ?Query:seq(?Expr:clauses(), ?Clause:body())),
        ?Expr:type(Expr) == list_gen],
    case ListGens of
        [ListGen] ->
            [ListExpr] =
                ?Query:exec(ListGen, ?Query:seq(?Expr:clause(2), ?Clause:body(1))),
            ListExpr;
        _ -> none
    end;
extract_listexpr(Node, mapapp) ->
    Arglist = ?Expr:fun_app_args(Node),
    ?Query:exec1(Arglist, ?Expr:child(2), bad_node).

is_rewrite_embedding(_Node, _Subtrees) ->
    false.
    %% FirstSubtree = hd(Subtrees),
    %% N2 = get_node(FirstSubtree),
    %% case FirstSubtree of
    %%     #seq{} -> false;
    %%     _ -> Node2 =
    %%              case is_list(N2) of
    %%                  true -> hd(N2);
    %%                  _    -> N2
    %%              end,
    %%          length(Subtrees) == 1 andalso Node2 == Node
    %% end.

max_nw_sum(Trees) ->
    lists:max(lists:map(fun (Tree) -> 
                                Attribs = get_attrib(Tree), 
                                Attribs#attrib.nw_sum
                        end, Trees)).

floor(X) ->
    T = erlang:trunc(X),
    case (X - T) of
        Neg when Neg < 0 -> T - 1;
        Pos when Pos > 0 -> T;
        _ -> T
    end.

rec_call_nodes(Fun) ->
    Fundef = ?Query:exec(Fun, ?Fun:definition()),
    Exprs = ?Query:exec([Fundef], ?Query:seq(?Form:clauses(),?Clause:exprs())),
    lists:flatten([?Query:exec(Fun, ?Fun:applications(Expr)) || Expr <- Exprs]).
    
calc_metric(Node) ->
    calc_rec_metric(Node) + 10*refusr_metrics:icc(Node) 
	+ (syn_metric(Node) + 1).

calc_rec_metric(Node)->
    case Node of 
	{'$gn', expr, _} ->
	    FNodes = ?Query:exec(Node, ?Expr:functions());
	_ -> FNodes = [Node]
    end,
    lists:sum(lists:map(fun rec_metric/1, FNodes)).

rec_metric(Node) -> 
    case refpp_predicate_lib:is_recursive_opt(Node) of
	true ->
	    50;
	false ->
	    0
    end.

syn_metric({'$gn', func, _} = Fun) ->
    [FunDef] = ?Query:exec(Fun, ?Fun:definition()),
    syn_metric(FunDef);
syn_metric(Node) ->
    SynCh = [ N || {L, N} <- ?Syn:children(Node), 
		   L /= elex, L /= flex, L/=clex],
    length(SynCh) + lists:sum(lists:map(fun syn_metric/1, SynCh)).

metrics_of_stages(MetricValues) ->
    %lists:sum(MetricValues) - length(MetricValues) + 1.
    lists:max(MetricValues).

validTseq(Tree) -> 
    #attrib{tseqcpu = Tseqcpu} = get_attrib(Tree),
    case Tseqcpu of
	undefined -> false;
	_ -> true
    end.    
