%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Eötvös Loránd University.
%%% Portions created  by Eötvös Loránd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Eötvös Loránd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.
%%%
%%% @author Viktoria Fordos <f-viktoria@elte.hu>
%%% @doc
%%% <b>Datasource</b>
%%% <b>It's important to keep the encoding of this file UTF-8!</b>
%%% @see https://developers.google.com/chart/interactive/docs/dev/implementing_data_source

-module(refpp_datasource).
-include("referl_paraphrase.hrl").

-define(DefResponseHandler, "google.visualization.Query.setResponse").
-define(CsvTextDelimiter, $`).
-define(CsvFieldDelimiter, $,).
-define(CsvRowDelimiter, $\n).

-export([handle/1]).


%%% ============================================================================
%%% Interface

handle(ParsedRequest)->
    % Example: tqx=version:0.6;reqId:1;sig:5277771;out:json; responseHandler:myQueryHandler
    try
        case proplists:get_value("tqx", ParsedRequest) of
            undefined ->
                bad_req();
            Tqx ->
                Query = proplists:get_value("tq", ParsedRequest, ""),
                process_req(Tqx, Query)
        end
    catch
        _:E -> ?d(E),?d(erlang:get_stacktrace()),bad_req()
    end.

%%% ============================================================================
%%% Internals

bad_req()->
    {json, string:join([?DefResponseHandler, "(",
                       convert_to_json([{status, error},
                                       {errors,[{reason,invalid_request}]}]),
                       ")"],"")}.

process_req(Tqx, Query)->
    case is_mandatory_params_given(Tqx) of
        true ->
            ReqParams = reqstr_to_proplist(Tqx),
            QueryParams = querystr_to_proplist(Query),
            generate_response(ReqParams, QueryParams);
        false ->
            bad_req()
    end.

generate_response(ReqParams, QueryParams)->
    case proplists:get_value(out, ReqParams, "json") of
        "json" ->
            generate_json_response(ReqParams, QueryParams);
        "csv" ->
            generate_csv_response(ReqParams, QueryParams);
        "html" ->
            generate_xml_response(ReqParams, QueryParams)
    end.

generate_xml_response(_ReqParams, _QueryParams)->
    {xml, generate_xml_document()}.

generate_xml_document()->
    UnFormattedRows = ?Persistence:query_data(tr_seq, 0),
    ColumnsInfo = make_key(?Persistence:columns_info(tr_seq)),
    Elements = [create_tr_seq_elements(ColumnsInfo, UnFormattedRows)],
    transform_to_be_delivered(xmerl:export_simple(Elements, xmerl_xml)).

make_key(ColumnsInfo)->
    {Labels, _} = lists:unzip(ColumnsInfo),
    lists:map(fun(Label)->
        SafeLabel = lists:filter(fun(Char)->
            (Char>=$a) andalso (Char=<$z)
        end,lists:flatten(string:tokens(string:to_lower(Label)," "))),
        list_to_atom(SafeLabel)
    end, Labels).

create_tr_seq_elements(ColumnsInfo, UnFormattedRows)->
    GenTime = calendar:now_to_universal_time(now()),
    {tr_seq_entities,
     [{timestamp, ?MISC:any_to_string(GenTime)},
      {db_hash, ?MISC:any_to_string(?MISC:database_hash())}],
     create_tr_seq_elements(ColumnsInfo, UnFormattedRows, [])}.

create_tr_seq_elements(_, [], Generated)->
    Generated;
create_tr_seq_elements(ColumnsInfo, [Row|UnFormattedRows], Generated)->
    Entity = create_tr_seq_element(ColumnsInfo, Row),
    create_tr_seq_elements(ColumnsInfo, UnFormattedRows, Generated++[Entity]).

create_tr_seq_element(ColumnsInfo, Row=[ID|_])->
    {tr_seq_entity, lists:zipwith(fun(ColI,Val)->
        {ColI, ?MISC:any_to_string(Val)}
    end, ColumnsInfo, Row), create_tr_comp_elements(ID)}.

create_tr_comp_elements(ID)->
    UnFormattedRows = ?Persistence:query_data(tr_comp, ID),
    ColumnsInfo = make_key(?Persistence:columns_info(tr_comp)),
    create_tr_comp_elements(ColumnsInfo, UnFormattedRows).

create_tr_comp_elements(ColumnsInfo, UnFormattedRows) ->
    [{tr_comp_entities, [], create_tr_comp_elements(ColumnsInfo, UnFormattedRows, [])}].

create_tr_comp_elements(_, [], Generated)->
    Generated;
create_tr_comp_elements(ColumnsInfo, [Row|UnFormattedRows], Generated)->
    Entity = create_tr_comp_element(ColumnsInfo, Row, length(Generated)),
    create_tr_comp_elements(ColumnsInfo, UnFormattedRows, Generated++[Entity]).

create_tr_comp_element(ColumnsInfo, Row, ID) ->
    {tr_comp_entity, [{id, ID} | lists:zipwith(fun(ColI,Val)->
        {ColI, ?MISC:any_to_string(Val)}
    end, ColumnsInfo, Row)],[]}.

generate_csv_response(_ReqParams, QueryParams)->
    {csv, generate_csv_table(QueryParams)}.

generate_csv_table({EntityType, EntityID})->
    UnFormattedRows = ?Persistence:query_data(EntityType, EntityID),
    ColumnsInfo = ?Persistence:columns_info(EntityType),
    generate_csv_table0(ColumnsInfo,UnFormattedRows).

generate_csv_table0(ColumnsInfo, Rows)->
    {Labels, _Types} = lists:unzip(ColumnsInfo),
    string:join(generate_csv_rows([Labels]++Rows, []),[?CsvRowDelimiter]).

generate_csv_rows([], Generated)->
    Generated;
generate_csv_rows([Row|Others], Generated)->
    generate_csv_rows(Others, Generated ++
    [string:join(lists:map(
        fun(Cel) when is_number(Cel); is_atom(Cel) ->
                io_lib:format("~p", [Cel]);
            (Cel=[C|_]) when is_integer(C) ->
                [?CsvTextDelimiter,
                  transform_to_be_delivered(remove_reserved_chars(Cel)),
                 ?CsvTextDelimiter];
            ([Cel=[C|_]]) when is_integer(C) ->
                [?CsvTextDelimiter,
                  transform_to_be_delivered(remove_reserved_chars(Cel)),
                 ?CsvTextDelimiter];
            ([])->
                [?CsvTextDelimiter,?CsvTextDelimiter];
            (Else)->
                ?d(Else),
                [?CsvTextDelimiter,
                 "error_while_exporting_the_value",
                 ?CsvTextDelimiter]
        end,Row),[?CsvFieldDelimiter])]).

generate_json_response(ReqParams, QueryParams)->
    Handler = proplists:get_value(responseHandler, ReqParams,
                                  ?DefResponseHandler),
    ReqID = proplists:get_value(reqId, ReqParams),
    Response0 =
        [{status, <<"ok">>},
         % This is the version of the implemented API
         {version, <<"0.6">>},
         {table, generate_json_table(QueryParams)}],
    % If and only if the reqID is given, we have to send it back
    Response = case ReqID of
        undefined -> Response0;
        _ -> Response0 ++ [{reqId, list_to_integer(ReqID)}]
    end,
    {json, string:join([Handler, "(", convert_to_json(Response), ")"], "")}.


generate_json_table({EntityType, EntityID})->
    UnFormattedRows = ?Persistence:query_data(EntityType, EntityID),
    ColumnsInfo = ?Persistence:columns_info(EntityType),
    generate_json_table0(ColumnsInfo, UnFormattedRows).

generate_json_table0(ColumnsInfo, UnFormattedRows)->
[{cols, generate_colinfo(ColumnsInfo, [])},
 {rows, generate_rows(ColumnsInfo, UnFormattedRows, [])}].

generate_colinfo([], Generated)->
    Generated;
generate_colinfo([{_, ignore}|Others], Generated)->
    generate_colinfo(Others, Generated);
generate_colinfo([ColumnInfo|Others], Generated)->
    {Label, Type} = ColumnInfo,
    Info = [[{id, unicode:characters_to_binary(Label)},
            {label, unicode:characters_to_binary(Label)},
            {type, Type}]],
    generate_colinfo(Others, Generated ++ Info).

generate_rows(_, [], Generated)->
    Generated;
generate_rows(ColumnsInfo, [UnFormattedRow|Others], Generated)->
    RowWithInfo = lists:zip(ColumnsInfo, UnFormattedRow),
    generate_rows(ColumnsInfo, Others,
                  Generated++[[{c, generate_cels(RowWithInfo, [])}]]).

generate_cels([], Generated)->
    Generated;
generate_cels([{{_,ignore}, _} | Cels], Generated)->
    generate_cels(Cels, Generated);
generate_cels([{{_,string}, Val} | Cels], Generated)->
    generate_cels(Cels, Generated++[[{v, unicode:characters_to_binary(Val)}]]);
generate_cels([{{_,boolean}, Val} | Cels], Generated)->
    generate_cels(Cels, Generated++[[{v, Val}]]);
generate_cels([{{_,number}, Val} | Cels], Generated)->
    generate_cels(Cels,
                  Generated++[[{v, Val},
                               {f, list_to_binary(io_lib:format("~p",[Val]))}]]).

reqstr_to_proplist(Tqx)->
    Params = string:tokens(Tqx, ";"),
    lists:foldl(fun(PStr, ParamsIn)->
        [Key, Val] = string:tokens(PStr, ":"),
        ParamsIn ++ [{list_to_atom(Key), Val}]
    end, [], Params).

querystr_to_proplist(Query)->
    % "WHERE kind = 'tr_seq' AND id = 5"
    case re:run(Query,
                "WHERE.*kind[\s]*=[\s]*'([a-z_]+)'.*id[\s]*=[\s]*([0-9]+)",
                [{capture,[1,2],list}]) of
        nomatch ->
            {tr_seq, 0};
        {match, [Kind, ID]} ->
            {list_to_atom(Kind),list_to_integer(ID)}
    end.

convert_to_json(Structure)->
    google_json:encode(Structure).

remove_reserved_chars(Cel)->
    lists:filter(
        fun(Char)->
            not lists:member(Char,[?CsvTextDelimiter,
                                   ?CsvFieldDelimiter,
                                   ?CsvRowDelimiter])
        end, Cel).

is_mandatory_params_given(_Tqx)->
    true.
% If the client requests for csv or html, then it doesn't send reqID
%    MandParams = ["reqId"],
%    lists:all(fun(P)->
%        string:str(Tqx, P) > 0
%    end, MandParams).

% temp fix to make export feature working again
transform_to_be_delivered(List)->
[case I of
        _ when is_integer(I), I > 255 -> $!;
        _ -> I
    end || I<-lists:flatten(List)].
