%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Eötvös Loránd University.
%%% Portions created  by Eötvös Loránd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Eötvös Loránd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.
%%%
%%% @author Viktoria Fordos <f-viktoria@elte.hu>
%%% @doc
%%% <b>Output handler and formatter module</b>
%%%
%%% @todo doc!

%% TODOs:
% - error handling
% - handle the case when the user've just closed the window without requesting a tr_seq
% - handle the case when there is an internal server error
% - make proper result format

-module(refpp_output).

-include("referl_paraphrase.hrl").
-compile(export_all).
% Interface functions for storing and showing the results
-export([display/0, display/1]).
% Interface functions for querying and rendering the results
-export([query_data/2, columns_info/1, has_stored_results/0]).
% Error handling
-export([error_text/2]).
% Handling user req.
-export([save_user_req/1, user_launched_session/0, user_aborted_session/0]).

-define(TrSeqTab, transformation_sequences_table).
-define(TrComponentTab, transformation_components_table).
-define(UserReqTab, requested_trseq_table).

-define(TrSeqTabPath, filename:join([?MISC:data_dir(),
                   "transformation_sequences_table"++
                   ?MISC:graph_based_postfix()])).

-define(TrComponentTabPath, filename:join([?MISC:data_dir(),
                        "transformation_components_table"++
                        ?MISC:graph_based_postfix()])).

-define(UserReqTabPath, filename:join([?MISC:data_dir(),
                        "requested_trseq_table"++
                        ?MISC:graph_based_postfix()])).

-define(UserReqWatcher, user_req_watcher_proc).


-record(tr_seq_tab_entry,{index              :: non_neg_integer(),
                         label               :: nonempty_string(),
                         mod                 :: nonempty_string(),
                         func                :: nonempty_string(),
                         arity               :: non_neg_integer(),
                         num_of_workers      :: non_neg_integer(),
                         %tseqcpu = 0.0      :: float(),
                         %tseqgpu = 0.0      :: float(),
                         %tparcpu = 0.0      :: float(),
                         %tpargpu = 0.0      :: float(),
                         speedupcpu = 1.0    :: float(),
                         %speedupgpu = 1.0    :: float(),
                         metric = 0.0        :: non_neg_integer(),
                         recommended = false :: boolean()}).

-record(tr_comp_tab_entry,{tr_seq_index      :: non_neg_integer(),
                           raw,
                           label             :: nonempty_string(),
                           loc_data          :: nonempty_string(),
                           expr_text         :: nonempty_string(),
                           num_of_workers    :: non_neg_integer(),
                           tseqcpu = 0.0     :: float(),
                           %tseqgpu = 0.0     :: float(),
                           tparcpu = 0.0     :: float(),
                           %tpargpu = 0.0     :: float(),
                           speedupcpu = 1.0  :: float(),
                           %speedupgpu = 1.0  :: float(),
                           metric = 0.0      :: non_neg_integer(),
                           list_length = 0 :: non_neg_integer()}).


%%% ============================================================================
%%% Interface

% @doc Show previously stored results via the web-based interface
display()->
    init_displaying_medium().

% @doc Drop any previously stored result, stores the new given results,
% then shows it via the web-based interface.
display([])->
    % todo!
    display0([]);

% @doc Temporal solution to asynchronously display previously stored results.
% It's required because PC Browser is synchronous.
% Made for EUC, should be removed after that.
% @todo Remove this func. clause!
display(asynch) ->
    spawn(fun ?MODULE:display/0),
    ok;

display(TrSeqs=[HTrSeq|_]) when is_record(HTrSeq, tr_seq) ->
    display0(TrSeqs);

display(_)->
    throw(?LocalErr0r(bad_data_type)).

has_stored_results()->
    query_table(?TrSeqTab,
                fun(E=#tr_seq_tab_entry{})->
                        true;
                    (_)->
                        continue
                end) =:= true.

query_data(tr_seq, 0)->
    query_table(?TrSeqTab,
                fun(E=#tr_seq_tab_entry{})->
                        {continue, lists:nthtail(1,tuple_to_list(E))};
                    (_)->
                        continue
                end);
query_data(tr_seq, TrSeq)->
    query_table(?TrSeqTab,
                fun(E=#tr_seq_tab_entry{index = ATrSeq}) when ATrSeq == TrSeq ->
                        E;
                    (_)->
                        continue
                end);
query_data(tr_comp, TrSeqId)->
    query_table(?TrComponentTab,
                fun(E = #tr_comp_tab_entry{tr_seq_index = ATrSeqId})
                    when TrSeqId == ATrSeqId ->
                        {continue, lists:nthtail(2,tuple_to_list(E))};
                    (_)->
                        continue
                end).

columns_info(tr_seq)->
    [{"ID", number}, {"Configuration", string}, {"Module", string},
     {"Function", string}, {"Arity", number}, {"Number of workers", number},
     %{"Sequential CPU time", number}, {"Sequential GPU time", number},
     %{"Parallel CPU time", number}, {"Parallel GPU time", number},
     {"Expected speedup (CPU)", number}, %{"Expected speedup (GPU)", number},
     {"Metric", number}, {"Recommended?", boolean}];

columns_info(tr_comp)->
    [{"raw", ignore},
     {"Configuration", string}, {"Location information", string},
     {"Program text", string}, {"Number of workers", number},
     {"Sequential CPU time", number}, %{"Sequential GPU time", number},
     {"Parallel CPU time", number}, %{"Parallel GPU time", number},
     {"Expected speedup (CPU)", number}, %{"Expected speedup (GPU)", number},
     {"Metric", number}, {"Used stream length", number}].

%%% ============================================================================
%%% Error handling

error_text(bad_data_type, _)->
    "The given input has wrong data type".

%%% ============================================================================
%%% Save user request

save_user_req(TrSeq)->
    open_tab(?UserReqTab),
    empty_tab(?UserReqTab),
    insert_tab(TrSeq),
    close_tab(?UserReqTab),
    user_finished(),
    ok.

raw(T) ->
    #tr_comp_tab_entry{raw=Raw} = list_to_tuple([tr_comp_tab_entry,-1|T]),
    Raw.

user_launched_session()->
    start_user_req_watcher() ! user_launched_session.

user_aborted_session()->
    start_user_req_watcher() ! aborted.

user_finished()->
    [_, TRComps] = query_table(?UserReqTab, doesnotmatterwhatiwritehere),
    try 
        ?silent([paraphrase(raw(T)) || T <- TRComps])
    catch 
        _A:_B ->
            ?d({_A, _B}),
            ok
            %spawn(?REFERL_NODE, io, format, ["\n The transformation can not be performed automatically. Please, try it manually!\n "])        
    end, 
    start_user_req_watcher() ! finished.

paraphrase({Node, farm, NW}) ->
    try
        ri:transform(rules, [{node, Node}, {name, introduce_farm}, {params, [NW]}])
    catch 
        _A:_B ->
            error({_A, _B})
    end;
paraphrase({_Node, seq, _NW}) ->
    ok;
paraphrase(X = {_Node, _SkelType, _NW}) ->
    ?out("\n The transformation can not be performed automatically. Please, try it manually!\n"),
    ok.%?d({paraphrse, Node, SkelType}).

%%% ============================================================================
%%% Internals

start_user_req_watcher()->
    case whereis(?UserReqWatcher) of
        undefined ->
            ToSend = self(),
            PID = spawn_link(fun()-> user_req_watcher_loop(idle, ToSend) end),
            register(?UserReqWatcher,PID),
            PID;
        PID ->
            PID
    end.

user_req_watcher_loop(aborted, ToSend) ->
    receive
        user_launched_session ->
            user_req_watcher_loop(user_in_session, ToSend);
        % TODO: is it possible?
        finished ->
            ToSend ! {self(), finish}
        after 5000 ->
            ToSend ! {self(), abort}
    end;

user_req_watcher_loop(idle, ToSend) ->
    receive
        user_launched_session ->
            user_req_watcher_loop(user_in_session, ToSend);
        % TODO: is it possible?
       finished ->
            ToSend ! {self(), finish}
        after 25000 ->
            ToSend ! {self(), abort}
    end;

user_req_watcher_loop(State, ToSend)->
    receive
        user_launched_session ->
            user_req_watcher_loop(user_in_session, ToSend);
        aborted ->
            user_req_watcher_loop(aborted, ToSend);
        finished ->
            ToSend ! {self(), finish}
        after 5000 ->
            case State of
                % PC Browser has been closed for 5 sec.
                aborted -> ToSend ! {self(), abort};
                _ -> user_req_watcher_loop(State, ToSend)
            end
    end.

display0(TrSeqs)->
    persist_data(TrSeqs),
    init_displaying_medium().

persist_data(TrSeqs)->
    reset_tabs(),
    extend_and_store_trseqs(TrSeqs).

init_displaying_medium()->
    Watcher = start_user_req_watcher(),
    Infos = ?WebSrv:start(),
    Url = proplists:get_value(url, ?WEB_SERVICES:service(output)),
    ServerDir = proplists:get_value(server_dir, Infos),
    instantiate_html_template(ServerDir),
    instantiate_css(ServerDir),
    case os:type() of
        {unix, darwin}->
            os:cmd(string:join(["open /Applications/Safari.app ",Url],""));
        {unix, _} ->
            spawn(fun()-> os:cmd(string:join(["firefox ", Url],"")) end);
        _ ->
            error_logger:info_msg(
                "Open a Firefox and enter the following address (~s) to browse pattern candidates~n",
                [Url])
    end,
    receive
        {Watcher, finish} ->
            exit(Watcher, kill),
            query_table(?UserReqTab, unused);
        {Watcher, abort} ->
            exit(Watcher, kill),
            abort
    end.

instantiate_html_template(ServerDir)->
    ToBeReplaced=[
        {"##DATASOURCE##",
            proplists:get_value(url, ?WEB_SERVICES:service(datasource))},
        {"##USERINPUTHANDLER##",
            proplists:get_value(url, ?WEB_SERVICES:service(user_input_handler))}],
    {ok, Source} = file:read_file(raw_html_template_path()),
    Instantiated =
        lists:foldl(fun({Pattern, Url}, Template)->
            re:replace(Template, Pattern, Url, [global])
        end, Source, ToBeReplaced),
    file:write_file(filename:join(ServerDir, "output.html"),
                    Instantiated).

instantiate_css(ServerDir)->
    Csss = filelib:wildcard(string:concat(raw_html_path(), "/*.css")),
    [file:copy(Css, filename:join(ServerDir, filename:basename(Css)))
        || Css <- Csss].



raw_html_path()->
    Path0 = filename:split(filename:dirname(code:which(?MODULE))),
    filename:join(lists:sublist(Path0,length(Path0)-1)++
                  ["html_template"]).

raw_html_template_path()->
    filename:join(raw_html_path(),
                  "output_template.html").


%%% ============================================================================
%%% Persisting data

extend_and_store_trseqs(TrSeqs)->
    [open_tab(Tab) || Tab <- [?TrSeqTab, ?TrComponentTab]],
    extend_and_store_trseqs0(?MISC:index_list(TrSeqs)),
    [close_tab(Tab) || Tab <- [?TrSeqTab, ?TrComponentTab]].

extend_and_store_trseqs0([])->
    ok;
extend_and_store_trseqs0([{TrSeq=#tr_seq{}, TrSeqIndex}|IndexedTrSeqsTail])->
    #tr_seq{attribs=Attribs, comps=Comps, label=Label, node_id=Node} = TrSeq,
    CPUSpeedUp = ratio(Attribs#attrib.tparcpu,Attribs#attrib.tseqcpu),
    GPUSpeedUp = ratio(Attribs#attrib.tpargpu,Attribs#attrib.tseqgpu),
    insert_tab(#tr_seq_tab_entry{
         index = TrSeqIndex,
         label = Label,
         mod = mod(Node),
         func = func(Node),
         arity = arity(Node),
         num_of_workers = Attribs#attrib.num_of_workers,
         %tseqcpu = Attribs#attrib.tseqcpu,
         %tseqgpu = Attribs#attrib.tseqgpu,
         %tparcpu = Attribs#attrib.tparcpu,
         %tpargpu = Attribs#attrib.tpargpu,
         speedupcpu = CPUSpeedUp,
         %speedupgpu = GPUSpeedUp,
         metric = Attribs#attrib.metric,
         recommended = is_recommended(CPUSpeedUp, GPUSpeedUp, 
				      Attribs#attrib.metric)}),
    extend_and_store_trcomps(TrSeqIndex, Comps),
    extend_and_store_trseqs0(IndexedTrSeqsTail).

num_of_ws([#comp{attribs = #attrib{num_of_workers = N}} | Tail], Acc) ->
    num_of_ws(Tail, Acc*N);
num_of_ws([], Acc)->
    Acc.

% this calc. should be done in a lazy manner in the future
extend_and_store_trcomps(_, [])->
    ok;
extend_and_store_trcomps(TrSeqIndex,[Comps = [#comp{}| _]])->
extend_and_store_trcomps(TrSeqIndex,Comps);
extend_and_store_trcomps(TrSeqIndex,[TrComp=#comp{}| TrCompsTail])->
    #comp{node_id=Node, label=Label, attribs=Attribs, type=Type} = TrComp,
    insert_tab(#tr_comp_tab_entry{
       tr_seq_index = TrSeqIndex,
       label = Label,
       raw = {Node, Type, Attribs#attrib.num_of_workers},
       loc_data = loc_data(Node),
       expr_text = expr_text(Node),
       num_of_workers = Attribs#attrib.num_of_workers,
       tseqcpu = Attribs#attrib.tseqcpu,
       %tseqgpu = Attribs#attrib.tseqgpu,
       tparcpu = Attribs#attrib.tparcpu,
       %tpargpu = Attribs#attrib.tpargpu,
       speedupcpu = ratio(Attribs#attrib.tparcpu,Attribs#attrib.tseqcpu),
       %speedupgpu = ratio(Attribs#attrib.tpargpu,Attribs#attrib.tseqgpu),
       metric = Attribs#attrib.metric,
       list_length = Attribs#attrib.list_length}),
    extend_and_store_trcomps(TrSeqIndex, TrCompsTail).

reset_tabs()->
    [begin
        open_tab(Tab),
        empty_tab(Tab),
        close_tab(Tab)
    end || Tab <- [?TrSeqTab, ?TrComponentTab, ?UserReqTab]].

open_tab(?UserReqTab)->
    dets:open_file(?UserReqTabPath,
                   [{type, set}, {keypos,1}]);
open_tab(?TrSeqTab)->
    dets:open_file(?TrSeqTabPath,
                   [{type, set}, {keypos,#tr_seq_tab_entry.index}]);
open_tab(?TrComponentTab)->
    dets:open_file(?TrComponentTabPath,
                   [{type, bag}, {keypos, #tr_comp_tab_entry.tr_seq_index}]).

close_tab(?UserReqTab)->
    dets:close(?UserReqTabPath);
close_tab(?TrSeqTab)->
    dets:close(?TrSeqTabPath);
close_tab(?TrComponentTab)->
    dets:close(?TrComponentTabPath).

empty_tab(?UserReqTab)->
    dets:delete_all_objects(?UserReqTabPath);
empty_tab(?TrSeqTab)->
    dets:delete_all_objects(?TrSeqTabPath);
empty_tab(?TrComponentTab)->
    dets:delete_all_objects(?TrComponentTabPath).
insert_tab(Elem = #tr_seq_tab_entry{})->
    dets:insert(?TrSeqTabPath, Elem);
insert_tab(Elem = #tr_comp_tab_entry{})->
    dets:insert(?TrComponentTabPath, Elem);
insert_tab(TrSeqId) when is_integer(TrSeqId), TrSeqId>0 ->
    dets:insert(?UserReqTabPath, {TrSeqId}).


query_table(?UserReqTab, _)->
    open_tab(?UserReqTab),
    Res = case dets:match_object(?UserReqTabPath, {'_'}) of
        [{TrSeq}] -> [query_data(tr_seq, TrSeq), query_data(tr_comp, TrSeq)];
        [] -> empty;
        % todo error handling
        _ -> error
    end,
    close_tab(?UserReqTab),
    Res;
query_table(?TrSeqTab, Fun)->
    query_table0(?TrSeqTab, ?TrSeqTabPath, Fun);
query_table(?TrComponentTab, Fun)->
    query_table0(?TrComponentTab, ?TrComponentTabPath, Fun).

query_table0(TabName, TabPath, Fun)->
    open_tab(TabName),
    Res = dets:traverse(TabPath, Fun),
    close_tab(TabName),
    Res.
%%% ============================================================================
%%% Queries

mod(Node) when ?IS_NODE(Node), element(2, Node) =:= expr->
    [ModNode] = ?Query:exec(Node,
                            ?Query:seq([?Expr:clause(), ?Clause:form(),
                                       ?Form:module()])),
    io_lib:format("~p", [?Mod:name(ModNode)]);
mod(Node) when ?IS_NODE(Node), element(2,Node) =:= func ->
    [ModNode] = ?Query:exec(Node,?Fun:module()),
    io_lib:format("~p", [?Mod:name(ModNode)]).

func(Node) when ?IS_NODE(Node), element(2, Node) =:= expr->
    [FuncNode] = ?Query:exec(Node,
                             ?Query:seq([?Expr:clause(), ?Clause:form(),
                                        ?Form:func()])),
    io_lib:format("~p", [?Fun:name(FuncNode)]);
func(Node) when ?IS_NODE(Node), element(2, Node) =:= func->
    io_lib:format("~p", [?Fun:name(Node)]).


arity(Node) when ?IS_NODE(Node), element(2, Node) =:= expr->
    [FuncNode] = ?Query:exec(Node,
                             ?Query:seq([?Expr:clause(), ?Clause:form(),
                                        ?Form:func()])),
    ?Fun:arity(FuncNode);
arity(Node) when ?IS_NODE(Node), element(2, Node) =:= func->
    ?Fun:arity(Node).

loc_data(Node) when ?IS_NODE(Node), element(2, Node) =:= expr->
    loc_data0(Node);
loc_data(FuncNode) when ?IS_NODE(FuncNode), element(2, FuncNode) =:= func->
    case ?Query:exec(FuncNode, ?Fun:definition()) of
        [Node] -> loc_data0(Node);
        [] -> lists:flatten([mod(FuncNode),[":"],func(FuncNode),["/"],
                          io_lib:format("~p",[arity(FuncNode)])])
    end.

expr_text(Node) when ?IS_NODE(Node), element(2, Node) =:= expr->
    flat_text0(Node);
expr_text(FuncNode) when ?IS_NODE(FuncNode), element(2, FuncNode) =:= func->
    case ?Query:exec(FuncNode, ?Fun:definition()) of
        [Node] -> flat_text0(Node);
        [] -> lists:flatten([mod(FuncNode),[":"],func(FuncNode),["/"],
                          io_lib:format("~p",[arity(FuncNode)])])
    end.

flat_text0(Node) when ?IS_NODE(Node), (element(2, Node) =:= expr orelse
                                     element(2, Node) =:= form)->
    string:strip(?Syn:flat_text(Node)).

loc_data0(Node) when ?IS_NODE(Node), (element(2, Node) =:= expr orelse
                                     element(2, Node) =:= form)->
    [File] = ?Syn:get_file(Node),
    [First] = ?Query:exec(Node, ?Syn:first_leaf()),
    [Last] = ?Query:exec(Node, ?Syn:last_leaf()),
    string:join(
        io_lib:format("~s~s~p~s~p",
            [?File:path(File),
            " : ",
            ?Token:pos(File,First,linecol),
            " - ",
            ?Token:pos(File, Last, linecol)]),"").

ratio(_, undefined)->
    -1; %undefined;
ratio(undefined, _)->
    -1; %undefined;
ratio(New, Old)->
    Old/New.

is_recommended(CPUSpeedUp, GPUSpeedUp, M) ->
    (CPUSpeedUp > 1.0) orelse (GPUSpeedUp > 1.0) orelse (M > 15).
