%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.
%%%
%%% @author Istvan Bozo  <bozoistvan@elte.hu>
%%% @doc
%%% <b>Predicate library for refactoring rule interpretation.</b>
%%%
%%% @todo doc!

-module(refpp_predicate_lib).

-export([has_base_case/1, has_base_case/2, is_recursive/1, is_recursive/2,
         is_recursive_opt/1, is_multiple_recursive/1, is_multiple_recursive/2,
         is_tail_recursive/1, is_tail_recursive/2, is_stream_processing/1,
         is_list_processing/1, is_list_processing/2, is_empty_list/1,
         is_arg_processed/1, is_recursive_list_construct/2,
         is_divide_and_conquer/1 ]).

-export([return_points/1, head/1, tail/1]).

-export([fun_cfg_info/1, cfg_paths/2]).

-export([bwd_closure/2, fwd_closure/2]).

-compile(export_all).

-include("referl_paraphrase.hrl").

%% types should be moved to the slicer app with the cfg functions
-type cfg_node() :: gnode()  | {ret, gnode()}.

-type cfg_path() :: [cfg_node()].

-type cfg_edge() :: {cfg_node(), cfg_node(), atom() | []}.

%%%%%%%%%%%%%%%%%%%%%% Interface functions %%%%%%%%%%%%%%%%%%%%%%

-spec has_base_case(FunForm :: gnode()) -> boolean().
has_base_case(FunForm) ->
    CFGEdges = fun_cfg_info(FunForm),
    Paths = cfg_paths(FunForm, CFGEdges),
    has_base_case(FunForm, Paths).

-spec has_base_case(FunForm :: gnode(), Paths :: [cfg_path()]) -> boolean().
has_base_case(FunForm, Paths) ->
    lists:any(fun(Path) ->
                      is_base_case_path(Path, FunForm)
              end, Paths).

-spec is_recursive_opt(Node :: gnode()) -> boolean().
is_recursive_opt({'$gn', func, _} = FunNode)->
    FwdCallClsr  = fwd_closure([FunNode], []),
    sets:is_element(FunNode, FwdCallClsr);
is_recursive_opt(FunForm) ->
    [FunNode]    = ?Query:exec(FunForm, ?Form:func()),
    FwdCallClsr  = fwd_closure([FunNode], []),
    sets:is_element(FunNode, FwdCallClsr).

-spec is_recursive(FunForm :: gnode()) -> boolean().
is_recursive(FunForm) ->
    CFGEdges = fun_cfg_info(FunForm),
    Paths = cfg_paths(FunForm, CFGEdges),
    is_recursive(FunForm, Paths).

-spec is_recursive(FunForm :: gnode(), Paths :: [cfg_path()]) -> boolean().
is_recursive(FunForm, Paths) ->
    lists:any(fun(Path) ->
                      is_recursive_path(Path, FunForm)
              end, Paths).

-spec is_multiple_recursive(FunForm :: gnode()) -> boolean().
is_multiple_recursive(FunForm) ->
    CFGEdges = fun_cfg_info(FunForm),
    Paths = cfg_paths(FunForm, CFGEdges),
    is_multiple_recursive(FunForm, Paths).

-spec is_multiple_recursive(FunForm :: gnode(), Paths :: [cfg_path()]) ->
                                   boolean().
is_multiple_recursive(FunForm, Paths) ->
    lists:any(fun(Path) ->
                      contains_multiple_rec_calls(Path, FunForm)
              end, Paths).

-spec is_tail_recursive(FunForm :: gnode()) -> boolean().
is_tail_recursive(FunForm) ->
    CFGEdges = fun_cfg_info(FunForm),
    Paths = cfg_paths(FunForm, CFGEdges),
    is_tail_recursive(FunForm, Paths).

-spec is_tail_recursive(FunForm :: gnode(), Paths :: [cfg_path()]) -> boolean().
is_tail_recursive(FunForm, Paths) ->
    case assort_execution_paths(Paths, [], []) of
        {_BasePaths, RecPaths} ->
            not lists:any(fun(Path) ->
                                  contains_multiple_rec_calls(Path, FunForm)
                          end, RecPaths) andalso
                lists:all(fun(X) -> X end,
                          [is_app_of_fun(return_point_from_path(RC), FunForm)
                           || RC <- RecPaths]);
        _ ->
            false
    end.

-spec is_stream_processing(FunForm :: gnode()) -> boolean().
is_stream_processing(FunForm) ->
    is_every_return_point_enclosed_in_receive(FunForm) andalso
        begin
            CFGEdges = fun_cfg_info(FunForm),
            Paths = cfg_paths(FunForm, CFGEdges),
            case assort_execution_paths(Paths, [], []) of
                {BasePaths, RecPaths} ->
                    Data = [ begin
                                 Ret = return_point_from_path(RC),
                                 {is_app_of_fun(Ret, FunForm), Ret}
                             end || RC <- RecPaths],
                    case compact_analysis_and_info_rec(Data, FunForm) of
                        {true, I} ->
                            RetPs = [ return_point_from_path(BP) ||
                                        BP <- BasePaths],
                            lists:all(
                              fun(R) ->
                                      is_node_identical_to_ith_arg_of_form(R, I)
                              end, RetPs);
                        _ ->
                            false
                    end;
                false ->
                    false
            end
        end.

-spec is_list_processing(FunForm :: gnode()) -> boolean().
is_list_processing(FunForm) ->
    CFGEdges = fun_cfg_info(FunForm),
    Paths = cfg_paths(FunForm, CFGEdges),
    is_list_processing(FunForm, Paths).

-spec is_list_processing(FunForm :: gnode(), Paths :: [cfg_path()]) ->
                                boolean().
is_list_processing(FunForm, Paths) ->
    case assort_execution_paths(Paths, [], []) of
	{BasePaths, RecPaths} ->
	    Res = lists:all(fun is_return_point_empty_list/1, BasePaths) andalso
		lists:all(fun is_appropriate_rec_return_point/1, RecPaths),
            case Res of
                true ->
                    case processed_arg_idx(FunForm) of
                        -1 ->
                            false;
                        I ->
                            Patterns =
                                ?Query:exec(FunForm,
                                            ?Query:seq([?Form:clauses(),
                                                        ?Clause:pattern(I)])),
                            BPatterns = 
                                ?MISC:intersect(Patterns,
                                                lists:append(BasePaths)),
                            lists:all(fun is_empty_list_pattern/1, BPatterns)
                    end;
                _ ->
                    false
            end;
	false -> false
    end.

-spec is_empty_list_pattern(Node :: gnode()) -> boolean().
is_empty_list_pattern(Node) when is_tuple(Node) ->
    CompactFwdReach = refanal_dataflow:reach([Node], [{back, false}], true),
    lists:any(fun is_empty_list_node/1, CompactFwdReach).

-spec is_empty_list(Node :: gnode()) -> boolean().
is_empty_list(Node) when is_tuple(Node) ->
    CompactBackReach = refanal_dataflow:reach([Node], [back], true),
    lists:all(fun is_empty_list_node/1, CompactBackReach).

-spec is_arg_processed(ArgNode :: gnode()) -> boolean().
is_arg_processed(ArgNode) ->
    [Form] = ?Query:exec(ArgNode, ?Query:seq([?Expr:clause(), ?Clause:form()])),
    case processed_arg_idx(Form) of
        -1 -> 
            false;
        I ->
            [{Link, Parent}] = ?Syn:parent(ArgNode),
            Index = ?Syn:index(Parent, Link, ArgNode),
            I == Index
    end.

-spec is_recursive_list_construct(Node :: gnode(), Fun :: gnode()) -> boolean().
is_recursive_list_construct(Node, Fun) ->
    [FunForm] = ?Query:exec(Fun, ?Fun:definition()),
    is_appropriate_list_node(Node, FunForm).

-spec is_divide_and_conquer(FunForm :: gnode()) -> boolean().
is_divide_and_conquer(FunForm) ->
    FunNode      = ?Query:exec(FunForm, ?Form:func()),
    FwdCallClsr  = fwd_closure(FunNode, FunNode),
    BwdCallClsr  = bwd_closure(FunNode, FunNode),
    Intersection =
        sets:del_element(FunNode,
                         sets:intersection(FwdCallClsr, BwdCallClsr)),
%    ElementList = sets:to_list(Intersection),
    Fun = fun(N) ->
                  Closure =
                      fwd_closure_with_dead_end([N], [], FunNode),
                  %% Chains = call_chains(N, FunNode),
                  %% F = fun(Ch) -> lists:member(N, Ch) end,
                  %% lists:any(F, Chains)
                  sets:is_element(N, Closure)
          end,
    sets:size(sets:filter(Fun, Intersection)) > 0.

-spec processed_arg_idx(Form :: gnode()) -> integer().
processed_arg_idx(Form) ->
    CFGEdges = fun_cfg_info(Form),
    Paths = cfg_paths(Form, CFGEdges),
    case assort_execution_paths(Paths, [], []) of
        {_BPaths, RecPaths} ->
            RetPoints = lists:map(fun return_point_from_path/1, RecPaths),
            CompactBackReach = refanal_dataflow:reach(RetPoints, [back], true),
            Data = lists:map(fun is_appropriate_list_node_h/1,
                             CompactBackReach),
            case lists:all(fun is_tuple/1, Data) of
                true ->
                    Res  = [check_right_expr(RExpr, Form) ||
                               {_LExpr, RExpr} <- Data],
                    case fold_result_of_rec_call_analysis(Res) of
                        {true, [I]} ->
                            I;
                        _ ->
                            -1
                    end;
                false ->
                    -1
            end;
        _ ->
            -1
    end.

-spec return_points(FunForm :: gnode()) -> [gnode()].
return_points(FunForm) ->
    CFGEdges = fun_cfg_info(FunForm),
    Paths = cfg_paths(FunForm, CFGEdges),
    [return_point_from_path(Path) || Path <- Paths].

-spec tail(Node :: gnode()) -> [gnode()].
tail(Node) ->
    FwdReach = ?Dataflow:reach([Node], [{back, false}], false),
    SelT = ?Query:exec(FwdReach, [sel_t]),
    case SelT of
        [] -> [];
        _List ->
            Reach = ?Dataflow:reach(SelT, [{back, false}], false),
            [Form] = ?Query:exec(Node, ?Query:seq([?Expr:clause(),
                                                   ?Clause:form()])),
            CFGEdges = fun_cfg_info(Form),

            %% TODO: refine patterns in CFG
            SelTP = ?Query:exec(SelT, [{sel_t, back}]),
            CFGClosure =
                closure(SelTP, sets:from_list(SelT),
                        fun(N) ->
                                [ E || {S,E,_} <- CFGEdges, S == N]
                        end, sets:new()),

            ?MISC:intersect(sets:to_list(CFGClosure), Reach)
    end.

-spec head(Node :: gnode()) -> [gnode()].
head(Node) ->
    FwdReach = ?Dataflow:reach([Node], [{back, false}], false),
    SelH = ?Query:exec(FwdReach, [sel_h]),
    ?Dataflow:reach(SelH, [{back, false}], false).

-spec cfg_paths(StartNode :: cfg_node(), Edges :: [cfg_edge()]) -> [cfg_path()].
cfg_paths(StartNode, Edges) ->
    Tab = ets:new(cfg, [duplicate_bag]),
    ets:insert(Tab,Edges),
    Paths = get_paths(Tab, _InitPaths = [[StartNode]], _Ready = []),
    ets:delete(Tab),
    neglect_error_paths(Paths).

-spec fun_cfg_info(Form :: gnode()) -> [cfg_edge()].
fun_cfg_info(Form) ->
    {Edges, _Apps} = refsc_cfg_utils:build_cfg(Form),
    Edges.

%%%%%%%%%%%%%%%%%%%%%% Helper functions %%%%%%%%%%%%%%%%%%%%%%

is_every_return_point_enclosed_in_receive(FunForm) ->
    [FunNode]    = ?Query:exec(FunForm, ?Form:func()),
    ReturnPoints = ?Query:exec(FunNode, ?Fun:return_points(FunNode)),

    IsPathWithRec =
        fun(Path) ->
                lists:any(fun({_, N}) -> is_receive_expr(N) end, Path)
        end,
    lists:all(IsPathWithRec, [ ?Syn:root_path(RP) || RP <- ReturnPoints]).

is_receive_expr(Node) ->
    ?Graph:class(Node) == expr andalso
        ?Expr:type(Node) == receive_expr.

is_node_identical_to_ith_arg_of_form(Node, I) ->
    IthArgs =
        ?Query:exec(Node, ?Query:seq([?Expr:clause(),?Clause:funcl(),
                                      ?Clause:pattern(I)])),
    ArgFwdReach = ?Dataflow:reach(IthArgs, [{back, false}], false),
    lists:member(Node, ArgFwdReach).



%% helper function for stream processing map-like function recursive cases
compact_analysis_and_info_rec(Data, FunForm) ->
    case  lists:all(fun({E, _}) -> E end, Data) of
        true ->
            A = [ analyse_rec_call(RetP, FunForm) || {_, RetP} <- Data],
            case fold_result_of_rec_call_analysis(A) of
                {true, [I]} ->
                    {lists:all(fun({_, B}) ->
                                       is_accumulating_in_ith_arg(B, I)
                               end, Data),
                     I};
                _ ->
                    false
            end;
        false ->
            false
    end.


%% The function is to determine wether an application is called with
%% the ith argument used as accumulator. The function returns true if
%% the ith argument of the applications is of form [NewVal | Acc] or
%% [NewVal] ++ Acc, otherwise false. Acc is a list coming unchanged
%% from the argument list of function definition.
is_accumulating_in_ith_arg(App, AccArgIdx) ->
    AppIthArg = ?Query:exec(?Expr:fun_app_args(App), ?Expr:child(AccArgIdx)),
    BackReach = ?Dataflow:reach(AppIthArg, [back], false),

    IthArgs =
        ?Query:exec(App, ?Query:seq([?Expr:clause(),?Clause:funcl(),
                                      ?Clause:pattern(AccArgIdx)])),
    IthData =
        ?Query:exec(?Dataflow:reach(IthArgs, [{back, false}], false), [cons_t]),
    ?MISC:intersect(BackReach, IthData) /= [].

% map-like list processing base case
is_return_point_empty_list(Path) when is_list(Path) ->
    Last = return_point_from_path(Path),
    is_empty_list(Last).

is_empty_list_node(Node) ->
    ?Expr:type(Node) == cons andalso ?Graph:path(Node, [esub]) == [].

% map-like list processing recursive case
is_appropriate_rec_return_point(Path) ->
    Last = return_point_from_path(Path),
    is_appropriate_rec_list_node(Last, hd(Path)).

is_appropriate_rec_list_node(Node, Form) ->
    CompactBackReach = ?Dataflow:reach([Node], [back], true),
    lists:all(fun(N) -> is_appropriate_list_node(N, Form) end,
              CompactBackReach).

%% accepts only: [Res | rec_call(...)] and [Res] ++ rec_call(...)  it
%% returns the corresponding {cons, Res, rec_call()} or {infix_expr,
%% Res, rec_call()} tuples and false otherwise
is_appropriate_list_node_h(Node) ->
    case ?Expr:type(Node) of
        cons ->
            case ?Graph:path(Node, [esub]) of
                [Head, Tail] ->
                    {Head, Tail};
                _ ->
                    false
            end;
        infix_expr ->
            case ?Expr:value(Node) of
                '++' ->
                    [Sub1, Sub2] = ?Graph:path(Node, [esub]),
                    case ?Query:exec(Sub1, ?Expr:children()) of
                        [SubSub1] ->
                            {SubSub1, Sub2};
                        _ ->
                            false
                    end;
                _  ->
                    false
            end;
        _ ->
            false
    end.

%% accepts only: [Res | rec_call(...)] and [Res] ++ rec_call(...)
is_appropriate_list_node(Node, Form) ->
    case is_appropriate_list_node_h(Node) of
        {LeftExpr, RightExpr} ->
            {Res, ProcArgIdx} = check_right_expr(RightExpr, Form),
            Res andalso is_singleton(ProcArgIdx) andalso
                check_left_expr(LeftExpr, Form, hd(ProcArgIdx));
        _ -> false
    end.

return_point_from_path(Path) ->
    [Last | _] = lists:dropwhile(fun(Node) -> element(1, Node) == ret end,
                                 lists:reverse(Path)),
    Last.
%% TODO: more sophisticated analysis
check_left_expr(Node, _Form, ProcArgIdx) ->
    Subs = ?Query:exec(Node, ?Expr:deep_sub()),
    DataBackReach = refanal_dataflow:reach(Subs, [back], false),

    IthArgs =
        ?Query:exec(Node, ?Query:seq([?Expr:clause(),?Clause:funcl(),
                                      ?Clause:pattern(ProcArgIdx)])),
    IthData = ?Query:exec(?Dataflow:reach(IthArgs, [{back, false}], false),
                          [sel_t]),

    IntSect = ?MISC:intersect(DataBackReach, IthData),
    IntSect == [].

check_right_expr(Node, Form) ->
    DataBackReach =
        refanal_dataflow:reach([Node], [back], false),

    RecCalls = lists:filter(fun(N) -> is_app_of_fun(N, Form) end,
                            DataBackReach),
    RecCallResults =
        [ analyse_rec_call(RC, Form) || RC <- RecCalls],
    fold_result_of_rec_call_analysis(RecCallResults).

is_app_of_fun(Node, Form) ->
    ?Expr:type(Node) == application andalso
        ?Query:exec(Node,?Query:seq([?Expr:dynfunction(),
                                     ?Fun:definition()])) == [Form].

%% Evaluates the result of the analyse_rec_call. The input is a list
%% of results of analyse_rec_call/2 function. The funnction returns
%% {B, List} if the
fold_result_of_rec_call_analysis(RecCallResults) ->
    EvalFun =
        fun({true, [I]}, {_, [I]}) ->
                {true, [I]};
           ({false, _}, {_, _}) ->
                {false, []};
           (_, _) ->
                {false, []}
        end,

    lists:foldl(EvalFun, hd(RecCallResults), tl(RecCallResults)).

%% Checks whether the recursive call is appropriate, that is every
%% argument is identical to the original arguments, and only one
%% differs from it (further analysis is postponed to following steps
%% whether this ?list?  is consumed properly). The function returns a
%% two-tuple: first element is a boolean value (true if the above
%% conditions hold), the second element is a list of indeces (if the
%% first is true, it is a singleton list, otherwise it should not be
%% used for further processing)
analyse_rec_call(App, _Form) ->
    AppArgs = ?Query:exec(?Expr:fun_app_args(App), ?Expr:children()),

    ArgsBackReach =
        [ ?Dataflow:reach([A], [back], false) || A <- AppArgs ],

    Cl = ?Query:exec(App, ?Query:seq([?Expr:clause(),?Clause:funcl()])),
    PatternList = ?Query:exec(Cl, ?Clause:patterns()),
    %% FunCls  = ?Query:exec(Form, ?Form:clauses()),
    %% PatternLists = [ ?Query:exec(F, ?Clause:patterns()) || F <- FunCls],

    %% Diffs = [ diff_indices(PL, ArgsBackReach, 1) || PL <- PatternLists],
    Diffs = diff_indices(PatternList, ArgsBackReach, 1),
    FlatUniqueIdxs = lists:usort(lists:flatten(Diffs)),

    {is_singleton(FlatUniqueIdxs), FlatUniqueIdxs}.

diff_indices([], [], _) -> [];
diff_indices([P|Ps], [A | As], N) ->
    PD = ?Dataflow:reach([P], [{back, false}], false),
    case ?MISC:intersect(PD, A) of
        [] ->
            [N | diff_indices(Ps, As, N+1)];
        _ ->
            diff_indices(Ps, As, N+1)
    end.

is_singleton([_]) ->
    true;
is_singleton(_) ->
    false.

-define(ReturnLabels, [ret, new_val]).

assort_execution_paths([], BaseCPths, RecPths) ->
    {BaseCPths, RecPths};
assort_execution_paths([Path | Paths], BaseCPths, RecPths) ->
    case is_base_case_path(Path, hd(Path)) of
        true ->
            assort_execution_paths(Paths, [Path | BaseCPths], RecPths);
        false ->
            case is_recursive_path(Path, hd(Path)) of %% mutually rec???
                true ->
                    assort_execution_paths(Paths, BaseCPths, [Path | RecPths]);
                X ->
                    X
            end
    end.

is_recursive_path(Path, FunForm) ->
    [FunNode] = ?Query:exec(FunForm, ?Form:func()),
    FunsCalledOnPath = called_functions(Path), %% not flat list

    lists:any(fun(FunGroup)->
                      FunClosureOnPath = fwd_closure(FunGroup, FunGroup),
                      sets:is_element(FunNode, FunClosureOnPath)
              end,FunsCalledOnPath).

% assumed that it is recursive
contains_multiple_rec_calls(Path, FunForm) ->
    contains_multiple_rec_calls_base(Path, [FunForm]).

contains_multiple_rec_calls_base(Path, FunForms) ->
    FunNodes = ?Query:exec(FunForms, ?Form:func()),
    FunsCalledOnPath = called_functions(Path), %% not flat list
    FunClosuresOnPath =
        [fwd_closure(FunGroup, FunGroup)
         || FunGroup <- FunsCalledOnPath],
    MemberInGroups =
        lists:filter(fun(Set) ->
                             lists:any(fun(FunNode) ->
                                               sets:is_element(FunNode, Set)
                                       end, FunNodes)
                     end,
                     FunClosuresOnPath),
    case MemberInGroups of
        [] -> false;
        [G] ->
            Funs = [F || {F, Group} <- lists:zip(FunsCalledOnPath,
                                                 FunClosuresOnPath),
                         Group == G],
            lists:any(fun(Fun) ->
                         [FunDef] =
                                  ?Query:exec(Fun, ?Fun:definition()),
                         case lists:member(FunDef, FunForms) of
                             true ->
                                 false;
                             false ->
                                 contains_multiple_rec_calls_deep(FunDef,
                                                                  [FunDef
                                                                   | FunForms])                         end
                      end, Funs);
        _N ->
            true
    end.

contains_multiple_rec_calls_deep(FunForm, PrevFunForms) ->
    CFGEdges = fun_cfg_info(FunForm),
    Complex = length([1 || {_,_, yes} <- CFGEdges]),
    if Complex > 70 -> true; % strict view for map-like
       true ->
	    Paths = cfg_paths(FunForm, CFGEdges),
	    lists:any(fun(Path) ->
			   contains_multiple_rec_calls_base(Path, PrevFunForms)
		      end, Paths)
    end.

is_base_case_path(Path, FunForm) ->
    [FunNode] = ?Query:exec(FunForm, ?Form:func()),
    CalledFunsOnPath = called_functions(Path), %% not flat list

    lists:all(fun(FunGroup)->
                      FunClosureOnPath = fwd_closure(FunGroup, FunGroup),
                      not sets:is_element(FunNode, FunClosureOnPath)
              end, CalledFunsOnPath).

called_functions(Path) ->
    [?Query:exec(Node, ?Expr:dynfunction())
     || Node <- Path,  ?Graph:is_gnode(Node) andalso
            ?Graph:class(Node) == expr andalso
            ?Expr:type(Node) == application].

%returns a set
fwd_closure_with_dead_end(Nodes, InitNodes, DeadEndNodes) ->
    InitSet    = sets:from_list(InitNodes),
    DeadEndSet = sets:from_list(DeadEndNodes),
    closure(Nodes, InitSet, fun(N) -> ?Query:exec(N, ?Fun:funcalls()) end,
            DeadEndSet).

%returns a set
fwd_closure(Nodes, InitNodes) ->
    InitSet = sets:from_list(InitNodes),
    closure(Nodes, InitSet, fun(N) -> ?Query:exec(N, ?Fun:funcalls()) end,
            sets:new()).

%returns a set
bwd_closure(Nodes, InitNodes) ->
    InitSet = sets:from_list(InitNodes),
    closure(Nodes, InitSet, fun(N) -> ?Query:exec(N, ?Fun:called()) end,
            sets:new()).

%returns a set
closure([], WorkSet, _Fun, _DeadEndSet) -> WorkSet;
closure([Elem | Elems], WorkSet, Fun, DeadEndSet) ->
    Result  = Fun(Elem),
%    Result  = reflib_query:exec(Node, PathExpr),
    FResult =
        lists:filter(fun(E) -> not sets:is_element(E, WorkSet) andalso
                                   not sets:is_element(E, DeadEndSet) andalso
                                   not lists:member(E, Elems) end
                    ,Result),
    closure(FResult ++ Elems,
            sets:union(sets:from_list(FResult), WorkSet), Fun, DeadEndSet).


neglect_error_paths(Paths) ->
    Fun =
        fun(Path) ->
                case lists:last(Path) of
                    {error, _} -> false;
                    _          -> true
                end
        end,
    lists:filter(Fun, Paths).

get_paths(_Tab, [], Ready) ->
    [lists:reverse(RP) || RP <- Ready];
get_paths(Tab, [ [H | _T] = HPath | TPaths], Ready) ->
    case ets:lookup(Tab, H) of
        [] ->
            get_paths(Tab, TPaths, [HPath | Ready]);
        List ->
            {Returns, Targets} = assort_targets(List, [], []),

            NewPaths =
                [ [Elem | HPath] || Elem <- Targets,
                                    not lists:member(Elem, HPath)],
            NewRetPaths =
                [ [Elem | HPath] || Elem <- Returns],

            RPaths = ret_path(Tab, NewRetPaths, []),
            get_paths(Tab, NewPaths ++ TPaths, RPaths ++ Ready)
    end.

ret_path(_Tab, [], Ready) ->
    [lists:reverse(RP) || RP <- Ready];
ret_path(Tab,  [[H | _T] = HPath | TPaths], Ready) ->
    case ets:lookup(Tab, H) of
        []   ->
            ret_path(Tab, TPaths, [HPath | Ready]);
        List ->
            case assort_targets(List, [], []) of
                {[], _Targets} ->
                    R = get_paths(Tab, [HPath], []),
                    ret_path(Tab, TPaths, R ++ Ready);
                {RetTargets, _} ->
                    NewRetPaths = [ [Ret | HPath] || Ret <- RetTargets],
                    ret_path(Tab, NewRetPaths ++ TPaths, Ready)
            end
    end.

assort_targets([], RetTargets, Targets) ->
    {RetTargets, Targets};
assort_targets([{_, T, Label}|Edges], RetTargets, Targets) ->
    case lists:member(Label, ?ReturnLabels) of
        true  -> assort_targets(Edges, [T | RetTargets], Targets);
        false -> assort_targets(Edges, RetTargets, [T | Targets])
    end.
