%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.
%%%
%%% @author Melinda Toth <tothmelinda@elte.hu>
%%% @author Istvan Bozo  <bozoistvan@elte.hu>
%%% @doc <b> The module searches for candidates of farm, pool,
%%% pipeline and d&c like functions and expressions.</b>
%%%
%%% @todo doc!

-module(refpp_skeleton).
-vsn("$Rev: 11697 $ ").

-export([find/0, find/1]).

-export([get_candidates/0, get_candidates/1,
         get_candidate_chain_combinations/0,
	 get_candidate_chain_combinations/1]).

-export([ print_chains/0]).

%% for internal use
-export([create_dot/0]).

-include("referl_paraphrase.hrl").

-define(Pred, refpp_predicate_lib).

-define(SkelTypes, [farm, pipe, dnc, pool]).

%%%%%%%%%%% Interface functions %%%%%%%%%%%%%%%%%%%%%%

%% @doc The function returns the possible skeleton candiadtes in the
%% human readable format.
-spec find() -> [{candidate(), {ElemInfo, {Path, StartPos,
                                                          EndPos}}}] when
      ElemInfo :: string() | tuple(),
      Path     :: string(),
      StartPos :: tuple(),
      EndPos   :: tuple().

find()->
    Modules = ?Query:exec(?Mod:all()),
    find(Modules).

%% @doc The function returns the possible skeleton candiadtes from the
%% given modules in the human readable format. The modules can be
%% either module names (atoms) or graph node ids.
-spec find(Modules) -> [{candidate(), {ElemInfo, {Path, StartPos,
                                                                 EndPos}}}] when
      Modules  :: [atom() | gnode()],
      ElemInfo :: string() | tuple(),
      Path     :: string(),
      StartPos :: tuple(),
      EndPos   :: tuple().

find(Modules) ->
    [begin
	 case Cand of
	     #farm{type = Type, node = Node} -> farm;
	     #pipe{type = Type, nodes = [Node]} -> pipe;
             #dnc{type = Type, node = Node} -> dnc;
	     #pool{node = Node} -> Type = func, pool
	 end,
         {Cand, get_info(Type, Node)}
     end || Cand <- get_candidates([{modules,Modules}])].

%% @doc The function returns skeleton candiadte found in the semantic
%% graph of any type.
-spec get_candidates() -> [candidate()].

get_candidates() ->
    get_candidates([]).

%% @doc The function returns skeleton candiadte list according to the
%% proplist options. Proplist options are:
%% Keys: modules, cand_types
%% Associated value with modules key is a list of either module nodes or
%% module names (atoms)
%% Associated value with cand_types key is a list of values `candidate_type()'.
%% If the properties are not defined, it searches the entire graph for
%% candidates of any type.

-spec get_candidates(PropList :: search_options()) -> [candidate()].

get_candidates(PropList) ->
    Modules = proplists:get_value(modules, PropList, ?Query:exec(?Mod:all())),
    ModNormFun =
        fun(M) when is_atom(M) -> ?Query:exec(?Mod:find(M));
           (M) when is_tuple(M) -> [M];
           (_) -> []
        end,
    ModuleNodes = lists:flatmap(ModNormFun, Modules),
    
    Types = proplists:get_value(cand_types, PropList, ?SkelTypes),

    RTypes = ?MISC:intersect(Types, ?SkelTypes),

    lists:flatmap(fun(T) -> find_candidates_of_type(T, ModuleNodes) end,RTypes).

%% @doc The function returns all the possible (valid) combinations of
%% the skeleton candiadte chains. The output is a list of candidates
%% including the embeddings.

-spec get_candidate_chain_combinations() -> [candidate()].

get_candidate_chain_combinations() ->
    get_candidate_chain_combinations([]).

%% @doc The function returns all the possible (valid) combinations of
%% the skeleton candiadte chains. The output is a list of lists that
%% contains the pattern proposal, the type of the node and the node
%% for each candidate. The input is a proplist of module nodes (where
%% to search for candidates), candidate types to find and an atom
%% rewrite or no_rewrite to apply the rewrite rules or not.  If the
%% properties are undefined, it searches the entire graph for
%% candidates of any type and creates chain combinations.

-spec get_candidate_chain_combinations(PropList :: search_options()) ->
                                              [candidate()].

get_candidate_chain_combinations(PropList) ->
    CandidateList = get_candidates(PropList),
    {WorkerCandDict, ChildDict} = get_compact_info(CandidateList),
    Nodes = dict:fetch_keys(ChildDict),
    ChildList = [ Ch || N <- Nodes, Ch  <- dict:fetch(N, ChildDict)],
    F = fun(C = #farm{}, Ns) -> lists:member(C#farm.node, Ns);
           (C = #pipe{}, Ns) -> lists:member(hd(C#pipe.nodes), Ns);
           (C = #dnc{}, Ns) ->  lists:member(C#dnc.node, Ns);
	   (C = #pool{}, Ns) ->  lists:member(C#pool.node, Ns)
        end, %% exploits that pipe is represented with a single node in the dict

    InitialSet = [ C || C <- CandidateList,
                        not F(C, Nodes -- ChildList)],
    Chains = lists:usort(longest_chains(WorkerCandDict, ChildDict, InitialSet) 
			 ++ CandidateList),
    case proplists:get_value(rewrite, PropList, false) of
	true ->
	    lists:flatmap( fun transform_chain/1, Chains);
	_ ->
	    Chains
    end.

%% @doc Prints the candidate chain combinations from the entire db.
print_chains() ->
    Chains = get_candidate_chain_combinations(),
    [print(Ch) || Ch <- Chains].

%%%%%%%%%%% Helper functions %%%%%%%%%%%%%%%%%%%%%%%%

find_candidates_of_type(pool, Modules) ->
    FarmCands = find_candidates_of_type(farm, Modules),
    lists:flatmap(fun farm_to_pool/1, FarmCands);
find_candidates_of_type(dnc, Modules) ->
    DNCLikeFuns   = find_funs_by_type(Modules, dnc),
    OtherDNCFuns  = find_divide_and_conquer_funs_other(Modules),
    lists:append(DNCLikeFuns,  OtherDNCFuns);
find_candidates_of_type(pipe, Modules) ->
    PipeLikeFuns  = find_funs_by_type(Modules, pipe),
    PipeLikeExprs = find_listcomps_by_type(Modules, pipe),
    lists:append(PipeLikeFuns, PipeLikeExprs);
find_candidates_of_type(farm, Modules) ->
    FarmLikeFuns  = find_funs_by_type(Modules, farm),
    FarmLikeExprs = find_listcomps_by_type(Modules, farm),
    FarmLikeApps  = find_map_app_candidates(Modules),

    lists:append([FarmLikeFuns, FarmLikeExprs, FarmLikeApps]).

farm_to_pool(#farm{node = N, job = J, type = T, attribs = A})->
    case N of
	{'$gn', func, _} ->
	    [];
	_ ->
	    [PoolFun] = ?Query:exec(N, ?Query:seq([?Expr:clause(),
						   ?Clause:form(),
						   ?Form:func()])),
	    case refusr_metrics:is_tail_recursive({function, PoolFun}) of
		1 -> [#pool{node = PoolFun, job = {N, []}, 
			    type = T, attribs = A,
			   embeds = []}]; 
		_ -> []
	    end
    end.


print(#farm{node=N, job = {Job, Embs}}) ->
    case ?Syn:class(N) of
        func ->
            [Form] = ?Query:exec(N, ?Fun:definition()),
            {refcore_syntax:flat_text(Form), {refcore_syntax:flat_text(Form),
                                              [print(E) || E <- Embs]}};

        expr ->
            {refcore_syntax:flat_text(N), {refcore_syntax:flat_text(Job),
                                           [print(E) || E <- Embs]}}
    end;
print(#pipe{nodes=[N], stages=Stages}) ->
    {refcore_syntax:flat_text(N), [{refcore_syntax:flat_text(Job),
     [print(E) || E <- Embs]} || {Job, Embs} <- Stages]};
print(#dnc{node=N, embeds = Embs}) ->
    {refcore_syntax:flat_text(N), [print(E) || E <- Embs]};

print(_)->
    {}.
    

transform_chain(F = #farm{job={J, Embeds}}) ->
    NewEmbeds = lists:flatmap(fun transform_chain/1, Embeds),
    N = lists:usort([select_node(Sub) || Sub <- NewEmbeds]),
    DiffCh = split3(N, [[]]),
    case NewEmbeds of
	[]     -> [F];
	Embeds -> [F];
	_      -> [F#farm{job = {J, New}} || New <- DiffCh]
    end;
transform_chain(P = #pipe{})->
    [P | transform_pipe(P)];
transform_chain(_)->
    [].

gen_farm_stages(Node, Type, [{N,E} | Stages])->
    [{N, [#farm{node=Node, 
		job={N, lists:flatmap(fun transform_chain/1,E)}, 
		type=Type}]} | gen_farm_stages(Node, Type, Stages)];
gen_farm_stages(_,_,[]) ->
    [].

gen_pipe_stages([{N,E} | Stages])->
    [{N, lists:flatmap(fun transform_chain/1, E)} | gen_pipe_stages(Stages)];
gen_pipe_stages([]) ->
    [].

transform_pipe(P = #pipe{nodes = [N], stages = Stages, type=T} )->
    [#pipe{nodes=[N], 
	   stages = gen_farm_stages(N, T, Stages),
	   type = T},
%% TODO: N is not the job !!!
     #farm{node=N, 
	   job = {N, [P#pipe{stages=gen_pipe_stages(Stages)}]}, 
	   type=T}];
transform_pipe(_) ->
    [].

split3([{N, C1}, {N, C2}, {N, C3} | Tail], Acc) ->
    NewAcc = [ [C|L] || L <- Acc, C <- [C1, C2, C3]],
    split3(Tail, NewAcc);
split3([{_, C} | T], Acc)->
    split3(T, [[C|L]|| L <- Acc]);
split3([], Acc) -> [ lists:reverse(L) || L <- Acc].

find_divide_and_conquer_funs_other(Modules) ->
    Funs = find_divide_and_conquer_funs(Modules, list_comp) ++
        find_divide_and_conquer_funs(Modules, map_app),
    [create_fun_cand1(dnc, F, []) || F <- Funs].

find_divide_and_conquer_funs(Modules, Type) ->
    {Nodes, Query} =
        case Type of
            list_comp ->
                ListComps = get_listcomps(Modules),
                QuerySeq  = ?Query:seq([?Expr:clause(1), ?Clause:body()]),
                {ListComps, QuerySeq};
            map_app ->
                MapApps   = find_map_apps(Modules),
                QuerySeq  = ?Query:seq([?Expr:child(2), ?Expr:child(1)]),
                {MapApps, QuerySeq}
        end,

    NodeInfos =
        [{Node, defining_function(Node)} || Node  <- Nodes],
    
    Fun =
        fun({Node, ContFunNode}) ->
                [Expr | _] = ?Query:exec(Node, Query),
                InvFuns    = get_possibly_invoked_funs(Expr),
                BwdClosure = ?Pred:bwd_closure([ContFunNode], [ContFunNode]),
                sets:intersection(InvFuns, BwdClosure)
        end,
    sets:to_list(sets:union([ Fun(I)|| I <- NodeInfos])).

defining_function(Node) ->
    [Func] = ?Query:exec(Node, ?Query:seq([?Expr:clause(), ?Clause:funcl(),
                                      ?Clause:form(), ?Form:func()])),
%% ?Query:exec(Expr, ?Query:seq([?Expr:clause(), ?Clause:form(),
%%                                            ?Form:func()])),
    Func.

% closure of function calls, based on an expression
get_possibly_invoked_funs(Expr) ->
    {FunExpr, Implicits} = get_funs_called_by_expr(Expr),
    ImplFun =
        fun(Impl) ->
                FunNodeL   = ?Query:exec(Impl, ?Expr:dynfunction()),
                ?Pred:fwd_closure(FunNodeL, FunNodeL)
        end,
    FunExprFun =
        fun(FExpr) ->
                sets:union(
                  [ ?Pred:fwd_closure([DFun], [DFun]) ||
                      App <- ?Query:exec(FExpr, ?Expr:deep_sub()),
                      DFun <- ?Query:exec(App, ?Expr:dynfunction())])
        end,
    sets:union(lists:map(ImplFun, Implicits) ++ lists:map(FunExprFun, FunExpr)).

get_funs_called_by_expr(Expr) ->
    Subs      = ?Query:exec(Expr, ?Expr:deep_sub()),
    BackReach = refanal_dataflow:reach(Subs, [back], false),
    AssortFun =
        fun(Node, {FunExpr, Implicit}) ->
                case ?Expr:type(Node) of
                    fun_expr ->
                        {[Node | FunExpr], Implicit};
                    implicit_fun ->
                        {FunExpr, [Node | Implicit]};
                    _ ->
                        {FunExpr, Implicit}
                end
        end,
    lists:foldl(AssortFun, {[],[]}, BackReach).

find_funs_by_type(Mods, Type) ->
    Funs   = [ F || F <- ?Query:exec(Mods, ?Mod:locals()),
                    ?Query:exec(F, ?Fun:definition()) /= []],

    FunCls = [{Fun, ?Query:exec(Fun, ?Query:seq(?Fun:definition(),
                                                ?Form:clauses()))} ||
                 Fun <- Funs],
    process_funs(Type, FunCls, []).
%    [Fun || A = {Fun, [_,_]} <- FunCls, filter_cl(A, Type)].
                                    %% the function has two clauses

find_listcomps_by_type(ModNodes, farm) ->
    lists:concat([ get_listcomps_for_farm(M) || M <- ModNodes]);
find_listcomps_by_type(ModNodes, pipe) ->
    lists:concat([ get_listcomps_for_pipe(M) || M <- ModNodes]).

find_map_app_candidates(Mods) ->
    SMods   = lists:usort(Mods),
    MapApps =
        case lists:usort(?Query:exec(?Mod:all())) of
            SMods ->
                find_map_apps();
            _ ->
                find_map_apps(Mods)
        end,

    [begin
         ArgList    = ?Expr:fun_app_args(MApp),
         [Job ,_L] = ?Query:exec(ArgList, ?Expr:children()),
	 #farm{node = MApp, job = {Job, []}, type = mapapp}
     end || MApp <- MapApps].

find_map_apps() ->
    ?Query:exec(?Query:seq([?Mod:find(lists), ?Fun:find(map, 2),
                            ?Fun:applications()])).

find_map_apps(Modules) ->
    Funs = [ F || F <- ?Query:exec(Modules, ?Mod:locals()),
                  lists:any(fun(E) ->
                                    ?Fun:mfa(E) == {lists, map, 2}
                            end,
                            ?Query:exec(F, ?Fun:funcalls()))],
    Exprs = ?Query:exec(Funs,
                        ?Query:seq([?Fun:definition(), ?Form:clauses(),
                                    ?Clause:body(), ?Expr:deep_sub()])),
    
    MApps = ?Query:exec(?Query:seq([?Mod:find(lists),
                                    ?Fun:find(map, 2),
                                    ?Fun:applications()])),
    
    ?MISC:intersect(Exprs, MApps).

process_funs(_, [], Acc) ->
    Acc;
process_funs(farm, [{Fun, _} | Funs], Acc) ->
    [FunForm] = ?Query:exec(Fun, ?Fun:definition()),
    IsFarm =
        case ?Pred:is_recursive_opt(FunForm) of
            true ->
                CFGEdges = ?Pred:fun_cfg_info(FunForm),
 		Complex = length([1 || {_,_, yes} <- CFGEdges]),
 		if Complex > 70 -> false;
 		   true ->	
 			Paths = ?Pred:cfg_paths(FunForm, CFGEdges),
 			not ?Pred:is_multiple_recursive(FunForm, Paths) 
			    andalso
 			    ?Pred:is_list_processing(FunForm, Paths)
 		end;
            false ->
                false
        end,

    if IsFarm ->
            Cand = create_fun_cand1(farm, Fun, {Fun, []}),
            process_funs(farm, Funs, [Cand | Acc]);
       true   ->
            process_funs(farm, Funs, Acc)
    end;
process_funs(dnc, [{Fun, _} | Funs], Acc) ->
    [FunForm] = ?Query:exec(Fun, ?Fun:definition()),
    case ?Pred:is_divide_and_conquer(FunForm) of
        true ->
            Cand = create_fun_cand1(dnc, Fun, []),
            process_funs(dnc, Funs, [Cand | Acc]);
        false ->
            process_funs(dnc, Funs, Acc)
    end;
process_funs(pipe, _ , _Acc) ->
    [].

create_fun_cand1(pipe, Nodes, Stages) ->
    #pipe{nodes = Nodes,
	  stages = Stages,
	  type = func};
create_fun_cand1(farm, Node, Job) ->
    #farm{node = Node, 
          job  = Job, 
	  type = func}; %%{Job, Embed}}.
create_fun_cand1(dnc, Node, _) ->
    #dnc{node = Node,
         type = func,
         embeds =  []}.

check_head([E], farm) ->
    case ?Expr:type(E) of
        application -> {true, {E, []}}; %% filter appropriate apps
        variable ->
            Values = refanal_dataflow:reach_1st([E], [back]),
            %% filter appropriate apps
            A = lists:filter(fun is_application/1, Values),
            case A of
                [App] ->
                    %{true, [App]};
		    {true, {App, []}};
                _ ->
                    false
            end;
            %%lists:all(fun(V)-> ?Expr:type(V) == application end, Values);
        _ -> false
    end;
check_head([E], pipe) ->
    case ?Expr:type(E) of
        application -> %% filter appropriate apps
            case ?Query:exec(E, ?Query:seq(?Expr:child(2), ?Expr:child(1))) of
                [E1] ->
                    case ?Expr:type(E1) of
                        application ->
                            {true, extend_pipe(E1, [{E1, []}, {E, []}])};
                        _           -> false
                    end;
                _    ->  false
            end;
        _ -> false
    end;
check_head(_,_)->
    false.

extend_pipe(E, Acc) ->
    case ?Query:exec(E, ?Query:seq(?Expr:child(2), ?Expr:child(1))) of
        [E1] ->
            case ?Expr:type(E1) of
                application ->
                    extend_pipe(E1, [{E1, []} | Acc]);
                _           -> Acc
            end;
        _ -> Acc
    end.

%% %% the tail of the list constructor is a function call to the analysed
%% %% function (map(F, T), pipeline(F, G, T))
%% check_tail(E, Fun) ->
%%     case {?Expr:type(E), ?Query:exec(E, ?Expr:function())} of
%%         {application, [Fun]} -> true;
%%         {variable, _} ->
%%             Values = refanal_dataflow:reach_1st([E], [back]),
%%                                %% filter appropriate apps
%%             lists:any(fun(V)->
%%                               ?Expr:type(V) == application andalso
%%                                   ?Query:exec(V, ?Expr:function()) == [Fun]
%%                       end, Values);
%%         _ -> false
%%     end.

get_listcomps(Modules) ->
    Exprs = ?Query:exec(Modules, ?Query:seq([?Mod:locals(), ?Fun:definition(),
                                            ?Form:clauses(), ?Clause:body(),
                                            ?Expr:deep_sub()])),
    [ Expr || Expr <- Exprs, ?Expr:type(Expr) == list_comp].
    

get_listcomps_for_farm(Module) ->
    ListcompExprs = get_listcomps(Module),
    get_listcomps_for_farm(ListcompExprs, []).

% checks wether the list generator has form like [app(X) || X <- SomeList]
get_listcomps_for_farm([], Acc) ->
    Acc;
get_listcomps_for_farm([Expr | Exprs], Acc) ->
   [HClause, Compr] = ?Query:exec(Expr, ?Expr:clauses()),
    [HNode]          = ?Query:exec(HClause, ?Clause:body()),
    List = ?Query:exec(Compr, ?Clause:body()),
    Appr = is_application(HNode) andalso
        lists:any(fun(E) -> ?Expr:type(E) == list_gen end, List),
    if Appr ->
	    Cand = #farm{node = Expr, job = {HNode, []}, type = listcomp},
            get_listcomps_for_farm(Exprs, [Cand | Acc]);
       true ->
            get_listcomps_for_farm(Exprs, Acc)
    end.

get_listcomps_for_pipe(Module) ->
    ListcompExprs = get_listcomps(Module),
    get_listcomps_for_pipe(ListcompExprs, []).

get_listcomps_for_pipe([], Acc) ->
    Acc;
get_listcomps_for_pipe([Expr | Exprs], Acc) ->
    [HCl, _Compr] = ?Query:exec(Expr, ?Expr:clauses()),
    HClBdy        = ?Query:exec(HCl, ?Clause:body()),
    case {check_head(HClBdy, pipe),
          ?Query:exec(HClBdy, ?Expr:function()),
          is_application(hd(HClBdy))} of
        {{true, Stages}, _, _} ->
	    Cand = #pipe{nodes = [Expr],
			 stages = Stages,
			 type =listcomp},
            get_listcomps_for_pipe(Exprs, [Cand|Acc]);
        {false, [Func], true} ->
            RetPoints = ?Query:exec(?Fun:return_points(Func)),
            FunRetPoints =
                [ ?Query:exec(RetPoint, ?Expr:function()) ||
                    RetPoint <- RetPoints],
            IsLegal =
                lists:all(fun([]) -> false;
                             (_) -> true end, FunRetPoints),
            
            case lists:usort(lists:concat(FunRetPoints)) of
                [_] when IsLegal ->
		    Cand = #pipe{nodes = [Expr],
				 stages = [{hd(HClBdy), []},
                                           {hd(RetPoints), []}],
				 type = listcomp},
                    get_listcomps_for_pipe(Exprs, [Cand|Acc]);
                _   ->
                    get_listcomps_for_pipe(Exprs, Acc)
            end;
        {false, _, _} -> 
            get_listcomps_for_pipe(Exprs, Acc)
    end.

is_application(Expr) ->
    ?Expr:type(Expr) == application.

%% is_infix(Expr) ->
%%     ?Expr:type(Expr) == infix_expr.

longest_chains(WorkerCandDict, ChildDict, Candidates) ->
    longest_chains(WorkerCandDict, ChildDict, Candidates, []).

longest_chains(WorkerCandDict, ChildDict, Candidates, TouchedNodes) ->
    lists:flatten([longest_chain(WorkerCandDict, ChildDict, Cand, TouchedNodes)
		   || Cand <- Candidates]).

longest_chain(WorkerCandDict, ChildDict, Cand=#farm{node=Node}, TNodes) ->
    case lists:member(Node, lists:flatten(TNodes)) of
        true ->
            [];
        false ->
            Job =
                case Cand#farm.job of
                    [] ->
                        Cand#farm.node;
                    {J, _} ->
                        J
                end,
            
            case dict:fetch({Job, farm}, ChildDict) of
                [] ->
                    [Cand];
                List ->
                    NewTNodes = [Cand#farm.node | TNodes],
                    CandList = [ C || E <- List,
                                      C <- [dict:fetch(E, WorkerCandDict)],
                                      not lists:member(ret_node(C), NewTNodes)],
                    N = lists:usort([select_node(Sub) || Sub <- CandList]),
                    DiffCh = split(N, [[]]),
                    Subs = [longest_chains(WorkerCandDict, ChildDict,
                                           CList, NewTNodes) || 
                               CList <- DiffCh],
                    [Cand#farm{job={Job, Sub}} || Sub <- Subs]
            end
    end;
longest_chain(WorkerCandDict, ChildDict, Cand=#pipe{nodes=N, stages=Stages}, TNodes) ->
    case lists:member(N, TNodes) of
        true ->
            [];
        false ->
            NStages =
                [case dict:fetch({St, pipe}, ChildDict) of
                     [] ->
                         {St,[]};
                     List ->
                         NewTNodes = [N | TNodes],
                         CandList = [ C || E <- List,
                                           C <- [dict:fetch(E, WorkerCandDict)],
                                           not lists:member(ret_node(C),
                                                            NewTNodes)],
                         Subs = longest_chains(WorkerCandDict, ChildDict,
                                               CandList, NewTNodes),
                         {St, Subs}
                 end || {St, _} <- Stages],
            [Cand#pipe{stages=NStages}]
    end;
longest_chain(WorkerCandDict, ChildDict, Cand=#dnc{node=Node}, TNodes) ->
    case lists:member(Node, lists:flatten(TNodes)) of
        true ->
            [];
        false ->
            case dict:fetch({Node, dnc}, ChildDict) of
                [] ->
                    [Cand];
                List ->
                    NewTNodes = [Cand#dnc.node | TNodes],
                    CandList = [ C || E <- List,
                                      C <- [dict:fetch(E, WorkerCandDict)],
                                      not lists:member(ret_node(C), NewTNodes)],
                    N = lists:usort([select_node(Sub) || Sub <- CandList]),
                    DiffCh = split(N, [[]]),
                    Subs = lists:flatten(
                             [longest_chains(WorkerCandDict, ChildDict,
                                             CList, NewTNodes) || 
                                 CList <- DiffCh]),
                    Cand#dnc{embeds=Subs}
            end
    end;
longest_chain(WorkerCandDict, ChildDict, Cand=#pool{node=Node}, TNodes) ->
    case lists:member(Node, lists:flatten(TNodes)) of
        true ->
            [];
        false ->
            case dict:fetch({Node, dnc}, ChildDict) of
                [] ->
                    [Cand];
                List ->
                    NewTNodes = [Cand#pool.node | TNodes],
                    CandList = [ C || E <- List,
                                      C <- [dict:fetch(E, WorkerCandDict)],
                                      not lists:member(ret_node(C), NewTNodes)],
                    N = lists:usort([select_node(Sub) || Sub <- CandList]),
                    DiffCh = split(N, [[]]),
                    Subs = lists:flatten(
                             [longest_chains(WorkerCandDict, ChildDict,
                                             CList, NewTNodes) || 
                                 CList <- DiffCh]),
                    Cand#pool{embeds=Subs}
            end
    end.



split([{N, C1}, {N, C2} | Tail], Acc) ->
    NewAcc = [ [C|L] || L <- Acc, C <- [C1, C2]],
    split(Tail, NewAcc);
split([{_, C} | T], Acc)->
    split(T, [[C|L]|| L <- Acc]);
split([], Acc) -> [ lists:reverse(L) || L <- Acc].

    

select_node(F = #farm{node=N})->
    {N, F};
select_node(P = #pipe{nodes=[N]}) ->
    {N, P};
select_node(P = #pipe{}) ->
    {[],P};
select_node(D = #dnc{node=N}) ->
    {N, D};
select_node(D = #pool{node=N}) ->
    {N, D}.



%% non_empty_subsequences([]) -> [];
%% non_empty_subsequences([X|Xs]) ->
%% 	[ [X] | lists:foldr(fun(L, R) -> [L, [X | L] | R] end, [],
%%                             non_empty_subsequences(Xs))].

merge([]) ->
    dict:new();
merge([Dict|DictList]) ->
    MergeFun =
        fun(D, AccD) ->
                MFun =
                    fun(_Key, Vals1, Vals2) ->
                            % if the order matters, change it
                            lists:usort(Vals1++Vals2)
                    end,
                dict:merge(MFun, D, AccD)
        end,
    lists:foldl(MergeFun, Dict, DictList).

traversal(Dict) ->
    Nodes = dict:fetch_keys(Dict),
    DictList = [ walk({simple_walk, Node}, Dict) || Node <- Nodes],
    merge(DictList).

walk({simple_walk, W = {Node, _Type}}, Dict) ->
    case ?Syn:node_type(Node) of
        func ->
            walk_back({Node, W, _VisitedApps = []}, Dict);
        expr ->
            [_ | PathToRoot] = lists:reverse(refcore_syntax:root_path(Node)),
            case find_ancestor_candidate(PathToRoot, Dict) of
                [] ->
                    walk_back({Node, W, _VisitedApps = []}, Dict);
                [List|_] ->
                    F = fun(Prev, AccDict) ->
                                dict:append(Prev, W, AccDict)
                        end,
                    lists:foldl(F, Dict, List)
            end
    end;
walk({back_walk, AppNode, Worker, VisitedApps}, Dict) ->
    PathToRoot = lists:reverse(refcore_syntax:root_path(AppNode)),
    case find_ancestor_candidate(PathToRoot, Dict) of
        [] ->
            walk_back({AppNode, Worker, VisitedApps}, Dict);
        [List|_] ->
            F = fun(Prev, AccDict) ->
                        dict:append(Prev, Worker, AccDict)
                end,
            lists:foldl(F, Dict, List)
    end.

find_ancestor_candidate(PathToRoot, Dict) ->
    Funs = [ F || {_, N} <- PathToRoot, ?Syn:node_type(N) == form,
                  F <- ?Query:exec(N, ?Form:func())], 
    L = [ [{E,T} || T <- [farm, pipe, dnc], dict:is_key({E, T}, Dict) ]
          || {_, E} <- PathToRoot],
    FunL = [[{F, T}] || F <- Funs,T <- [farm, pipe, dnc],
                        dict:is_key({F, T}, Dict)],
    lists:dropwhile(fun([]) -> true;
                       (_) -> false
                    end, L ++ FunL).

walk_back({Node, Worker, VisitedApps}, Dict) ->
    case ?Syn:node_type(Node) of
        func ->
            walk_back_func({Node, Worker, VisitedApps}, Dict);
        expr ->
            walk_back_expr({Node, Worker, VisitedApps}, Dict)
    end.

walk_back_expr({Node, Worker, VisitedApps}, Dict) ->
    DefiningFun = defining_function(Node),

    case {dict:is_key({DefiningFun, farm}, Dict), 
          dict:is_key({DefiningFun, pipe}, Dict)} of
        {true, true} ->
            D = dict:append({DefiningFun,farm}, Worker, Dict),
            dict:append({DefiningFun, pipe}, Worker, D);
        {true, _} ->
            dict:append({DefiningFun,farm}, Worker, Dict);
        {_, true} ->
            dict:append({DefiningFun, pipe}, Worker, Dict);
        _ ->
            walk_back_func({DefiningFun, Worker, VisitedApps}, Dict)
    end.

walk_back_func({Node, Worker, VisitedApps}, Dict) ->
    case ?Query:exec(Node, ?Query:all([?Fun:applications(), 
				       ?Fun:implicits()])) of
        [] ->
            Dict;
        Apps ->
            DictList =
                [walk({back_walk, App, Worker, [App|VisitedApps]}, Dict) ||
                    App <- Apps,
                    not lists:member(App, VisitedApps)],
            merge([Dict|DictList])
    end.

get_info(Type, Node) ->
    [F1, L1, L2] =
        case Type of
            func ->
                refusr_sq_format:file_and_tokens(function, Node);
            _    ->
                refusr_sq_format:file_and_tokens(expression, Node)
        end,
    EntityInfo   = get_entity_info(Type, Node),
    {EntityInfo, {?File:path(F1), element(1,reflib_token_gen:linecol(L1)),
                  element(2,reflib_token_gen:linecol(L2))}}.

get_entity_info(func, Fun) ->
    {_, MFA} = ?Fun:mod_fun_arity(Fun),
    MFA;
get_entity_info(_, Expr) ->
    refcore_syntax:flat_text(Expr).

%%%% Comapact view for internal use

%%% @private
get_compact_info(CandList) ->
    Fun =
        fun(Cand, {AccIn1, AccIn2}) ->
            ListOfPairs =
                case Cand of
                    #farm{job = {Job, _}} ->
                        [{Job, farm, Cand}];
                    #pipe{stages = Stages} ->
                        [{Stage, pipe, Cand} || {Stage, _} <- Stages];
                    #dnc{node = N} ->
                        [{N, dnc, Cand}];
		    #pool{node = N} ->
			[{N, dnc, Cand}]
                end,
                F = fun({N, T, C}, {A1, A2}) ->
                        {dict:store({N, T}, [], A1), dict:store({N, T}, C, A2)}
                    end,
                lists:foldl(F, {AccIn1, AccIn2}, ListOfPairs)
        end,
    {InitChildDict, WorkerCandDict} =
        lists:foldl(Fun, {dict:new(), dict:new()}, CandList),

    ChildDict = traversal(InitChildDict),
    UpdateFun = 
	fun(K, Dict) ->
		Elems = dict:fetch(K, Dict),
		Nodes = [{E, ret_node(dict:fetch(E, WorkerCandDict))} || 
			    E <- Elems],
		NewList = update(ret_node(dict:fetch(K, WorkerCandDict)),Nodes),
		dict:store(K, NewList, Dict)
	end,
    Keys = dict:fetch_keys(ChildDict),
    ChildDictNew = lists:foldl(UpdateFun, ChildDict, Keys),
    {WorkerCandDict, ChildDictNew}.

update(KCand, [ {_E, KCand} | Tail ]) ->
    update(KCand, Tail);
update(KCand, [ {E, _} | Tail]) ->
    [E | update(KCand, Tail)];
update(_, []) ->
    [].


ret_node(#farm{node = N})-> N;
ret_node(#pipe{nodes = N})-> hd(N);
ret_node(#dnc{node = N}) -> N;
ret_node(#pool{node = N}) -> N. 
    

%%% @private
%% creates a chain.dot file in pwd from the actual content of the database
create_dot() ->
    CandidateList = get_candidates(),

    {PropDict, ChildDict} = get_compact_info(CandidateList),

    FileId =
        case file:open("chain.dot", [write]) of
            {ok, FI} -> FI;
            {error, Reason1} ->
                throw(Reason1)
        end,
    ChildList = dict:to_list(ChildDict),
    PropList  = dict:to_list(PropDict),
    %io:format("PropList:~p~n", [PropList]),
    %io:format("ChildList:~p~n", [ChildList]),
    io:fwrite(FileId, "digraph chains{~n",[]),
    io:fwrite(FileId, "node [shape=record];~n",[]),
    [ io:fwrite(FileId, "struct~p [label=\"<f0> ~p |<f1> ~p\"];~n",
                [NodeId, NodeId, PList])
      || {{_,_,NodeId}, PList} <- PropList],

    [ io:fwrite(FileId, "struct~p -> struct~p;~n",
                [NodeId, ChId])
      || {{_,_,NodeId}, ChList} <- ChildList, {_,_,ChId} <- ChList],

    io:fwrite(FileId, "}",[]).
