%%% -*- coding: utf-8 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Eötvös Loránd University.
%%% Portions created  by Eötvös Loránd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Eötvös Loránd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.
%%%
%%% @author Judit Koszegi <koszegijudit@elte.hu>

-module(refpp_timestamp).

-include("referl_paraphrase.hrl").

-export([measure/3, ts_receiver/4]).

measure(Mode, Node, Nodes) ->
    NodesWithEmptyTimeList = lists:map(fun(N) -> {N, []} end, Nodes),
    register(ts_receiver, 
             spawn(refpp_timestamp, ts_receiver, 
                   [{Mode, NodesWithEmptyTimeList}, start, start, self()])),
    ts_receiver ! start,
    refpp_benchmark:time(Node, Nodes),
    ts_receiver ! stop,
    receive
        NodesWithTimeList -> 
            lists:map(fun({N, TimeList}) -> {N, mean(TimeList)} end, NodesWithTimeList)
    end.

%% In case of pipes we measure tseq for each stages.
ts_receiver({pipe, Nodes}, First_tag, Prev, Pid) ->
     receive
         stop -> 
             Pid ! Nodes;
         start ->
             ts_receiver({pipe, Nodes}, start, start, Pid);
         e ->
             Ts = erlang:now(),
             ts_receiver({pipe, Nodes}, First_tag, {e, Ts}, Pid);
         s ->
             Ts = erlang:now(),
             ts_receiver({pipe, Nodes}, First_tag, {s, Ts}, Pid);
         First_tag ->
             Ts = erlang:now(),
             ts_receiver({pipe, Nodes}, First_tag, {First_tag, Ts}, Pid);
         Tag ->
             Ts = erlang:now(),
             case Prev of
                 start -> 
                     ts_receiver({pipe, Nodes}, Tag, {Tag, Ts}, Pid);
                 {s, _} -> 
                     ts_receiver({pipe, Nodes}, Tag, {Tag, Ts}, Pid);
                 _ ->
                     New_Nodes = process_ts_pipe(Prev, {Tag, Ts}, Nodes),
                     ts_receiver({pipe, New_Nodes}, First_tag, {Tag, Ts}, Pid)
              end
     end;
%% In case of embedded skeletons we measure the "before" tseq for each embedded node.
%% To be able to store the tseq after the last embedded node,
%% we intorduce a special "end node": {'$gn', 0, 0}.
ts_receiver({embedded, Nodes}, First_tag, Prev, Pid) ->
    receive
        stop -> 
            Pid ! Nodes;
        start ->
            Ts = erlang:now(),
            ts_receiver({embedded, [{{'$gn', 0, 0},[]} | Nodes]},  
                        First_tag, {{e, start}, Ts}, Pid);
        s ->
            Ts = erlang:now(),
            ts_receiver({embedded, Nodes}, First_tag, {{e, start}, Ts}, Pid);
        Tag ->
            Tag2 =
                case Tag of
                    e -> {s,{0,0}};
                    _ -> Tag
                end,
            Ts = erlang:now(),
            NewNodes = process_ts_embedded(Prev, {Tag2, Ts}, Nodes),
            ts_receiver({embedded, NewNodes}, First_tag, {Tag2, Ts}, Pid)
    end;
%% In case of maplike functions we measure the whole recursive function,
%% then we divide the time by the number of recursive calls.
ts_receiver({maplike, Nodes}, First_tag, Prev, Pid) ->
    ts_receiver({maplike, Nodes, 0}, First_tag, Prev, Pid);
ts_receiver({maplike, Nodes, Count}, First_tag, Prev, Pid) ->
    receive
        stop -> 
            Pid ! Nodes;
        s ->
            Ts = erlang:now(),
            ts_receiver({maplike, Nodes, Count}, First_tag, {s,Ts}, Pid);
        {s, _} ->
            ts_receiver({maplike, Nodes, Count+1}, First_tag, Prev, Pid);
        e ->
            Ts = erlang:now(),
            NewNodes = process_ts_maplike(Prev, {e, Ts}, Nodes, Count),
            ts_receiver({maplike, NewNodes, 0}, First_tag, Prev, Pid);
        _ -> % start / {e, _}
            ts_receiver({maplike, Nodes, Count}, First_tag, Prev, Pid)
    end.

process_ts_pipe({{Mode1, Node1}, Ts1}, {{Mode2, Node2}, Ts2}, Nodes) ->
    Seq_num1 = seq_number(Nodes, Node1), 
    Seq_num2 = seq_number(Nodes, Node2),
    Diff = timer:now_diff(Ts2, Ts1),
    if
        (Seq_num1 =< Seq_num2) andalso (not (Mode1 == e andalso Mode2 == s)) ->
            Node =
                case Mode2 of
                    s -> Node1;
                    e -> Node2
                end,
            put_new_time(Nodes, Node, Diff);
        true ->
            Nodes
    end.

process_ts_embedded({{Mode1, _Node1}, Ts1}, {{Mode2, Node2}, Ts2}, Nodes) ->
    Diff = timer:now_diff(Ts2, Ts1),
    if
        Mode2 == s -> put_new_time(Nodes, Node2, Diff);
        true -> Nodes
    end.

process_ts_maplike({_, Ts1}, {_, Ts2}, Nodes, Count) ->
    case Count of
        0 -> Nodes;
        _ ->
            Diff = timer:now_diff(Ts2, Ts1),
            {{'$gn', X, Y}, _} = hd(Nodes),
            put_new_time(Nodes, {X,Y}, Diff/Count)
    end.

%% The sequence number of the stage within the pipe.
seq_number(Nodes, NodeIDPair) ->
    seq_number0(Nodes, NodeIDPair, 1).

seq_number0([], _NodeIDPair, _Num) -> 
    0;
seq_number0([{{'$gn', X, Y},_} | _Nodes], {X, Y}, Num) -> 
    Num; 
seq_number0([_Node | Nodes], NodeIDPair, Num) ->
    seq_number0(Nodes, NodeIDPair, Num+1).

%% Insert new tseq for a given Node.
put_new_time([], _Node, _Time) -> 
    [];
put_new_time([{{'$gn', X, Y} = Node, TimeList} | Nodes], {X, Y}, Time) ->  
    [{Node, [Time | TimeList]}  | Nodes]; 
put_new_time([ N | Nodes], Node, Time) ->
    [N | put_new_time(Nodes, Node, Time)].

mean(List) ->
    if 
        length(List) == 0 -> undefined;
        true -> lists:sum(List)/length(List)
    end.

   
