%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Eötvös Loránd University.
%%% Portions created  by Eötvös Loránd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Eötvös Loránd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.
%%%
%%% @author Viktoria Fordos <f-viktoria@elte.hu>
%%% @doc
%%% <b>Webserver's controller module</b>
%%%
%%% @todo doc!

-module(refpp_web_srv_controller).
-export([start/0, stop/0, info/0, server_info/0]).

-export([error_text/2]).

-include("referl_paraphrase.hrl").

-define(PARA_WEB_SRV_NAME,"paraphrase-result-browser").
-define(PARA_WEB_SRV_ALIAS, "/paraphrase").
-define(PARA_WEB_SRV_MOD, ?WEB_SERVICES).

%%% ============================================================================
%%% Interface

start() ->
    case info() of
        not_started ->
            do_init(),
            start0();
        Else -> Else
    end.

stop()->
    case inets:stop() of
        ok -> ok;
        {error,{not_started,inets}} -> ok;
        Else -> Else
    end.

info()->
    info0(server_info()).

server_info()->
    case  is_inets_started() andalso proplists:get_value(httpd, inets:services()) of
        false -> not_started;
        undefined -> not_started;
        Pid when is_pid(Pid)-> get_para_srv_datas(Pid)
    end.

%%% ============================================================================
%%% Error handling
error_text(inets_startup_failure, Else)->
    io_lib:format("The webserver could not be started. Reason: ~p~n",[Else]);
error_text(server_dir, [Reason, ServerDir])->
    io_lib:format(string:join(["The webserver could not be started, ",
                  "because the creation of the required directory (",
                   ServerDir,
                  ") failed. ",
                  "Reason of the failure: ~p~n"],""),[Reason]).

%%% ============================================================================
%%% Internals

info0(CoreInfo) when is_list(CoreInfo)->
    CoreInfo ++ [{services, ?WEB_SERVICES:services()}];
info0(Else)->
    Else.

do_init()->
    case application:start(inets) of
        ok -> ok;
        {error,{already_started,inets}} -> ok;
        Else -> throw(?LocalError(inets_startup_failure, Else))
    end,
    info().

is_inets_started()->
    lists:any(fun({inets,_,_})->
                true;
                (_)->false
            end, application:which_applications()).
start0()->
    {ok, Pid}=
    inets:start(httpd, [
                {modules, [
                    mod_alias,
                    mod_auth,
                    mod_esi,
                    mod_actions,
                    mod_cgi,
                    mod_dir,
                    mod_get,
                    mod_head,
                    mod_log,
                    mod_disk_log]},
                {port,0},
                {server_name,?PARA_WEB_SRV_NAME},
                {server_root, ensure_server_dirs()},
                {document_root, ensure_server_dirs()},
                {erl_script_alias,
                    {?PARA_WEB_SRV_ALIAS,
                     [?PARA_WEB_SRV_MOD]}},
                {error_log, "error.log"},
                {security_log, "security.log"},
                {transfer_log, "transfer.log"},
                {mime_types,[
                    {"html","text/html"},
                    {"css","text/css"},
                    {"js","application/x-javascript"},
                    {"json","application/json"},
                    {"csv", "text/csv"},
                    {"erl", "text/plain"}]}]),
    info0(get_para_srv_datas(httpd:info(Pid))).

get_para_srv_datas(Pid) when is_pid(Pid)->
    get_para_srv_datas(httpd:info(Pid));
get_para_srv_datas(Config) when is_list(Config)->
    case proplists:get_value(server_name, Config) of
        ?PARA_WEB_SRV_NAME ->
            [{port, proplists:get_value(port, Config)},
            {url, base_url(proplists:get_value(port, Config))},
            {server_dir, server_dir()}];
        _ ->[]
    end.


base_url(Port)->
    string:join(
        io_lib:format("~s~p~s~s~p",
            ["http://localhost:",
            Port,
            ?PARA_WEB_SRV_ALIAS,
            "/",
            ?PARA_WEB_SRV_MOD]),"").

server_dir()->
    filename:join([?MISC:data_dir(),"paraphrase_ui"]).

ensure_server_dirs()->
    ServerDir = server_dir(),
    case file:make_dir(ServerDir) of
        ok ->
            ServerDir;
        {error,eexist} ->
            ServerDir;
        {error, Reason} ->
            throw(?LocalError(server_dir, [Reason, ServerDir]))
    end.

