%%%-------------------------------------------------------------------
%%% @author Gabor Olah <olikas@elte.hu>
%%% @copyright (C) 2014, ELTE
%%% @doc
%%% Calculates valuetypes for an Erlang source file.
%%% @end
%%% Created : 17 Mar 2014 by Gabor Olah <olikas@elte.hu>
%%%-------------------------------------------------------------------
-module(valuetypes).

-include("core_parse.hrl").

-export([
         start/1,
         print/1
        ]).

-define(FUN_STORE, function_store_table).
-define(RECURSION_LIMIT, 15).
-define(complete, complete).
-define(incomplete, incomplete).

%-define(d(F, As), io:format("~w - ~w: ~p~n", [?LINE, F, As])).
-define(d(F, As), ok).

-record(f,
        {name,                %% {name, arity}
         type,                %% erl_type()  
         state = ?incomplete, %% complete | incomplete
         body}).              %% cerl_tree()
-record(state,
        {env,                 %% type environment = dict()  
         fun_name,
         callgraph,           %% digraph()
         complete}).         


start(FileOrForms) ->
    {ok, _, Tree} = compile_to_core(FileOrForms),
    ets:new(?FUN_STORE, [named_table, set, {keypos, 2}]),
    try
        analyze_module(Tree)
    after
        ets:delete(?FUN_STORE)
    end.


analyze_module(Tree) ->
    ModuleName = cerl:module_name(Tree),
    %% Exports = cerl:module_exports(Tree),
    Defs = cerl:module_defs(Tree),
    Funs = [ #f{name  = cerl:var_name(Var), 
                type  = referl_erl_types:t_fun(),
                state = ?complete,
                body  = FunBody} || {Var, FunBody} <- Defs ],
    ets:insert(?FUN_STORE, Funs),
    loop(?RECURSION_LIMIT).


compile_to_core({forms, Forms}) ->
    compile:forms(Forms, [to_core, binary, no_line_info]);
compile_to_core(FileName) ->
    compile:file(FileName, [to_core, binary, no_line_info]).

loop(?RECURSION_LIMIT = N) ->
    ?d(loop, {first_run}),
    Funs = ets:tab2list(?FUN_STORE),
    FunNames = [ F#f.name || F <- Funs],
    Graph = digraph:new(),
    [ begin 
          digraph:add_vertex(Graph, FN),
          io:format("Added Vertex: ~p~n", [FN])
      end || FN <- FunNames ],
    
    [ begin 
          {RetType, RetState} = visit(F#f.body, 
                                      #state{env = dict:new(),
                                             fun_name = F#f.name,
                                             callgraph = Graph}),
          ets:insert(?FUN_STORE, 
                 F#f{type = RetType,
                     state = RetState#state.complete})
      end || F <- Funs ],
    case digraph_utils:topsort(Graph) of
        false ->
            loop(FunNames, N-1);
        Sorted ->
            loop(Sorted, N - 1)
    end.

loop(Funs, 0) ->
    ?d(loop, {last_run}),
    [ begin 
          [FF] = ets:lookup(?FUN_STORE, F),
          {FF#f.name, FF#f.type}
      end || F <- Funs ];

loop(Funs, N) when is_integer(N), N > 0 ->
    ?d(loop, {run, N}),
    [ begin
          [FF] = ets:lookup(?FUN_STORE, F),
          {RetType, RetState} = visit(FF#f.body, #state{env=dict:new()}),
          ets:insert(?FUN_STORE, 
                 FF#f{type = RetType,
                      state = RetState#state.complete})
      end || F <- Funs ],
    loop(Funs, N-1).


visit(T, St) ->
    case cerl:type(T) of
        literal ->
            %?d(visit, literal),
            {referl_erl_types:t_from_term(cerl:concrete(T)), St};
        var ->
            Name = cerl:var_name(T),
            case dict:find(Name, St#state.env) of
                {ok, X} ->
                    %?d(visit, {var, Name, X}),
                    {X, St};
                error ->
                    %?d(visit, {var, Name, error_case, none}),
                    %% If the var is not in the env already, then
                    %% something went wrong or there is a mistake int
                    %% he program. Initialize it to none.
                    Env1 = dict:store(Name, referl_erl_types:t_none(), St#state.env),
                    St1 = St#state{env = Env1},
                    {referl_erl_types:t_none(), St1}
            end;
        'fun' ->
            Vars = cerl:fun_vars(T),
            Env1 = lists:foldl(fun(V, StAcc) ->
                                       dict:store(cerl:var_name(V), 
                                                  referl_erl_types:t_any(),
                                                  StAcc)
                               end, St#state.env, Vars),
            Body = cerl:fun_body(T),
            {BodyType, St2} = visit(Body, St#state{env = Env1}),
            ArgTypes = [ dict:fetch(cerl:var_name(V), St2#state.env) || V <- Vars],
            %?d(visit, {'fun', ArgTypes, BodyType}),
            {referl_erl_types:t_fun(ArgTypes, BodyType), St2};
        values ->
            {Xs, St1} = visit_list(cerl:values_es(T), St),
            {referl_erl_types:t_product(Xs), St1};
        cons ->
            {[X1, X2], St1} = visit_list([cerl:cons_hd(T), cerl:cons_tl(T)], St),
            {referl_erl_types:t_cons(X1, X2), St1};
        tuple ->
            {Xs, St1} = visit_list(cerl:tuple_es(T), St),
            {referl_erl_types:t_tuple(Xs), St1};
        'let' ->
            {X, St1} = visit(cerl:let_arg(T), St),
            LetVars = cerl:let_vars(T),
            St1Vars = St1#state.env,
            Env = case referl_erl_types:t_is_any(X) orelse referl_erl_types:t_is_none(X) of
                      true ->
                          bind_vars_single(LetVars, X, St1Vars);
                      false ->
                          bind_vars(LetVars, referl_erl_types:t_to_tlist(X), St1Vars)
                  end,
            visit(cerl:let_body(T), St1#state{env = Env});
        seq ->
            {_, St1} = visit(cerl:seq_arg(T), St),
            visit(cerl:seq_body(T), St1);
        apply ->
            %% Always an in-module call, the type can be calculated
            Op = cerl:apply_op(T),
            case St#state.callgraph of
                undefined -> ok;
                Graph ->
                    Fn = St#state.fun_name,
                    %io:format("Apply: add edge: ~p -> ~p~n", [Fn,  cerl:var_name(Op) ]),
                    digraph:add_edge(Graph, Fn, cerl:var_name(Op))
            end,
            Args = cerl:apply_args(T),
            {_F, St1} = visit(Op, St),
            {As, St2} = visit_list(Args, St1),
            {FunType, FComplete} = get_local_type(cerl:var_name(Op)),
            DomainTypes = case referl_erl_types:t_fun_args(FunType) of
                              unknown ->
                                  lists:duplicate(length(Args), referl_erl_types:t_any());
                              Domain ->
                                  Domain
                          end,
            %?d(visit, {apply, funtype, FunType}),
            Env1 = lists:foldl(
                     fun({Arg, ArgT}, EnvAcc) ->
                             %?d(visit, {apply, fold, arg, Arg, argt, ArgT}),
                             bind_pat_vars(Arg, ArgT, EnvAcc, fun meet/2)
                     end, 
                     St2#state.env,
                     lists:zip(Args, DomainTypes)),
            %?d(visit, {env1, Env1}),
            Complete = case FComplete of
                           ?incomplete ->
                               ?incomplete;
                           Else ->
                               FComplete
                       end,
            {referl_erl_types:t_fun_range(FunType), St2#state{env = Env1,
                                                       complete = Complete}};
        call ->
            %% It can be a bif of a an external call. If BIF, then the 
            %% type comes from referl_erl_bif_types:type(Module, Function, Arity, Arguments)
            M = cerl:call_module(T),
            F = cerl:call_name(T),
            As = cerl:call_args(T),
            {[X1, X2], St1} = visit_list([M, F], St),
            {Xs, St2} = visit_list(As, St1),
            case {referl_erl_types:t_atom_vals(X1), referl_erl_types:t_atom_vals(X2)} of
                {[M1], [F1]} ->
                    A = length(As),
                    %% TODO inner module call (mymodule:myfunction)
                    ArgTypes = 
                        case arg_type(M1, F1, A) of
                            unknown ->
                                lists:duplicate(length(As), referl_erl_types:t_any());
                            Domain ->
                                Domain
                        end,
                    Env1 = 
                        lists:foldl(
                          fun({Arg, ArgT}, EnvAcc) ->
                                  bind_pat_vars(Arg, ArgT, EnvAcc, fun meet/2)
                          end, 
                          St2#state.env,
                          lists:zip(As, ArgTypes)),
                    {call_type(M1, F1, A, Xs), St2#state{env = Env1}};
                _ ->
                    {referl_erl_types:t_any(), St2}
            end;
        primop ->
            As = cerl:primop_args(T),
            {Xs, St1} = visit_list(As, St),
            F = cerl:atom_val(cerl:primop_name(T)),
            A = length(As),
            X = primop_type(F, A, Xs),
            {X, St1};
        'case' ->
            Arg = cerl:case_arg(T),
            Clauses = [ C || C <- cerl:case_clauses(T), 
                             C#c_clause.anno =/= [compiler_generated]],
            {X, St1} = visit(Arg, St),
            Xs = case referl_erl_types:t_is_any(X) orelse referl_erl_types:t_is_none(X) of
                     true ->
                         %?d(visit, {'case', arg, true, X}),
                         [X || _ <- Clauses];
                     false ->
                         %?d(visit, {'case', arg, false, X}),
                         referl_erl_types:t_to_tlist(X)
                 end,
            {PatType, {RetType, St2}} = 
                join_visit_clauses(Xs, Clauses, St1, referl_erl_types:t_none()),
            %?d(visit, {'case', patType, PatType, xs, Xs, retType, RetType}),
            Env2 = bind_pat_vars(Arg, PatType, St2#state.env, fun meet/2),
            {RetType, St2#state{env = Env2}};
        'receive' ->
            Any = referl_erl_types:t_any(),
            {_PatType, {X1, St1}} = join_visit_clauses([Any], cerl:receive_clauses(T),
                                           St, referl_erl_types:t_none()),
            {X2, St2} = visit(cerl:receive_timeout(T), St1),
            case referl_erl_types:t_is_atom(X2) andalso 
                (referl_erl_types:t_atom_vals(X2) =:= [infinity]) of
                true ->
                    {X1, St2};
                false ->
                    {X3, St3} = visit(cerl:receive_action(T), St2),
                    {join(X1, X3), St3}
            end;
        'try' ->
            {X, St1} = visit(cerl:try_arg(T), St),
            Any = referl_erl_types:t_any(),
            Atom = referl_erl_types:t_atom(),
            TryVars = cerl:try_vars(T),
            St1Env = St1#state.env,
            Env = case referl_erl_types:t_is_any(X) orelse referl_erl_types:t_is_none(X) of
                       true ->
                           bind_vars_single(TryVars, X, St1Env);
                       false ->
                           bind_vars(TryVars, referl_erl_types:t_to_tlist(X), St1Env)
                   end,
            {X1, St2} = visit(cerl:try_body(T), St1#state{env = Env}),
            EVars = bind_vars(cerl:try_evars(T), [Atom, Any, Any], St2#state.env),
            {X2, St3} = visit(cerl:try_handler(T), St2#state{env = EVars}),
            {join(X1, X2), St3};
        'catch' ->
            {_, St1} = visit(cerl:catch_body(T), St),
            {referl_erl_types:t_any(), St1};
        binary ->
            {_, St1} = visit_list(cerl:binary_segments(T), St),
            {referl_erl_types:t_binary(), St1};
        bitstr ->
            %% FIXME it do not understand HiPe, bitstring should be handled correctly.
            {_, St1} = visit(cerl:bitstr_val(T), St),
            {_, St2} = visit(cerl:bitstr_size(T), St1),
            {referl_erl_types:t_none(), St2};
        alias ->
            visit(cerl:alias_pat(T), St);
        letrec ->
            %% TODO
            %% In my understanding it should not occur in the code,
            %% since it could turn up as a top level structure in a
            %% module. There is no way to write recursive fun, only in
            %% a module.
            io:format("letrec: ~p~n", [T]),
            error_letrec;
        module ->
            %% should not happen, since we deal with modules in the upper level function.
            io:format("module: ~p~n", [T]),
            error_module
            
    end.

%% We assume correct value-list typing.

visit_list([T | Ts], St) ->
    ?d(visit_list, {t, T, st, St}),
    {X, St1} = visit(T, St),
    {Xs, St2} = visit_list(Ts, St1),
    {[X | Xs], St2};
visit_list([], St) ->
    {[], St}.


join_visit_clauses(Xs, [T | Ts], St, PatType) ->
    %?d(join_visit_clauses, {xs, Xs, patType, PatType}),
    {PT, {X1, St1}} = visit_clause(T, Xs, St),
    %?d(join_visit_clauses, {pt, PT, x1, X1, st1, St1, patType, PatType}),
    %?d(jvc, {pt, PT}),
    {NewPT, {X2, St2}} = join_visit_clauses(Xs, Ts, St1, join(PT, PatType)),
    ClauseType = join(X1, X2),
    %?d(join_visit_clauses, {x1, X1, x2, X2, clauseType, ClauseType}),
    {NewPT, {ClauseType, St2}};
join_visit_clauses(_, [], St, PatType) ->
    {PatType, {referl_erl_types:t_none(), St}}.

visit_clause(T, Xs, St) ->
    Patterns = cerl:clause_pats(T),
    Env1 = bind_pats(Patterns, Xs, St#state.env, fun meet/2),
    %?d(visit_clause, {xs, Xs, env1, Env1, patterns, Patterns}),
    {PatTypes, St1} = visit_list(Patterns, St#state{env=Env1}),
    Env2 = St1#state.env,
    %?d(visit_clause, {patTypes, PatTypes, env2, Env1}),
    % bind_pats([Arg], PatTypes, St1#state.env, fun meet/2),
    G = cerl:clause_guard(T),
    {_, St2} = visit(G, St1#state{env = Env2}),
    Env3 = guard_filters(G, Env2),
    {BodyType, St3} = B = visit(cerl:clause_body(T), St2#state{env=Env3}),
    ?d(visit_clause, {b,B, t, T}),
    {UpdatedPatTypes, _} = visit_list(Patterns, St3),
    {referl_erl_types:t_product(UpdatedPatTypes), 
     B}.

bind_pats(Ps, Xs, Vars, Way) ->
    if length(Xs) =:= length(Ps) ->
            %?d(bind_pats, {list}),
            bind_pats_list(Ps, Xs, Vars, Way);
       true ->
            %?d(bind_pats, {singe, ps, Ps, xs, Xs}),
            %% FIXME
            %% A corner case when somehow we do not have a the same
            %% number of patterns as types in Xs. The why is a
            %% mistery. It should be investigated. By that time, I
            %% changed the correspondng type to any() from
            %% none(). This makes some cases work.
            bind_pats_single(Ps, referl_erl_types:t_any(), Vars, Way)
    end.

bind_pats_list([P | Ps], [X | Xs], Vars, Way) ->
    Vars1 = bind_pat_vars(P, X, Vars, Way),
    bind_pats_list(Ps, Xs, Vars1, Way);
bind_pats_list([], [], Vars, _Way) ->
    Vars.

bind_pats_single([P | Ps], X, Vars, Way) ->
    bind_pats_single(Ps, X, bind_pat_vars(P, X, Vars, Way), Way);
bind_pats_single([], _X, Vars, _Way) ->
    Vars.

bind_pat_vars(P, X, Vars, Way) ->
    case cerl:type(P) of
        var ->
            Name = cerl:var_name(P),
            NewType = case dict:find(Name, Vars) of
                          {ok, X1} ->
                              %?d(bind_pat_vars, {vars, found, Name, way, Way, vars, Vars}),
                              Way(X, X1);
                          error ->
                              %?d(bind_pat_vars, {vars, not_found, Name, x, X, vars, Vars}),
                              %?d(bind_pat_vars, {Name, X}), 
                              X %referl_erl_types:t_any()
                      end,
            dict:store(Name, NewType, Vars);
        literal ->
            Vars;
        call ->
            Vars;
        apply ->
            Vars;
        cons ->
            case referl_erl_types:t_is_cons(X) of
                true ->
                    %% If X is "nonempty proper list of X1", then the
                    %% head has type X1 and the tail has type "proper
                    %% list of X1". (If X is just "cons cell of X1",
                    %% then both head and tail have type X1.)
                    Vars1 = bind_pat_vars(cerl:cons_hd(P), referl_erl_types:t_cons_hd(X),
                                          Vars, Way),
                    bind_pat_vars(cerl:cons_tl(P), referl_erl_types:t_cons_tl(X), Vars1, Way);
                false ->
                    case referl_erl_types:t_is_list(X) of
                        true ->
                            %% If X is "proper list of X1", then the
                            %% head has type X1 and the tail has type
                            %% "proper list of X1", i.e., type X.
                            Vars1 = bind_pat_vars(cerl:cons_hd(P),
                                                  referl_erl_types:t_list_elements(X),
                                                  Vars, Way),
                            bind_pat_vars(cerl:cons_tl(P), X, Vars1, Way);
                        false ->
                            case referl_erl_types:t_is_maybe_improper_list(X) of
                                true ->
                                    %% If X is "cons cell of X1", both
                                    %% the head and tail have type X1.
                                    X1 = referl_erl_types:t_list_elements(X),
                                    Vars1 = bind_pat_vars(cerl:cons_hd(P),
                                                          X1, Vars, Way),
                                    bind_pat_vars(cerl:cons_tl(P), X1,
                                                  Vars1, Way);
                                false ->
                                    bind_vars_single(cerl:pat_vars(P),
                                                     top_or_bottom(X),
                                                     Vars)
                            end
                    end
            end;
        tuple ->
            %?d(tuple, {x, X, p, P}),
            case referl_erl_types:t_is_tuple(X) of
                true ->
                    %?d(tuple, {x, X, p, P}),
                    case referl_erl_types:t_tuple_subtypes(X) of
                        unknown -> 
                            bind_vars_single(cerl:pat_vars(P), top_or_bottom(X), 
                                             Vars);
                        [Tuple] ->
                            case referl_erl_types:t_tuple_size(Tuple) =:= cerl:tuple_arity(P) of
                                true ->
                                    %?d(tuple, {all_true, x, X, p, P}),
                                    bind_pats_list(cerl:tuple_es(P), 
                                                   referl_erl_types:t_tuple_args(Tuple), 
                                                   Vars, Way);
                                
                                false ->
                                    bind_vars_single(cerl:pat_vars(P), 
                                                     top_or_bottom(X), Vars)
                            end;
                        List when is_list(List) ->
                            bind_vars_single(cerl:pat_vars(P), top_or_bottom(X), 
                                             Vars)
                    end;
                false ->
                    bind_vars_single(cerl:pat_vars(P), top_or_bottom(X), Vars)
            end;
        binary ->
            bind_pats_single(cerl:binary_segments(P), referl_erl_types:t_none(), Vars, Way);
        bitstr ->
            %% Only the Value field is a new binding. Size is already
            %% bound, and the other fields are constant literals.
            %% We could create a filter for Size being an integer().
            Size = cerl:bitstr_size(P),
            ValType = 
                case cerl:concrete(cerl:bitstr_type(P)) of
                    float -> referl_erl_types:t_float();
                    binary -> referl_erl_types:t_binary();
                    integer ->
                        case cerl:is_c_int(Size) of
                            false -> referl_erl_types:t_integer();
                            true -> 
                                SizeVal = cerl:int_val(Size),
                                Flags = cerl:concrete(cerl:bitstr_flags(P)),
                                case lists:member(signed, Flags) of
                                    true -> 
                                        referl_erl_types:t_from_range(-(1 bsl (SizeVal - 1)),
                                                     1 bsl (SizeVal - 1) - 1);
                                    false -> 
                                        referl_erl_types:t_from_range(0,1 bsl SizeVal - 1)
                                end
                        end;
                    _Unicode ->
                        %% I do not know how to deal with unicode bitstrs. :(
                        referl_erl_types:t_any()
                end,
            bind_pat_vars(cerl:bitstr_val(P), ValType, Vars, Way);
        alias ->
            P1 = cerl:alias_pat(P),
            Vars1 = bind_pat_vars(P1, X, Vars, Way),
            dict:store(cerl:var_name(cerl:alias_var(P)), pat_type(P1, Vars1),
                       Vars1);
        values ->
            case referl_erl_types:t_is_any(X) orelse referl_erl_types:t_is_none(X) of
                true ->
                    case cerl:values_es(P) of
                        [T | Ts] ->
                            Env1 = bind_pat_vars(T, X, Vars, Way),
                            bind_pat_vars(cerl:c_values(Ts), X, Env1, Way);
                        [] ->
                            Vars
                    end;
                false ->
                    case {cerl:values_es(P), referl_erl_types:t_to_tlist(X)} of
                        {[T | Ts], [Ty | Tys]} ->
                            Env1 = bind_pat_vars(T, Ty, Vars, Way),
                            bind_pat_vars(cerl:c_values(Ts), 
                                          referl_erl_types:t_product(Tys), Env1, Way);
                        {[],[]} ->
                            Vars
                    end
            end
    end.

pat_type(P, Vars) ->
    case cerl:type(P) of
        var ->
            dict:fetch(cerl:var_name(P), Vars);
        literal ->
            referl_erl_types:t_from_term(cerl:concrete(P));
        cons ->
            referl_erl_types:t_cons(pat_type(cerl:cons_hd(P), Vars), 
                             pat_type(cerl:cons_tl(P), Vars));
        tuple ->
            referl_erl_types:t_tuple([pat_type(E, Vars) || E <- cerl:tuple_es(P)]);
        binary ->
            referl_erl_types:t_binary();
        alias ->
            pat_type(cerl:alias_pat(P), Vars)
    end.

top_or_bottom(T) ->
    case referl_erl_types:t_is_none(T) of
        true ->
            T;
        false ->
            referl_erl_types:t_any()
    end.

guard_filters(T, Env) ->
    guard_filters(T, Env, dict:new()).

guard_filters(T, Env, Vars) ->
    case cerl:type(T) of
        call ->
            M = cerl:call_module(T),
            F = cerl:call_name(T),
            case cerl:is_c_atom(M) andalso cerl:is_c_atom(F) of
                true ->
                    As = cerl:call_args(T),
                    case {cerl:atom_val(M), cerl:atom_val(F), length(As)} of
                        {erlang, 'and', 2} ->
                            [A1, A2] = As,
                            guard_filters(A1, guard_filters(A2, Env));
                        {erlang, is_atom, 1} ->
                            filter(As, referl_erl_types:t_atom(), Env);
                        {erlang, is_binary, 1} ->
                            filter(As, referl_erl_types:t_binary(), Env);
                        {erlang, is_float, 1} ->
                            filter(As, referl_erl_types:t_float(), Env);
                        {erlang, is_function, 1} ->
                            filter(As, referl_erl_types:t_fun(), Env);
                        {erlang, is_integer, 1} ->
                            filter(As, referl_erl_types:t_integer(), Env);
                        {erlang, is_list, 1} ->
                            filter(As, referl_erl_types:t_maybe_improper_list(), Env);
                        {erlang, is_number, 1} ->
                            filter(As, referl_erl_types:t_number(), Env);
                        {erlang, is_pid, 1} ->
                            filter(As, referl_erl_types:t_pid(), Env);
                        {erlang, is_port, 1} ->
                            filter(As, referl_erl_types:t_port(), Env);
                        {erlang, is_reference, 1} ->
                            filter(As, referl_erl_types:t_reference(), Env);
                        {erlang, is_tuple, 1} ->
                            filter(As, referl_erl_types:t_tuple(), Env);
                        _ ->
                            Env
                    end;
                false ->
                    Env
            end;
        var ->
            case dict:find(cerl:var_name(T), Vars) of
                {ok, T1} ->
                    guard_filters(T1, Env, Vars);
                error ->
                    Env
            end;
        'let' ->
            case cerl:let_vars(T) of
                [V] ->
                    guard_filters(cerl:let_body(T), Env,
                                  dict:store(cerl:var_name(V), cerl:let_arg(T),
                                             Vars));
                _ ->
                    Env
            end;
        values ->
            case cerl:values_es(T) of
                [T1] ->
                    guard_filters(T1, Env, Vars);
                _ ->
                    Env
            end;
        _ ->
            Env
    end.

filter(As, X, Env) ->
    [A] = As,
    case cerl:type(A) of
        var ->
            V = cerl:var_name(A),
            case dict:find(V, Env) of
                {ok, X1} ->
                    dict:store(V, meet(X, X1), Env);
                error ->
                    dict:store(V, X, Env)
            end;
        _ ->
            Env
    end.



bind_vars(Vs, Xs, Vars) ->
    if length(Vs) =:= length(Xs) ->
	    bind_vars_list(Vs, Xs, Vars);
       true ->
	    bind_vars_single(Vs, referl_erl_types:t_none(), Vars)
    end.

bind_vars_list([V | Vs], [X | Xs], Vars) ->
    bind_vars_list(Vs, Xs, dict:store(cerl:var_name(V), X, Vars));
bind_vars_list([], [], Vars) ->
    Vars.

bind_vars_single([V | Vs], X, Vars) ->
    bind_vars_single(Vs, X, dict:store(cerl:var_name(V), X, Vars));
bind_vars_single([], _X, Vars) ->
    Vars.


primop_type(match_fail, 1, _) -> referl_erl_types:t_none();
primop_type(_, _, Xs) -> strict(Xs, referl_erl_types:t_any()).

strict(Xs, T) ->
    case referl_erl_types:any_none(Xs) of
        true ->
            referl_erl_types:t_none();
        false ->
            T
    end.

meet(X, Y) -> referl_erl_types:t_inf(X, Y).

%%join(X, Y) when is_list(X), is_list(Y) ->
%%    referl_erl_types:t_sup(X ++ Y);
join([X], [Y]) ->
    %?d(join, {list, x, X, y, Y}),
    referl_erl_types:t_sup(X, Y);

join(X, Y) -> 
    %?d(join, {other, x, X, y, Y}),
    referl_erl_types:t_sup(X, Y).

join_list([Xs | Xss]) ->
    join(Xs, join_list(Xss));
join_list([]) ->
    referl_erl_types:t_none().

call_type(M, F, A, Xs) ->
    referl_erl_bif_types:type(M, F, A, Xs).

arg_type(M, F, A) ->
    referl_erl_bif_types:arg_types(M, F, A).

call_site(DomainTypes, ArgsTypes, _Args, St) ->
    bind_args(DomainTypes, ArgsTypes, St).

%% If the arity does not match the call, nothing is done here.

bind_args(Vs, Xs, Vars) ->
    if length(Vs) =:= length(Xs) ->
            bind_args(Vs, Xs, Vars);
       true ->
            Vars
    end.

%% bind_args([V | Vs], [X | Xs], Vars) ->
%%     Vars1 = bind_arg(V, X, Vars),
%%     bind_args(Vs, Xs, Vars1);
%% bind_args([], [], Vars) ->
%%     Vars.

%% bind_arg(L, X, Vars, Limit) ->
%%     bind_arg(L, X, Vars, Limit, false).

%% bind_arg(L, X, Vars) ->
%%     X0 = dict:fetch(L, Vars),
%%     X1 = join(X, X0),
%%     case X0 =:= X1 of
%%         true ->
%%             Vars;
%%         false ->
%%             {dict:store(L, X1, Vars), true}
%%     end.


get_local_type(VarName) ->
    %io:format("get_type: ~p~n",[VarName]),
    case ets:lookup(?FUN_STORE, VarName) of
        [F] ->
            %?d(get_local_type, {varname, VarName, f, F#f.type}),
            {F#f.type, F#f.state};
        [] ->
            {referl_erl_types:t_fun(), ?complete}
    end.











% @private
% For debug purposes. It prints the cerl_tree, for inspection.
print(FileName) ->
    {ok, _, Tree} = compile_to_core(FileName),
    io:format("~p~n", [Tree]).

