%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

%%% @author Lilla Hajs <lya@elte.hu>

-module(refusr_sq).
-vsn("$Rev: 12312 $ ").

-export([run/3]).
-export([format_nodes/2]).
-export([prepare/1, error_text/2]).
-export([closure_worker/5, chains_worker/5, pmap_helper/2]).

-include("user.hrl").
-include("sq_lib.hrl").
-define(Lib, refusr_sq_lib).
-define(Format, refusr_sq_format).


%%% ============================================================================
%%% Errors

error_text(lexical_error, Error) ->
    refusr_sq_lexer:format_error(Error);
error_text(syntax_error, Error) ->
    refusr_ac_parser:format_error(Error);
error_text(illegal_property, Params) ->
    io_lib:format("illegal ~p property: ~p", Params);
error_text(illegal_selector, Params) ->
    io_lib:format("illegal ~p selector: ~p  ", Params);
error_text(illegal_operator, Params) ->
    io_lib:format("illegal ~p operator: ~p  ", Params);
error_text(type_mismatch, [At, Of, ExpType]) ->
    [ case At of
          {Pl,De} -> io_lib:format("type mismatch at ~p ~p :\n    ", [Pl,De]);
          Pl -> io_lib:format("type mismatch at ~p:\n    ", [Pl])
      end,
      case Of of
          {Ent, Type} -> io_lib:format("the type of ~p is ~s, not ~s",
                                       [Ent, Type, ExpType]);
          Ent -> io_lib:format("the type of ~p is not ~s", [Ent, ExpType])
      end ];
error_text(no_property_in_comparison, []) ->
    "no property in comparison";
error_text(statistics_error, []) ->
    "statistics are only available for properties with numeric values";
error_text(bad_regexp, [])->
    "illegal regexp";
error_text(bad_typing,[]) ->
    "bad grouping type, else the value is less than or equal to 0.";    
error_text(grouping_error, []) ->
     "grouping value run out of the query's length";
error_text(mult_error, []) ->
     "multiplicity must be greater than 0".
%%% ============================================================================
%%% Callbacks

%% @spec run(DisplayOpt::proplist(), Params::proplist(), Query::string()) ->
%%           QueryResult::term()
%% @doc Returns the result of `Query' starting from the initial state given by
%%      `Params'. The format of the result is determined by `DisplayOpt'.
%%
%%      `Params' contains either the
%%          - (optional) `file' and `position' keys or
%%          - the `node_list' and (optional) `node_type' key.
%%       The possible values for
%%          - `node_type': file|function|record|field|macro|variable|expression
%%
%%      `DisplayOpt' contains the keys `positions' and `output'.
%%      The possible values for
%%          - `positions': none|scalar|linecol|both
%%          - `output': stdio|{iodev,io_device()}|msg|other|nodes
%%
%%      The `QueryResult' depends on the `output' key in `DisplayOpt'.
%%          - stdio: a formatted text written to stdio
%%          - {iodev,Dev::io_device()}: a formatted text written to Dev
%%          - msg: a message containing a list with the following types of
%%                 elements: {eq, Name, Value} |
%%                           {list, [{Position, Text}]} |
%%                           {chain, [{Position, Text}], PostWS} |
%%                           {group_by, {Position, Text}, eq, Name, Value} |
%%                           {group_by, {Position, Text},
%%                            list, [{Position, Text}]}
%%          - other: the same list that is otherwise sent by a message
%%          - nodes: a proplist with the keys: `nodes' for a list of nodes and
%%                                             `text' for a formatted text
%%      The format of positions depends on the `positions' key in `DisplayOpt'.
%%          - none: nopos
%%          - scalar: {File::string(), PosFrom::integer(), PosTo::integer()}
%%          - linecol: {File::string(), PosFrom::{integer(), integer()},
%%                                      PosTo::{integer(), integer()}}
run(DisplayOpt, Params, Query) when is_list(Query) ->
    Tokens   = tokenize(?Lib:replace_special_chars(Query)),
    SynTree  = parse(Tokens),
    SynTree2 = make_internal_representation(SynTree),
    Result   = process_semantic_query(Params, SynTree2, DisplayOpt),%?d(Result),
    ?Format:result(Result, DisplayOpt).

make_internal_representation(SynTree) ->
    case refusr_ac:validate_query(SynTree) of
        {ok, Res} ->
            Res;
        {error, E} ->
            throw(?LocalError(syntax_error, E))
    end.

parse(Tokens) ->
    case refusr_ac_parser:parse(Tokens) of
        {ok, SynTree} ->
            SynTree;
        {error, {_, _, Err}} ->
            case lists:flatten(Err) of
                "{quoted," ++ A ->
                    A1 = lists:reverse(tl(lists:reverse(A))),
                    throw(?LocalError(syntax_error, A1));
                _ ->
                    throw(?LocalError(syntax_error, Err))
            end
    end.

tokenize(Query) ->
    case refusr_sq_lexer:string(Query) of
        {ok, Tokens, _} ->
            Tokens;
        {error, {_, _, Error}, _} ->
            throw(?LocalError(lexical_error, Error))
    end.

%% @private
prepare(Args) ->
    fun () ->
        [DisplayOpt, StartOpt, QueryStr] =
            ?MISC:pgetu([display_opt, start_opt, querystr], Args),
        run(DisplayOpt, StartOpt, QueryStr)
    end.

%%% ============================================================================
%%% Implementation

process_semantic_query(Params, SemanticQuery, DisplayOpt) ->
    Split =
    lists:foldr(
      fun(Filter = {filter, _F}, {undef, undef, [], Last}) ->
          {undef, undef, [], [Filter|Last]};
         ({initial_selector, InitialSelector}, {undef, Type, Mid, Last}) ->
          {InitialSelector, Type, Mid, Last};
         (Elem, {undef, undef, [], Last}) ->
          {undef, element(1,Elem), [], [Elem|Last]};
         (Filter = {filter, _F}, {undef, Type,Mid, Last}) ->
          {undef, Type, [Filter|Mid], Last};
         (Closure = {closure, _S, _M}, {undef, closure, Mid, Last}) ->
          {undef, closure, [], [Closure|Mid++Last]};
         (Iteration = {iteration, _S, _M}, {undef, iteration, Mid, Last}) ->
          {undef, iteration, [], [Iteration|Mid++Last]};
         (Elem, {undef, _Type, Mid, Last}) ->
          {undef, mid, [Elem|Mid], Last}
      end,
      {undef, undef, [], []},
      SemanticQuery),

    {InitialSelector, Seq, Last} =
    case Split of
        {undef, _, [{set_op, _}=H|T], L} ->
            {H, T, L};
        {undef, _, [], [{set_op, _}=H|T]} ->
            {H, [], T};
        {I, _, S, L} ->
            {I, S, L}
    end,

    SeqLength = length(Seq++Last) -
                length(proplists:get_all_values(filter, Seq++Last)),
    GroupBy = case proplists:get_value(groupby,DisplayOpt) of
        undefined ->
            SeqLength;
        GroupByOpt ->
            ?Check(is_integer(GroupByOpt) andalso (GroupByOpt>0),
                    ?LocalError(bad_typing, [])),
            ?Check(SeqLength >= GroupByOpt,
                    ?LocalError(grouping_error, [])),
            GroupByOpt
    end,

    Init = initial_state(#state{params=Params}, InitialSelector),
    InitState = 
    case (Seq/=[] orelse Last/=[]) andalso is_group_state(Init) of
        true ->
            Init#state{res=lists:flatten(Init#state.res),
                        groupby_type=[], groupby_res=[],
                        groupby_place=GroupBy, entitynumber=1};
        _ ->
            Init#state{groupby_place=GroupBy, entitynumber=1}
    end,

    SeqState = lists:foldl(fun check/2, InitState, Seq),
    ChkState = lists:foldl(fun check/2, SeqState#state{checked_query=[]}, Last),

    case find_help(ChkState) of
        [] -> 
            State = process_query_seq(InitState, SeqState#state.checked_query),
            LastState = process_last_query(State, ChkState#state.checked_query),
            case SeqLength==0 orelse SeqLength==GroupBy of
                true ->
                    LastState;
                _ ->
                    grouped_prop_query_naming(LastState, GroupBy-1, Seq++Last)
            end;
        [Help|_] -> Help
    end.

process_query_seq(State, QuerySeq) ->
      lists:foldr(fun process/2, State, QuerySeq).

process_last_query(State, LastQuery) ->
    lists:foldr(fun process_last/2, State, LastQuery).

find_help(E) when is_list(E) ->
    lists:append(lists:map(fun find_help/1, E));

find_help(E) when is_tuple(E) ->
    case E of
        {help, {_, _}} -> [E];
        {help, _} -> [{help, {initial_selectors, []}}];
        _ -> find_help(tuple_to_list(E))
    end;

find_help(_) ->
    [].

grouped_prop_query_naming(#state{action=property_query}=State, Skip, QuerySeq) ->
    case is_group_state(State) of
        true ->
            Converted = [query_selectors_to_string(Sel) || Sel<-QuerySeq],
            Filtered = lists:filter(fun(L)->length(L)>0 end, Converted),
            Seq = lists:append(lists:nthtail(Skip, Filtered)),
            State#state{type = Seq};
        _ ->
            State
    end;

grouped_prop_query_naming(State, _Skip, _Query) ->
    State.

query_selectors_to_string(Selector) ->
    case Selector of
        {selector, Sel} ->
            "."++atom_to_list(Sel);
        {initial_selector, Sel} ->
            atom_to_list(Sel);
        {iteration, {seq, Sels}, {mult, Mult}} ->
            ".{" ++
            tl(lists:append([query_selectors_to_string(Sel) || Sel<-Sels])) ++
            "}" ++ integer_to_list(Mult);
        {closure, {seq, Sels}, {mult, Mult}} ->
            ".(" ++
            tl(lists:append([query_selectors_to_string(Sel) || Sel<-Sels])) ++
            ")" ++
            case Mult of
                infinite -> 
                    "+";
                N ->
                    integer_to_list(N)
            end;
        {set_op, {_, Q1, Q2}} ->
            case hd(Q1) of
                {initial_selector, _} ->
                    lists:append([query_selectors_to_string(Sel) || Sel<-Q2]);
                _ ->
                    lists:append([query_selectors_to_string(Sel) || Sel<-Q1])
            end;
        _ ->
            []
    end.

semantic_query_last_type(SemanticQuery) ->
    Query = case find_help(SemanticQuery) of
                [{help, _}]=H -> H;
                _ -> SemanticQuery
            end,
    case Query of
        [{help, _}] -> help;

        [{set_op, {_, Q1, Q2}} | Seq] -> 
            Type1 = semantic_query_last_type(Q1),
            Type2 = semantic_query_last_type(Q2),
            Type = merged_type(Type1, Type2),
            ?Check(Type /= null,
                    ?LocalError(type_mismatch, [hd(SemanticQuery), Q2, Type1])),
            LastState = lists:foldl(fun check/2, #state{type=Type}, Seq),
            LastState#state.type;
        [{initial_selector, InitialSelector}| Seq] ->
            Init_sel_type = fun({quoted, InitSel}) ->
                         ?Lib:init_sel_type(InitSel);
                        (InitSel) ->
                         ?Lib:init_sel_type(InitSel)
                     end,
            InitType = Init_sel_type(InitialSelector),
            LastState = lists:foldl(fun check/2, #state{type=InitType}, Seq),
            LastState#state.type;
        _ -> any
    end.

merged_type(Type1, Type2) ->
    case Type1 of
        Type2 -> Type2;
        _ ->
            case ordsets:is_subset(
                    ordsets:from_list([Type1, Type2]),
                    [atom, bool, int, string]) of
                true -> string;
                _ -> 
                    case {Type1, Type2} of
                        {any, T} -> T;
                        {T, any} -> T;
                        {help, _} -> help;
                        {_, help} -> help;
                        _ -> null
                    end
            end
    end.


%%% ============================================================================
%%% Preprocessing of queries

merge_results(SetOp, #state{}=Res1, #state{}=Res2) ->
    merge_results(SetOp, Res1#state.res, Res2#state.res);

merge_results(SetOp, Res1, Res2) ->
    case SetOp of
        union -> ordsets:union(
                    ordsets:from_list(Res1),
                    ordsets:from_list(Res2)
                 );
        intersect -> ordsets:intersection(
                    ordsets:from_list(Res1),
                    ordsets:from_list(Res2)
                 );
        minus -> ordsets:subtract(
                    ordsets:from_list(Res1),
                    ordsets:from_list(Res2)
                 )
    end.

combine_same_groups(Group1, Group2) ->
    Same_added =
    [{ordsets:from_list(Ents ++
        case lists:keyfind(G, 2, Group2) of
            false -> [];
            {E, _} -> E
        end
        ), G} || {Ents, G} <- Group1],

    Only_in_G2 =
    lists:filter(
        fun({_, Key})->
            not(lists:keymember(Key, 2, Group1))
        end,
        Group2),

    lists:unzip(Same_added ++ Only_in_G2).

group_by_mod(State) ->
    St = 
    case State#state.type of
        file ->
            State#state{
                groupby_type = file,
                groupby_res = lists:append(State#state.res),
                res=[[E]||E <- lists:append(State#state.res)]
            };
        _ -> case State#state.groupby_type of
                [] -> State#state{groupby_res = State#state.res};
                _ -> State
             end
    end,

    Mods = [?Lib:node_file(N) || N <- St#state.groupby_res],

    Merge_same =
    fun(Groups) ->
        Gs = lists:keysort(1, Groups),
        lists:ukeysort(1,
            [ {K, ordsets:union([ordsets:from_list(E)|| {Key, E}<-Gs, Key==K])}
                || {K, _} <- Gs]
        )
    end,

    {PrevRes, Res} = 
        lists:unzip(
            Merge_same(lists:zip(Mods, St#state.res))
        ),
    St#state{groupby_type=file, groupby_res=PrevRes, res=Res}.

merge_states(SetOp, S1, S2) ->
    case {to_group_state(S1), to_group_state(S2)} of
        {State1, State2} when SetOp/=union ->
            State1#state{
                res = [merge_results(SetOp, E, lists:append(State2#state.res))
                        || E<-State1#state.res]
            };

        {#state{action=Action, type=Type, groupby_type=GType} = A,
            #state{action=Action, type=Type, groupby_type=GType} = B} ->
                {F_res, F_groupby_res} =
                    combine_same_groups(
                        lists:zip(A#state.res, A#state.groupby_res),
                        lists:zip(B#state.res, B#state.groupby_res)
                    ),
                A#state{
                    res = filter_same_results(F_res),
                    groupby_res = F_groupby_res};

        {#state{action = Action, groupby_type = GType} = A,
            #state{action=Action, groupby_type=PrevType} = B}
            when GType==[] orelse PrevType==[] orelse GType /= PrevType ->
                merge_states(SetOp, group_by_mod(A), group_by_mod(B));

        {A, B} ->
            case A#state.res of
                [] -> B;
                _  -> A
            end
    end.

property_to_string({Type, State1}, {Type, State2}) ->
    {State1, State2};

property_to_string({_, State1}, {_, State2}) ->
    {property_to_string(State1), property_to_string(State2)}.

property_to_string(#state{action=property_query}=State) ->
    State#state{
        type = value,
        res = 
            case is_group_state(State) of
                true -> [[?MISC:any_to_string(E) || E<-G] || G<-State#state.res];
                _    -> [?MISC:any_to_string(E) || E<-State#state.res]
            end
    };

property_to_string(State) ->
    State.

all_to_string(L) when is_list(L) ->
    [to_string(E) || E<-L].

to_string({quoted, Val}) ->
    to_string(Val);

to_string(E) -> ?MISC:any_to_string(E).

to_group_state(State) ->
    case is_group_state(State) of
        true -> State;
        _ ->
            State#state{
                res = [[E] || E <- State#state.res]
            }
    end.

is_group_state(State) ->
    case State#state.groupby_type of
        [] -> false;
        _ ->
            case State#state.action of
                property_query ->
                    lists:any(
                        fun ([H|_]) -> not(is_integer(H));
                            (_) -> false
                        end, State#state.res);
                _ -> true
            end
    end.

filter_empty([[]])->[];
filter_empty(L)->L.

filter_same_results(Results) ->
    case Results of
        [H|T] -> 
            [[E||E<-H, not(lists:member(E, lists:append(T)))] | filter_same_results(T)];
        _ -> Results
    end.

%todo:  node list <-> query format check
initial_state(State, {set_op, {SetOp, Q1, Q2}=Q}) ->
    ?Check( SetOp /= any_in andalso SetOp /= all_in,
            ?LocalError(illegal_operator, [SetOp, Q])),

    Type1 = semantic_query_last_type(Q1),
    Type2 = semantic_query_last_type(Q2),

    ?Check( merged_type(Type1, Type2) /= null,
            ?LocalError(type_mismatch, [SetOp, {Q1, Type1}, Type2])),

    SQ1 = process_semantic_query(State#state.params, Q1, []),
    SQ2 = process_semantic_query(State#state.params, Q2, []),

    {Res1, Res2} = property_to_string({Type1, SQ1}, {Type2, SQ2}),

    merge_states(SetOp, Res1, Res2);

initial_state(State, {quoted, InitialSelector}) ->
    initial_state(State, InitialSelector);

initial_state(#state{params=Params}=State, InitialSelector)->
    {NodeType, NodeList} =
        case proplists:get_value(node_list, Params) of
            undefined ->
                case InitialSelector of
                    undef -> {none, []};
                    _ ->
                        ?Lib:init_sel([{ask_missing,false} |Params],
                                      InitialSelector)
                end;
            [] ->
                {none, []};
            [HNode|TNodes] = Nodes ->
                Type = proplists:get_value(node_type, Params,
                                           ?Lib:node_type(HNode)),

                NT = [{Node, ?Lib:node_type(Node)}|| Node <- TNodes],
                Diffs = lists:filter(fun({_N, NType}) -> NType/=Type end, NT),
                case Diffs of
                    [] ->
                        {Type, Nodes};
                    _ ->
                        Params = ["initial state", hd(Diffs), Type],
                        throw(?LocalError(type_mismatch, Params))
                end
        end,
    State#state{type=NodeType, res=ordsets:from_list(NodeList)}.

check({set_op, {SetOp, Q1, Q2}}, State) ->
    Check_query_seq =
    fun([{Sel, _}|_]=QuerySeq) when Sel==initial_selector ->
            #state{
                type = semantic_query_last_type(QuerySeq),
                checked_query = {sem_query, QuerySeq}
            };
        (QuerySeq) -> 
            lists:foldl(fun check/2, State#state{checked_query=[]}, QuerySeq)
    end,
    CheckedQ1 = Check_query_seq(Q1),
    CheckedQ2 = Check_query_seq(Q2),
    Type1 = CheckedQ1#state.type,
    Type2 = CheckedQ2#state.type,

    ?Check( merged_type(Type1, Type2) /= null,
            ?LocalError(type_mismatch, [SetOp, {Q1, Type1}, Type2])),

    State#state{
        type = merged_type(Type1, Type2),
        checked_query = [
            {set_op, 
                {SetOp,
                    CheckedQ1#state.checked_query,
                    CheckedQ2#state.checked_query
                }} | State#state.checked_query]};

check({selector, Sel}, #state{type=Type, checked_query=Lst}=State) ->
    Selector =
        case Sel of
            {quoted, Atom} -> Atom;
            Atom -> Atom
        end,
    case ?Lib:sel_type(Type, Selector) of
        [SelType] ->
            State#state{
                type = SelType,
                checked_query = [{selector, Selector, SelType}| Lst]};
        [] ->
            case ?Lib:prop_type(Type, Selector) of
                [PropType] ->
                    State#state{
                        type = PropType,
                        checked_query = [{property, Selector, PropType}| Lst]};
                [] ->
                    throw(?LocalError(illegal_selector, [Type, Selector]))
            end
    end;

check({Action, {seq, Seq}, {mult, Mult}}, #state{type=Type, checked_query=Lst}=State) -> 
    case lists:foldl(fun check/2, State#state{checked_query = []}, Seq) of
        #state{type=Type, checked_query=QSLst} -> ?Check( (Mult>0),
                       ?LocalError(mult_error, [])),
            case Mult of
                0 -> State#state{checked_query = Lst};
                _ -> State#state{checked_query = [{Action, QSLst, Mult}| Lst]}
            end;
        #state{type=BadType} ->
            throw(?LocalError(type_mismatch, [Action, {Seq, BadType}, Type]))
    end;

check({filter, Filter}, #state{checked_query=Lst}=State) ->
    State#state{checked_query = [{filter, check_filter(State, Filter)}|Lst]};

check({statistics, Stat}, #state{type=PropType}=State) ->
    Statistics =
        case Stat of
            {quoted, S} -> S;
            S           -> S
        end,
    case PropType of
        any ->
            State#state{type=int, checked_query = [{statistics, Statistics}]};
        int ->
            State#state{type=int, checked_query = [{statistics, Statistics}]};
        _ ->
        throw(?LocalError(statistics_error, []))
    end;

%todo: filterekben help?
check({help, HelpType}, #state{type=Type}=State) ->
    State#state{type=help, checked_query={help,
        case HelpType of
            initial_selectors -> {initial_selectors, []};
            queries           -> {selectors, Type};
            statistics        -> {statistics, []};
            filters           -> {properties, Type}
        end}
    }.


%% @private
%% @spec check_filter(Type::atom(), Filter::atom()|tuple()) -> atom()|tuple()
check_filter(_State, 'true') ->
    'true';

check_filter(_State, 'false') ->
    'false';

check_filter(State, {'or', Filter1, Filter2}) ->
    {'or', check_filter(State, Filter1), check_filter(State, Filter2)};

check_filter(State, {'and', Filter1, Filter2}) ->
    {'and', check_filter(State, Filter1), check_filter(State, Filter2)};

check_filter(State, {'not', Filter}) ->
    {'not', check_filter(State, Filter)};

check_filter(#state{type=Type} = State, {set_op, {SetOp, Q1, Q2}=Filter}) ->
    Check_filter_setop =
    fun(F) -> 
        case F of
            {seq, Q} ->
                Seq = lists:foldl(fun check/2, State#state{checked_query=[]}, Q),
                {Seq#state.type, {seq, Seq#state.checked_query}};
            {cons,_} -> 
                {any, F};
            [{initial_selector, _}|_] = Q ->
                {semantic_query_last_type(Q), {'query', F}};
            Q when is_list(Q) ->
                Seq = lists:foldl(fun check/2, State#state{checked_query=[]}, Q),
                {Seq#state.type, {seq, Seq#state.checked_query}};
            Property when is_atom(Property) ->
                PropType = case ?Lib:prop_type(Type, Property) of
                            [T] -> T;
                            _ -> null
                        end,
                {PropType, Property}
        end
    end,

    {Type1, Q1Q} = Check_filter_setop(Q1),
    {Type2, Q2Q} = Check_filter_setop(Q2),

    ?Check( merged_type(Type1, Type2) /= null,
        ?LocalError(type_mismatch, [{filter, Filter}, {Q2}, Type1])),

    {set_op, {SetOp, Q1Q, Q2Q}};

check_filter(State, {'in', P1, P2})->
    check_filter(State, {set_op, {all_in, P1, P2}});

check_filter(State, {seq, Seq}) ->
    St = lists:foldl(fun check/2, State#state{checked_query=[]}, Seq),
    {seq, St#state.checked_query};

check_filter(_State, {_CompOp, {quoted, _}, {quoted, _}}) ->
    throw(?LocalError(no_property_in_comparison, []));

check_filter(State, {CompOp, {quoted, _} = F1, F2}) ->
    check_filter(State, {CompOp, F2, F1});

check_filter(#state{type=Type}=State, {CompOp, F1, {quoted, _} = F2}) ->
    if
        is_tuple(F1) -> {CompOp, check_filter(State, F1), F2};
        is_atom(F1)  ->
            PropType1 = ?Lib:prop_type(Type, F1),
            ?Check(ordsets:is_element(
                PropType1,
                [[atom], [bool], [int], [string]]
            ),
                   ?LocalError(no_property_in_comparison, [])),
            {CompOp, F1, F2};
        true ->
            throw(?LocalError(no_property_in_comparison, []))
    end;

check_filter(State, {CompOp, F1, F2}) when is_tuple(F1) andalso is_tuple(F2) ->
    {CompOp, check_filter(State, F1), check_filter(State, F1)};

check_filter(#state{type=Type}=State, {CompOp, F1, F2}) when is_tuple(F1) ->
    ?Check(?Lib:prop_type(Type, F2) == [bool] orelse is_boolean(F2),
           ?LocalError(type_mismatch, [{filter,{CompOp, F1, F2}}, F2, bool])),
    {CompOp, check_filter(State, F1), F2};

check_filter(State, {CompOp, F1, F2}) when is_tuple(F2) ->
    {CompOp, CheckedF2, F1} = check_filter(State, {CompOp, F2, F1}),
    {CompOp, F1, CheckedF2};

check_filter(#state{type=Type}, {CompOp, F1, F2}) when is_atom(F1) andalso is_atom(F2)->
    PropType1 = ?Lib:prop_type(Type, F1),
    PropType2 = ?Lib:prop_type(Type, F2),
    ?Check(PropType1 /= [] orelse PropType2 /= [],
           ?LocalError(no_property_in_comparison, [])),

    Type1 = case PropType1 of [] -> atom; [H] -> H end,
    Type2 = case PropType2 of [] -> atom; [Hd] -> Hd end,

    ?Check(merged_type(Type1, Type2) /= null,
            ?LocalError(type_mismatch, [{filter, {CompOp, F1, F2}},
                                                  {F1, Type1}, Type2])),
    {CompOp, F1, F2};

check_filter(#state{type=Type}, {CompOp, F1, F2}) when is_atom(F1) ->
    PropType = ?Lib:prop_type(Type, F1),
    ?Check(PropType /= [],
           ?LocalError(no_property_in_comparison, [])),

    ?Check(merged_type(hd(PropType), hd(type_of(F2))) /= null,
           ?LocalError(type_mismatch,
                       [{filter, {CompOp, F1, F2}}, F2, hd(PropType)])),

    {CompOp, F1, F2};

check_filter(State, {CompOp, F1, F2}) when is_atom(F2) ->
    check_filter(State, {CompOp, F2, F1}),
    {CompOp, F1, F2};

check_filter(_Type, {_CompOp, _F1, _F2}) ->
    throw(?LocalError(no_property_in_comparison, []));

check_filter(#state{type=Type}, Filter) ->
    PropType = ?Lib:prop_type(Type, Filter),
    case PropType of
        [bool] ->
            Filter;
        [] ->
            throw(?LocalError(illegal_property, [Type, Filter]));
        _ ->
            throw(?LocalError(type_mismatch,
                              [{filter,Filter}, {Filter, hd(PropType)}, bool]))
    end.

type_of(E) when is_integer(E) -> [int];
type_of(true) -> [bool];
type_of(false) -> [bool];
type_of(E) when is_atom(E) -> [atom];
type_of([H|_]) when is_integer(H) -> [string];
type_of(E) when is_list(E) -> [list];
type_of(_) -> [any].
%%% ============================================================================
%%% Processing of queries

process_set_op_operand(State, Q) ->
    case Q of
        {sem_query, SQ} ->
            SemQuery = process_semantic_query(State#state.params, SQ, []),
            Res = case is_group_state(SemQuery) of
                false -> SemQuery;
                _  -> SemQuery#state{res=lists:append(SemQuery#state.res)}
            end,
            fun(_) -> Res end;
        _ ->
            fun([]) -> process_query_seq(
                        #state{
                            res=[],
                            type=State#state.type,
                            params=State#state.params}, Q);
                (E) -> process_query_seq(
                        #state{
                            res=[E],
                            type=State#state.type,
                            params=State#state.params}, Q)
            end
    end.

type_set_op_operand(State, Q) ->
    case Q of
        {sem_query, SQ} ->
            semantic_query_last_type(SQ);
        _ ->
            (process_query_seq(#state{res=[], type=State#state.type}, Q))#state.type
    end.

process({set_op, _}=SetOp, State=#state{entitynumber=EntNum, 
                                        groupby_place=GbyPlace}) ->
    Res=process_last(SetOp, State),
    E=Res#state{
        res =   case EntNum>=GbyPlace of
                    true -> Res#state.res;
                    _ -> ordsets:union(Res#state.res)
                end},
    grouping(E, State);
      
process({selector, Selector, SelType}, St=#state{type= Type, res= Entities, 
                                                 entitynumber=EntNum,
                                                groupby_place=GbyPlace}) ->
    [Fun] = ?Lib:sel_fun(Type, Selector),
    Res = pmap({?MODULE, pmap_helper},
                [fun(E)->ordsets:from_list(lists:flatten(Fun(E))) end],
                Entities),

    Result = #state{
        action = selection,
        type = SelType,
        res  =  case EntNum>=GbyPlace of
                    true -> Res;
                    false -> ordsets:union(Res)
                end},
    grouping(Result, St);           

process({_Action, _QSLst, 0}, State) -> State;

process({iteration, QSLst, Mult}, State=#state{entitynumber=EntNum}) ->
    {F,R} = case QSLst of
        [H] ->
            {[H],[]};
        [_|_] ->
            {[lists:last(QSLst)],lists:reverse(tl(lists:reverse(QSLst)))}
    end,
    First = process_query_seq(State, F),

    ModFirst = if
        State#state.groupby_place > EntNum ->
            First#state{entitynumber=-Mult*length(QSLst)};
        true ->
            First
    end,

    (process_query_seq(ModFirst,
        lists:append(lists:duplicate(Mult-1, QSLst)) ++ R)
    )#state{entitynumber=EntNum+1};

process({closure, QSLst, Mult}, #state{res=Ents,
                                      entitynumber=EntNum,
                                      groupby_place=GbyPlace}=St)->
    Tab = ets:new(store,
        [public, ordered_set, {write_concurrency, true}, {keypos, 2}]),

    case EntNum>=GbyPlace of 
    true->
        Entities = case EntNum of
            GbyPlace -> [[E]||E<-Ents];
            _ -> Ents
        end, 
        [[ets:insert(Tab, {N, E}) || E <- lists:nth(N, Entities)] ||
            N<-lists:seq(1, length(Entities))],

        KeyList = [ rpc:async_call(node(), ?MODULE, closure_worker,
            [St#state{res=lists:nth(N, Entities)}, QSLst, Mult, Tab, N]) ||
                N<-lists:seq(1, length(Entities))],
        [ok = rpc:yield(Key) ||Key<-KeyList],
        Res = [ets:select(Tab, [{{N, '$2'},[], ['$2']}]) ||
                N <- lists:seq(1, length(Entities))];
    false ->
        [ets:insert(Tab, {0, E}) || E <- Ents],
        EntityLists = group_entities(Ents),
        KeyList =[ rpc:async_call(node(), ?MODULE,closure_worker,
                [St#state{res=E},QSLst, Mult, Tab, 0])|| E <- EntityLists],
        [ ok = rpc:yield(Key) || Key <- KeyList],
        Res = ets:select(Tab, [{{0, '$2'},[], ['$2']}])
    end,

    Result = St#state{action = closure, res = Res},
    ets:delete(Tab),
    grouping(Result, St);

process({filter, Filter}, #state{res=Res, entitynumber=EntNum,
                                 groupby_place=GbyPlace}=St) ->
    St#state{
        res=case GbyPlace < EntNum of 
            true ->
                pmap({?MODULE, pmap_helper},
                    [fun(Ent) -> filter(Filter,St#state{res=Ent}) end], 
                    Res);
            false ->
                filter(Filter,St) 
        end};

process({property, Prop, PropType}, #state{type=Type, 
                                              res=Entities,params=Params}) ->
    #state{action    = property_query,
           type      = PropType,
           res       = pmap({?MODULE, pmap_helper},
                            ?Lib:prop_fun(Type, Prop),
                            lists:flatten(Entities)),
           params    = Params}.
           
group_entities(Entities) ->
    NumberOfData = length(Entities),
    Divide =
        case erlang:system_info(logical_processors_available) of
            unknown ->
                (NumberOfData div 10) + 1;
            N ->
                (NumberOfData div (N*10)) + 1
        end,
    ?MISC:slice_list(Divide, Entities).

%% @private
closure_worker(_State, _QSLst, 0, _Tab, _Group) ->
    ok;
closure_worker(#state{res=[]}, _QSLst, _Mult, _Tab, _Group) ->
    ok;
closure_worker(#state{}=State, QSLst, Mult, Tab, Group) ->
    NewEntities = 
        (process_query_seq(State#state{entitynumber=-1}, QSLst))#state.res,
    ReallyNewEntities =
        [begin
             ets:insert(Tab, {Group, NEntity}),
             NEntity
         end || NEntity <- NewEntities, not ets:member(Tab, NEntity)],
    NewMult = case Mult of infinite = I -> I; M -> M-1 end,
    closure_worker(State#state{res=ReallyNewEntities}, QSLst, NewMult, Tab, Group).

grouping(NewState, #state{groupby_place=GroupBy, entitynumber=EntNum}=PrevState) ->
    if
        EntNum == GroupBy ->
            NewState#state{
                groupby_type = PrevState#state.type,
                groupby_res = PrevState#state.res,
                entitynumber = EntNum+1,
                groupby_place = GroupBy,
                params = PrevState#state.params};
        true ->
            NewState#state{
                groupby_type = PrevState#state.groupby_type,
                groupby_res = PrevState#state.groupby_res,
                entitynumber = EntNum+1,
                groupby_place = GroupBy,
                params = PrevState#state.params}
    end.

process_last(_Query, #state{res=[]}=State) -> State#state{groupby_res=[]};

process_last({set_op, {SetOp, Q1, Q2}}, #state{res=Entities}=State) ->
    {First, ToProc} = case Entities of [] -> {[],[]}; [H|T] -> {H,T} end,

    Proc1 = process_set_op_operand(State, Q1),
    Proc2 = process_set_op_operand(State, Q2),

    TypeQ1 = type_set_op_operand(State, Q1),
    TypeQ2 = type_set_op_operand(State, Q2),

    FirstQ1 = Proc1(First),
    FirstQ2 = Proc2(First),

    Merged = FirstQ1#state{
        res = [merge_results(SetOp, FirstQ1, FirstQ2)] ++
                pmap({?MODULE, pmap_helper},
                    [fun(E) -> merge_results(SetOp, Proc1(E), Proc2(E)) end],
                    ToProc)},

    Res = case TypeQ1 of
        TypeQ2 -> grouping(Merged, State);
        _ -> property_to_string(grouping(Merged, State))
    end,
    case FirstQ1#state.action of
        property_query ->
            Res#state{res = filter_empty(Res#state.res)};
        _ ->
            Res#state{res = filter_empty(filter_same_results(Res#state.res))}
    end;

process_last({selector, Selector, SelType}, St=#state{type=Type}) ->
    [Fun] = ?Lib:sel_fun(Type, Selector),
    Res = #state{
        action = selection,
        type = SelType,
        res = pmap({?MODULE, pmap_helper},
                [fun(E)->ordsets:from_list(lists:flatten(Fun(E))) end],
                St#state.res)},
    grouping(Res, St);

process_last({property, Prop, _PropType}, St=#state{type=Type}) ->
    [Fun]= ?Lib:prop_fun(Type, Prop),
    Res = #state{
        action = property_query,
        type = Prop,
        res = case St#state.groupby_type of 
            [] ->
                pmap({?MODULE, pmap_helper}, [Fun], St#state.res);
            _ ->
                pmap({?MODULE, pmap_helper},
                    [fun(G)->[Fun(E)||E<-G] end], St#state.res)
        end},
    grouping(Res, St);

process_last({Action, QSLst, Mult}, #state{type=Type, res=Res}=State) ->
    InitialChains =
        case State#state.action of
            selection -> #chains{incomplete=[[Entity]|| Entity<-lists:flatten(Res)]};
            Action -> Res
        end,
    Result = #state{
        action = Action,
        type = Type, 
        res = chains(State#state{action=Action}, InitialChains, QSLst, Mult)},
    grouping(Result, State);

process_last({filter, Filt}, #state{action=iteration, res=Res}=St) ->
    NewIncomplete = lists:filter(
                      fun(Chain) ->
                                 filter(Filt, St#state{res= [hd(Chain)]}) /= [] 
                       end,
                      Res#chains.incomplete),

    St#state{res = Res#chains{incomplete = NewIncomplete}};

process_last({filter, Filt}, #state{action=closure, res=Res}=St) ->
    NewComplete = filter_chain(St, Filt, Res#chains.complete),
    NewIncomplete = filter_chain(St, Filt, Res#chains.incomplete),
    NewRecursive = filter_chain(St, Filt, Res#chains.recursive),

    NewChains = Res#chains{complete   = NewComplete,
                           incomplete = NewIncomplete,
                           recursive  = NewRecursive},

    St#state{res = NewChains};

process_last({filter, Filt},St=#state{groupby_res=[]}=St) ->
    St#state{res=filter(Filt, St)};

process_last({filter, Filt}, St=#state{res=Res}=St) ->
    St#state{
        res = pmap({?MODULE, pmap_helper},
                [fun(E)->filter(Filt, St#state{res=E}) end], Res)};

% todo: preproc
process_last({statistics, Stat}, #state{res=PropValues}) ->
    NonNum = lists:filter(fun(Val) -> not is_number(Val) end, PropValues),

    ?Check(NonNum == [],
           ?LocalError(type_mismatch, [{statistics,Stat}, hd(NonNum), number])),

    Fun = ?Lib:stat_fun(Stat),
    #state{action = statistics, type = Stat, res = Fun(PropValues)}.

%%% ============================================================================
%%% Helper functions
pmap({Module, Func}, ExtraArgs, List) ->
    % rpc:pmap({Module, Func}, ExtraArgs, List). %slower than ?MISC:parallelise
     ?MISC:parallelise(List, Module, Func,
                       fun(Entities)-> [Entities|ExtraArgs] end, false).

pmap_helper(Entities, Fun) when is_list(Entities) ->
     [Fun(Entity) || Entity <- Entities];
pmap_helper(Entity, Fun)->
     pmap_helper([Entity], Fun).

chains_worker(_State, [] = _Chains, _QSLst, _Mult, _Tab) ->
    ok;
chains_worker(_State, Chains, _QSLst, 0, Tab) ->
    ets:insert(Tab, [{incomplete, Chain} || Chain <- Chains]),
    ok;
chains_worker(#state{type=Type}=St,Chains, QSLst, Mult, Tab) ->
    NewChains =
        lists:flatmap(
            fun([H|_] = Chain) ->
                #state{type=Type, res=Res} =
                    process_query_seq(St#state{res=[H]}, QSLst),
                next_chain(St#state{res=lists:flatten(Res)}, Chain)
            end,
            Chains),
    SplitFun =
        fun({complete, _} = C, {Incomp, Comp, Rec}) ->
                {Incomp, [C | Comp], Rec};
           ({recursive, _} = R, {Incomp, Comp, Rec}) ->
                {Incomp, Comp, [R | Rec]};
           (List, {Incomp, Comp, Rec}) ->
                {[List | Incomp], Comp, Rec}
        end,

    {Incomplete, Completed, Recursive} =
        lists:foldl(SplitFun, {[], [], []}, NewChains),

    ets:insert(Tab, Completed ++ Recursive),

    NewMult = case Mult of infinite -> infinite; Mult -> Mult-1 end,
    chains_worker(St,Incomplete, QSLst, NewMult, Tab).

chains(_State, #chains{incomplete = []} = Chains, _QSLst, _Mult) ->
    Chains;
chains(State, #chains{incomplete = Chains} = ChainsRec, QSLst, Mult) ->
    Tab  = ets:new(store, [public, bag, {write_concurrency, true}]),
    ChainGroups = group_entities(Chains),
    Keys = [rpc:async_call(node(), ?MODULE, chains_worker,
                           [State, Ch, QSLst, Mult, Tab])
            || Ch <- ChainGroups],
    [ ok = rpc:yield(K) || K <- Keys],
    Recursive =
        ets:select(Tab, [{{'$1', '$2'},[{'==', '$1', recursive}], ['$2']}]),
    Completed =
        ets:select(Tab, [{{'$1', '$2'},[{'==', '$1', complete}], ['$2']}]),
    Incomplete =
        ets:select(Tab, [{{'$1', '$2'},[{'==', '$1', incomplete}], ['$2']}]),
    ets:delete(Tab),
    ChainsRec#chains{complete   = Completed ++ ChainsRec#chains.complete,
                     incomplete = Incomplete,
                     recursive  =  Recursive ++ ChainsRec#chains.recursive}.

next_chain(#state{action=iteration, res = Result}, Chain) ->
    lists:foldl(fun(Entity, Acc) -> [[Entity| Chain]|Acc] end, [], Result);
next_chain(#state{action=closure, res=[]}, Chain) ->
    [{complete, Chain}];
next_chain(#state{action=closure, res=Result}, Chain) ->
    lists:foldl(
      fun(Entity, Acc) ->
              case lists:member(Entity, Chain) of
                  true -> [{recursive, [Entity|Chain]}|Acc];
                  _    -> [[Entity| Chain]|Acc]
              end
      end,
      [],
      Result).

filter_chain(State, Filter, Chain) ->
    NewChainWithEmpties = [ filter(Filter, State#state{res=List}) ||
                            List <- Chain],
    lists:filter( fun(List) -> List /= [] end, NewChainWithEmpties).

%%% ============================================================================
%%% Filters

%% TODO: entity listakat rendezni filter elott!!!!
%% @private
%% @spec filter(Filter::term(), #state{}) -> ordset()
filter(_Filter, #state{res=[]}) -> [];

filter('true', #state{res=Entities}) -> Entities;

filter('false', _State) -> [];

filter({'not', Filter}, #state{res=Entities} = State) ->
    ordsets:subtract(Entities, filter(Filter, State));

filter({'or', Filter1, Filter2}, State) ->
    ordsets:union(filter(Filter1, State),
                  filter(Filter2, State));

filter({'and', Filter1, Filter2}, #state{type=EntityType} = State) ->
    filter(Filter2,#state{type= EntityType, res = filter(Filter1, State)});

filter({set_op, {SetOp, Q1, Q2}}, State) ->
    Get_val =
        fun({cons, L}) ->
            List = ordsets:from_list(all_to_string(L)),
            fun(_) -> List end;
        ({seq, S}) ->
            fun(E) ->
                Seq = process_query_seq(
                        State#state{
                            res = [E],
                            entitynumber = -length(S)
                        }, S),
                Conv = property_to_string(Seq),
                ordsets:from_list(Conv#state.res)
            end;
        ({'query', SemQuery}) ->
            SemanticQuery = process_semantic_query(State#state.params, SemQuery, []),
            Conv = property_to_string(SemanticQuery),
            Res = 
            case is_group_state(SemanticQuery) of
                false -> ordsets:from_list(Conv#state.res);
                _  -> ordsets:from_list(lists:append(Conv#state.res))
            end,
            fun(_) -> Res end;
        (Property) when is_atom(Property) ->
            PropFun = ?Lib:prop_fun(State#state.type, Property),
            case PropFun of
                [] -> fun(_) -> [to_string(Property)] end;
                [Fun] -> fun(E) -> [to_string(Fun(E))] end
            end
    end,
    Op1 = Get_val(Q1),
    Op2 = Get_val(Q2),

    lists:filter(
        fun(E) ->%?d({Op1(E), Op2(E)}),
            case SetOp of
                any_in -> not( ordsets:is_disjoint(Op1(E), Op2(E)) );
                all_in -> ordsets:is_subset(Op1(E), Op2(E))
            end
        end,
        State#state.res);

filter({seq, QuerySeq}, #state{res=Entities,entitynumber=EntNum} = State) ->
    ordsets:filter(
      fun(Entity) ->
              St = process_query_seq(
                    State#state{
                        res = [Entity],
                        entitynumber=EntNum-length(QuerySeq)}, QuerySeq),
                St#state.res /= []
      end,
      Entities);

%% Comparison works on atom, int and string.
%% todo: regexp match -> if a string doesn't match
filter({CompOp, Filt1, Filt2}, #state{type=EntityType,res=Entities} = State) ->
    Comp1 = case Filt1 of
                {quoted, A} -> A;
                _ when is_tuple(Filt1) -> filter(Filt1, State);
                _  -> Filt1
            end,

    Comp2 = case Filt2 of
                {quoted, B} -> B;
                _ when is_tuple(Filt2) -> filter(Filt2, State);
                _ -> Filt2
            end,

    PropFun1 = ?Lib:prop_fun(EntityType, Comp1),
    PropFun2 = ?Lib:prop_fun(EntityType, Comp2),

    ordsets:filter(
      fun(Entity) ->
              CompL = case PropFun1 of 
                                    [] -> Comp1; 
                                    [Fun1] ->Fun1(Entity) 
                      end,
              CompR = case PropFun2 of 
                                    [] -> Comp2; 
                                    [Fun2] -> Fun2(Entity) 
                      end,
                case type_of(CompL)==type_of(CompR) of
                    true ->
                        compare(CompOp, CompL, CompR);
                    _ -> compare(CompOp, to_string(CompL), to_string(CompR))
                end
      end,
     Entities);

filter(Filter, #state{type=EntityType, res=Entities})->
    ordsets:filter(prop_fun(EntityType, Filter), Entities).


%%% ============================================================================
%%% Helper functions

compare(like, CompL, CompR) ->
    Distance =
        if
            is_list(CompL) andalso is_list(CompR) ->
                refusr_strm:getDistance(CompL, CompR);
            is_atom(CompL) andalso is_atom(CompR) ->
                refusr_strm:getDistance(atom_to_list(CompL),
                                        atom_to_list(CompR));
            true -> false
        end,

    case Distance of
         {lev, 0} -> true;
         {lev, 1} -> true;
         {lev, 2} ->
            Length = if is_atom(CompL) -> length(atom_to_list(CompL));
                        true           -> length(CompL)
                     end,
            case Length > 5 of true -> true; _ -> false end;
        _ -> false
    end;

compare('~', CompL, CompR) when is_atom(CompR)->
    compare('~', CompL, atom_to_list(CompR));
compare('~', CompL, CompR) when is_atom(CompL)->
    compare('~', atom_to_list(CompL), CompR);
compare('~', CompL, {quoted, A}) when is_atom(A)->
    compare('~', CompL, atom_to_list(A));
compare('~', {quoted, A}, CompR) when is_atom(A)->
    compare('~', atom_to_list(A), CompR);
compare('~', CompL, CompR) ->
    RegExp = case re:compile(CompR) of
        {ok, MP} -> MP;
        {error, _} -> throw(?LocalErr0r(bad_regexp))
    end,
    case re:run(CompL, RegExp) of
        {match, _} -> true;
        nomatch -> false
    end;


compare(CompOp, {quoted, A}, CompR) ->
    erlang:CompOp(A, CompR);

compare(CompOp, CompL, {quoted, A}) ->
    erlang:CompOp(CompL, A);

compare(CompOp, CompL, CompR) ->
    erlang:CompOp(CompL, CompR).

prop_fun(EntityType, Filter) ->
    case ?Lib:prop_fun(EntityType, Filter) of
        [] -> throw(?LocalError(illegal_property, [EntityType, Filter]));
        [Fun]   -> Fun
    end.

%% @spec format_nodes(Nodes::[entity()], Positions::atom()) -> string()
%%       Positions = none|scalar|linecol
%% @doc Returns a textual representation for a list of nodes.
format_nodes(Nodes, Position) -> ?Format:nodes(Nodes, Position).
