%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

%%% @author Dniel Horpcsi <daniel-h@elte.hu>

-module(erltex).
-export([tokens/1]).
-export([expr/1]).

-define(SKIP, ++ [$\\ , $;] ++).
-define(KW(X), "\\textbf{" ++ X ++ "}").

%% scan_and_parse(X) ->
%%     S = case lists:last(X) of
%%             $. -> ok;
%%             _  -> X ++ "."
%%         end,
%%     {ok, Ts, _} = erl_scan:string(S),
%% %    io:format("~p~n", [Ts]),
%%     {ok, Tree } = erl_parse:parse_exprs(Ts),
%% %    io:format("~p~n", [Tree]),
%%     Tree.

tokens(Ts) ->
    %%io:format("~p~n", [Ts]),
    {ok, Tree} = erl_parser:parse_exprs(Ts ++ [{dot, 0}]),
    %%io:format("~p~n", [Tree]),
    translate(Tree).

translate(Trees) when is_list(Trees) ->
    [translate(T) || T <- Trees];
translate(Tree) ->
    try form(Tree)
    catch _:_ ->
            expr(Tree)
    end.

form({attribute,_Line,_Attr,_Val}) -> [];
form({function,Line,Name0,Arity0,Clauses0}) ->
    {Name,Arity,Clauses} = function(Name0, Arity0, Clauses0),
    {function,Line,Name,Arity,Clauses};
form({error,_E}) -> [];
form({warning,_W}) -> [];
form({eof,_Line}) -> [].

function(Name, Arity, Clauses0) ->
    Clauses1 = clauses(Clauses0),
    {Name,Arity,Clauses1}.

clauses([C0|Cs]) ->
    C1 = clause(C0),
    [C1|clauses(Cs)];
clauses([]) -> [].

clause({clause,_,H0,[],B0}) ->
    H1 = head(H0),
    B1 = exprs(B0),
    "(" ++ H1 ++ ")" ?SKIP "\\rightarrow" ?SKIP B1;
clause({clause,_,H0,G0,B0}) ->
    H1 = head(H0),
    G1 = guard(G0),
    B1 = exprs(B0),
    "(" ++ H1 ++ ")" ?SKIP "when" ?SKIP G1 ?SKIP "\\rightarrow" ?SKIP B1.

head(Ps) -> patterns(Ps).

patterns([P0|Ps]) ->
    P1 = pattern(P0),
    [P1|patterns(Ps)];
patterns([]) -> [].

pattern(X) -> expr(X).

pattern_grp([{bin_element,L1,E1,S1,T1} | Fs]) ->
    S2 = case S1 of
	     default ->
		 default;
	     _ ->
		 expr(S1)
	 end,
    T2 = case T1 of
	     default ->
		 default;
	     _ ->
		 bit_types(T1)
	 end,
    [{bin_element,L1,expr(E1),S2,T2} | pattern_grp(Fs)];
pattern_grp([]) ->
    [].

bit_types([]) ->
    [];
bit_types([Atom | Rest]) when is_atom(Atom) ->
    [Atom | bit_types(Rest)];
bit_types([{Atom, Integer} | Rest]) when is_atom(Atom), is_integer(Integer) ->
    [{Atom, Integer} | bit_types(Rest)].



%% -type pattern_list([Pattern]) -> [Pattern].
%%  These patterns are processed "in parallel" for purposes of variable
%%  definition etc.

pattern_list([P0|Ps]) ->
    P1 = pattern(P0),
    [P1|pattern_list(Ps)];
pattern_list([]) -> [].

%% -type pattern_fields([Field]) -> [Field].
%%  N.B. Field names are full expressions here but only atoms are allowed
%%  by the *linter*!.

pattern_fields([{record_field,Lf,{atom,La,F},P0}|Pfs]) ->
    P1 = pattern(P0),
    [{record_field,Lf,{atom,La,F},P1}|pattern_fields(Pfs)];
pattern_fields([{record_field,Lf,{var,La,'_'},P0}|Pfs]) ->
    P1 = pattern(P0),
    [{record_field,Lf,{var,La,'_'},P1}|pattern_fields(Pfs)];
pattern_fields([]) -> [].

%% -type guard([GuardTest]) -> [GuardTest].

guard([G0|Gs]) when is_list(G0) ->
    [guard0(G0) | guard(Gs)];
guard(L) ->
    guard0(L).

guard0([G0|Gs]) ->
    G1 =  guard_test(G0),
    [G1|guard0(Gs)];
guard0([]) -> [].

guard_test(Expr={call,Line,{atom,La,F},As0}) ->
    case erl_internal:type_test(F, length(As0)) of
	true -> 
	    As1 = gexpr_list(As0),
	    {call,Line,{atom,La,F},As1};
	_ ->
	    gexpr(Expr)
    end;
guard_test(Any) ->
    gexpr(Any).

%% Before R9, there were special rules regarding the expressions on
%% top level in guards. Those limitations are now lifted - therefore
%% there is no need for a special clause for the toplevel expressions.
%% -type gexpr(GuardExpr) -> GuardExpr.

gexpr({call,Line,{atom,La,F},As0}) ->
    case erl_internal:guard_bif(F, length(As0)) of
	true -> As1 = gexpr_list(As0),
		{call,Line,{atom,La,F},As1}
    end;
% Guard bif's can be remote, but only in the module erlang...
gexpr({call,Line,{remote,La,{atom,Lb,erlang},{atom,Lc,F}},As0}) ->
    case erl_internal:guard_bif(F, length(As0)) or
	 erl_internal:arith_op(F, length(As0)) or 
	 erl_internal:comp_op(F, length(As0)) or
	 erl_internal:bool_op(F, length(As0)) of
	true -> As1 = gexpr_list(As0),
		{call,Line,{remote,La,{atom,Lb,erlang},{atom,Lc,F}},As1}
    end;
gexpr({bin,Line,Fs}) ->
    Fs2 = pattern_grp(Fs),
    {bin,Line,Fs2};
gexpr({op,Line,Op,A0}) ->
    case erl_internal:arith_op(Op, 1) or 
	 erl_internal:bool_op(Op, 1) of
	true -> A1 = gexpr(A0),
		{op,Line,Op,A1}
    end;
gexpr({op,Line,Op,L0,R0}) when Op =:= 'andalso'; Op =:= 'orelse' ->
    %% R11B: andalso/orelse are now allowed in guards.
    L1 = gexpr(L0),
    R1 = gexpr(R0),			%They see the same variables
    {op,Line,Op,L1,R1};
gexpr({op,Line,Op,L0,R0}) ->
    case erl_internal:arith_op(Op, 2) or
	  erl_internal:bool_op(Op, 2) or 
	  erl_internal:comp_op(Op, 2) of
	true ->
	    L1 = gexpr(L0),
	    R1 = gexpr(R0),			%They see the same variables
	    {op,Line,Op,L1,R1}
    end;
gexpr(X) -> expr(X).


%% -type gexpr_list([GuardExpr]) -> [GuardExpr].
%%  These expressions are processed "in parallel" for purposes of variable
%%  definition etc.

gexpr_list([E0|Es]) ->
    E1 = gexpr(E0),
    [E1|gexpr_list(Es)];
gexpr_list([]) -> [].

grecord_inits([{record_field,Lf,{atom,La,F},Val0}|Is]) ->
    Val1 = gexpr(Val0),
    [{record_field,Lf,{atom,La,F},Val1}|grecord_inits(Is)];
grecord_inits([{record_field,Lf,{var,La,'_'},Val0}|Is]) ->
    Val1 = gexpr(Val0),
    [{record_field,Lf,{var,La,'_'},Val1}|grecord_inits(Is)];
grecord_inits([]) -> [].

%% -type exprs([Expression]) -> [Expression].
%%  These expressions are processed "sequentially" for purposes of variable
%%  definition etc.

exprs([E0|Es]) ->
    E1 = expr(E0),
    [E1|exprs(Es)];
exprs([]) -> [].

%% TODO: LATEX _

fix_underscores(S) -> re:replace(S, "_", "\\\\_", [global, {return, list}]).

expr({var,freshvar,V}) -> "\\freshvar{" ++ fix_underscores(atom_to_list(V)) ++ "}";
expr({var,_,V}) ->
    case atom_to_list(V) of
        [$M,$E,$T,$A,$L,$I,$S,$T|V2] -> "\\metalist{" ++ fix_underscores(V2) ++ "}";
        V2 -> "\\metavar{" ++ fix_underscores(V2) ++ "}"
    end;
expr({integer,_,I}) -> hd(io_lib:format("~p", [I]));
expr({float,_,F}) -> hd(io_lib:format("~p", [F]));
expr({atom,_,A}) -> "\\text{" ++ fix_underscores(hd(io_lib:format("~p", [A]))) ++ "}";
expr({string,_,S}) -> "\\text{" ++ hd(io_lib:format("~p", [S])) ++ "}";
expr({char,_,C}) -> hd(io_lib:format("~p", [C]));
expr({nil,_}) -> "[]";
expr({cons,_,H0,T0}) ->
    H1 = string:join(lists:map(fun expr/1, H0), ", "),
    T1 = expr(T0),
    case T1 of
        "[]" ->
            "[" ++ H1 ++ "]";
        _ ->
            "[" ++ H1 ++ "|" ++ T1 ++ "]"
    end;
expr({lc,_,E0,Qs0}) ->
    Qs1 = lc_bc_quals(Qs0),
    E1 = expr(E0),
    "[" ?SKIP E1 ?SKIP "||" ?SKIP Qs1 ?SKIP "]";
expr({bc,_,E0,Qs0}) ->
    Qs1 = lc_bc_quals(Qs0),
    E1 = expr(E0),
    "<<" ?SKIP E1 ?SKIP "||" ?SKIP Qs1 ?SKIP ">>";
expr({tuple,_,Es0}) ->
    Es1 = expr_list(Es0),
    "\\{" ++ Es1 ++ "\\}";
%% expr({record_index,Line,Name,Field0}) ->
%%     Field1 = expr(Field0),
%%     {record_index,Line,Name,Field1};
%% expr({record,Line,Name,Inits0}) ->
%%     Inits1 = record_inits(Inits0),
%%     {record,Line,Name,Inits1};
%% expr({record_field,Line,Rec0,Name,Field0}) ->
%%     Rec1 = expr(Rec0),
%%     Field1 = expr(Field0),
%%     {record_field,Line,Rec1,Name,Field1};
%% expr({record,Line,Rec0,Name,Upds0}) ->
%%     Rec1 = expr(Rec0),
%%     Upds1 = record_updates(Upds0),
%%     {record,Line,Rec1,Name,Upds1};
%% expr({record_field,Line,Rec0,Field0}) ->
%%     Rec1 = expr(Rec0),
%%     Field1 = expr(Field0),
%%     {record_field,Line,Rec1,Field1};
expr({block,_,Es0}) ->
    Es1 = expr_list(Es0),
    ?KW("begin") ?SKIP Es1 ?SKIP ?KW("end");
expr({'if',_,Cs0}) ->
    Cs1 = icr_clauses(Cs0),
    ?KW("if") ?SKIP Cs1 ?SKIP ?KW("end");
expr({'case',_,E0,Cs0}) ->
    E1 = expr(E0),
    Cs1 = icr_clauses(Cs0),
    ?KW("case") ?SKIP E1 ?SKIP ?KW("of") ?SKIP Cs1 ?SKIP ?KW("end");
expr({'receive',_,Cs0}) ->
    Cs1 = icr_clauses(Cs0),
    ?KW("receive") ?SKIP Cs1 ?SKIP ?KW("end");
expr({'receive',_,Cs0,To0,ToEs0}) ->
    To1 = expr(To0),
    ToEs1 = exprs(ToEs0),
    Cs1 = icr_clauses(Cs0),
    ?KW("receive") ?SKIP Cs1 ?SKIP ?KW("after") ?SKIP To1 ?SKIP "\\rightarrow" ?SKIP ToEs1 ?SKIP ?KW("end");
expr({'try',_,Es0,Scs0,Ccs0,[]}) ->
    Es1 = exprs(Es0),
    Scs1 = icr_clauses(Scs0),
    Ccs1 = icr_clauses(Ccs0),
    ?KW("try") ?SKIP Es1 ?SKIP ?KW("of") ?SKIP Scs1 ?SKIP ?KW("catch") ?SKIP Ccs1 ?SKIP ?KW("end");
expr({'try',_,Es0,Scs0,Ccs0,As0}) ->
    Es1 = exprs(Es0),
    Scs1 = icr_clauses(Scs0),
    Ccs1 = icr_clauses(Ccs0),
    As1 = exprs(As0),
    ?KW("try") ?SKIP Es1 ?SKIP ?KW("of") ?SKIP Scs1 ?SKIP ?KW("catch") ?SKIP Ccs1 ?SKIP ?KW("after") ?SKIP As1 ?SKIP ?KW("end");
expr({'fun',_,Body}) ->
    case Body of
	{clauses,Cs0} ->
	    Cs1 = fun_clauses(Cs0),
            ?KW("fun") ?SKIP Cs1 ?SKIP ?KW("end");
	{function,F,A} ->
            ?KW("fun") ?SKIP expr(F) ++ "/" ++ expr(A);
	{function,M,F,A} when is_atom(M), is_atom(F), is_integer(A) ->
            ?KW("fun") ?SKIP expr({atom, 1, M}) ++ ":" ++ expr({atom, 1, F}) ++ "/" ++ expr({integer, 1, A});
	{function,M0,F0,A0} ->
	    %% R15: fun M:F/A with variables.
	    M = expr(M0),
	    F = expr(F0),
	    A = expr(A0),
            ?KW("fun") ?SKIP M ++ ":" ++ F ++ "/" ++ A
    end;
expr({call,_,F0,As0}) ->
    F1 = expr(F0),
    As1 = expr_list(As0),
    F1 ++ "(" ++ As1 ++ ")";
expr({'catch',_,E0}) ->
    E1 = expr(E0),
    ?KW("catch") ?SKIP E1;
expr({match,_,P0,E0}) ->
    E1 = expr(E0),
    P1 = pattern(P0),
    P1 ++ "=" ++ E1;
%expr({bin,Line,Fs}) ->
%    Fs2 = pattern_grp(Fs),
%    {bin,Line,Fs2};
expr({op,_,Op,A0}) ->
    A1 = expr(A0),
    atom_to_list(Op) ++ A1;
expr({op,_,Op,L0,R0}) ->
    L1 = expr(L0),
    R1 = expr(R0),
    L1 ++ atom_to_list(Op) ++ R1;
expr({remote,_,M0,F0}) ->
    M1 = expr(M0),
    F1 = expr(F0),
    M1 ++ ":" ++ F1.

expr_list(L) ->
    string:join([expr(X) || X <- L], ", ").
icr_clauses(L) ->
    Ss = [clause(X) || X <- L],
%%    Length = lists:sum(lists:map(fun length/1, Ss)),
%%    if Length > 10 ->
%%            "\\\\ &&&" ++ string:join(Ss, "; \\\\ &&& ");
%%       true ->
    string:join(Ss, "; ").
%%    end.
fun_clauses(L) ->
    string:join([clause(X) || X <- L], "; ").
lc_bc_quals(L) ->
    string:join([lc_bc_qual(X) || X <- L], ", ").

lc_bc_qual({generate,_,P0,E0}) ->
    E1 = expr(E0),
    P1 = pattern(P0),
    P1 ++ " \\leftarrow " ++ E1;
lc_bc_qual({b_generate,_,P0,E0}) ->
    E1 = expr(E0),
    P1 = pattern(P0),
    P1 ++ " \\leftarrow " ++ E1;
lc_bc_qual(E0) ->
    expr(E0).
