%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

%%% @author Dniel Horpcsi <daniel-h@elte.hu>

-module(rule_compiler).

-export([file/1, file/2, format_error/1]).
-export([string/1, string/2]).

-import(lists, [concat/1, delete/2, flatmap/2, foldl/3, foldr/3,
                foreach/2, keydelete/3, keysearch/3, map/2, member/2,
                sort/1, usort/1]).

-include("rules.hrl").

-define(ETSTAB,            egg_data).
-define(ASSUMED_EXTENSION, ".refac").
-define(S,                 erl_syntax).

-define(STEP(P), {P, fun P/0}).

%%% ============================================================================
%%% Interface

%% Running with the default options (reporting errors and warnings).

file(File) -> file(File, [report_errors, report_warnings]).
string(File) -> string(File, []).

%% Options:
%% - out:             specifying the name of the resulting Erlang file
%% - report_errors:   report all errors occuring during the process
%%   report_warnings: report all warnings occuring during the process
%% - return_errors:   return errors as Erlang terms
%%   return_warnings: return warnings as Erlang terms
%% - time:            enable timing of the process steps
%% - verbose:         (currently no functionality)
%% - return:          return_errors + return_warnings
%%   report:          report_errors + report_warnings
%% - compile:         compile the resulting Erlang module

file(File, Options) ->
    case is_filename(File) of
        no -> erlang:error(badarg, [File, Options]);
        _  -> ok
    end,
    ets:new(?ETSTAB, [public, bag, named_table]),
    try set_options(Options)
    catch
        throw:badarg -> erlang:error(badarg, [File, Options])
    end,
    Self = self(),
    Pid = spawn_link(fun() -> main(Self, File) end),
    receive
        {Pid, Rep} ->
            receive after 1 -> ok end,
            process_flag(trap_exit, process_flag(trap_exit, false)),
            ets:delete(?ETSTAB),
            Rep
    end.

string(String, Options) ->
    ets:new(?ETSTAB, [public, bag, named_table]),
    try set_options(Options)
    catch
        throw:badarg -> erlang:error(badarg, [Options])
    end,
    Self = self(),
    Pid = spawn_link(fun() -> main3(Self, String) end),
    receive
        {Pid, Rep} ->
            receive after 1 -> ok end,
            process_flag(trap_exit, process_flag(trap_exit, false)),
            ets:delete(?ETSTAB),
            Rep
    end.

%% -----------------------------------------------------------------------------
%% Error formatting

format_error(E) ->
    io_lib:fwrite("Unexpected error: ~p", [E]).

%%% ============================================================================

%% -----------------------------------------------------------------------------
%% Option parsing

-define(Options, [out, report_errors, report_warnings, return_errors,
                  return_warnings, time, verbose, return, report, compile,

                  return_ast]).

set_options(Opts) when is_list(Opts) ->
    validate_options(proplists:get_keys(Opts)),
    [option(Key, Opts) || Key <- ?Options].

validate_options(Keys) ->
    Valid =  fun(K) ->  lists:member(K, ?Options) end,
    case lists:all(Valid, Keys) of
        true  -> ok;
        false -> throw(badarg)
    end.

option(report, Opts) ->
    case proplists:is_defined(report, Opts) of
        true ->
            Val = proplists:get_bool(report, Opts),
            option(report_errors,   Val),
            option(report_warnings, Val);
        false ->
            ok
    end;
option(return, Opts) ->
    case proplists:is_defined(return, Opts) of
        true ->
            Val = proplists:get_bool(return, Opts),
            option(return_errors,   Val),
            option(return_warnings, Val);
        false ->
            ok
    end;
option(out, Opts) ->
    Out = proplists:get_value(out, Opts, default(out)),
    case proplists:is_defined(return, Opts) of
        true ->
            case is_filename(Out) of
                no       -> throw(badarg);
                Filename -> option(out, Filename)
            end;
        false ->
            set_data(out, Out)
    end;
option(Key, Opts) when is_list(Opts) ->
    set_data(Key, proplists:get_value(Key, Opts, default(Key)));
option(Key, Value) ->
    set_data(Key, Value).

default(out)             -> "";
default(report_errors)   -> true;
default(report_warnings) -> true;
default(return_errors)   -> false;
default(return_warnings) -> false;
default(time)            -> false;
default(verbose)         -> false;
default(compile)         -> false;
default(_)               -> false.

%% -----------------------------------------------------------------------------

main(Parent, File) ->
    case get_data(out) of
        {stdout, Rulename} ->
            set_data(out, Rulename),
            main2(Parent, File);
        _ ->
            main1(Parent, File)
    end.

main1(Parent, File) ->
    prepare_filenames(File),
    try prepare_input(),
        try prepare_output(),
            try generate()
            after
                file:close(get_data(outport))
            end
        after
            file:close(get_data(inport))
        end,
        finish()
    catch
        throw:step_failed ->
            ':(';
        throw:{input_error, Reason} ->
            add_error(get_string(infile), none, {file_error, Reason});
        throw:{output_error, Reason} ->
            add_error(get_string(outfile), none, {file_error, Reason}),
            catch file:delete(get_data(outfile));
        exit:Reason ->
            add_error({internal_error, Reason}),
            catch file:delete(get_data(outfile))
    end,
    Parent ! {self(), summary()}.

main2(Parent, File) ->
    prepare_filenames(File),
    try prepare_input(),
         set_data(outport, standard_io),
         try generate()
         after
             file:close(get_data(inport))
         end,
         finish()
    catch
        throw:step_failed ->
            ':(';
        throw:{input_error, Reason} ->
            add_error(get_string(infile), none, {file_error, Reason});
        exit:Reason ->
            add_error({internal_error, Reason}),
            catch file:delete(get_data(outfile))
    end,
    Parent ! {self(), summary()}.

main3(Parent, String) ->
    try
        prepare_input(String),
        set_data(outport, standard_io),
        generate(),
        finish()
    catch
        throw:step_failed ->
            ':(';
        exit:Reason ->
            add_error({internal_error, Reason}),
            catch file:delete(get_data(outfile))
    end,
    Parent ! {self(), summary()}.

prepare_filenames(File) ->
    Infile = assure_extension(File, ?ASSUMED_EXTENSION),
    Out = get_data(out),
    OutName = case Out of
                   [] -> filename:rootname(Infile, ?ASSUMED_EXTENSION);
                   _  -> Out
               end,
    Outfile = assure_extension(OutName, ".tex"),
    Module = list_to_atom(filename:basename(Outfile, ".tex")),
    set_data(module,  Module),
    set_data(infile,  Infile),
    set_data(outfile, Outfile).

prepare_input(String) when is_list(String) ->
    set_data(inport, String).

prepare_input() ->
    case file:open(get_string(infile), [read, read_ahead]) of
        {ok, Inport}    -> set_data(inport, Inport);
        {error, Reason} -> throw({input_error, Reason})
    end.

prepare_output() ->
    case file:open(get_string(outfile), [write, delayed_write]) of
        {ok, Outport}   -> set_data(outport, Outport);
        {error, Reason} -> throw({output_error, Reason})
    end.

finish() ->
    case get_datalist(errors) of
        [] -> ok;
        _ ->  file:delete(get_string(outfile))
    end.

timing_fun(Pre) ->
    case get_data(time) of
        true  -> Pre(), fun timeit/2;
        false -> fun(_Name, Fn) -> Fn() end
    end.

run_step(F, Name, Fun) ->
    F(Name, Fun),
    case get_datalist(errors) of
        [] -> ok;
        _  -> throw(step_failed)
    end.

generate() ->
    Pre = fun() ->
                  io:fwrite(<<"Generating output... ~s\n">>,
                            [format_filename(get_string(infile))])
          end,
    TF = timing_fun(Pre),
    Ps =  [?STEP(parse_rules), ?STEP(write_file)],
    [run_step(TF, Name, Fun) || {Name, Fun} <-Ps].

timeit(Name, Fun) ->
    {Before, _} = statistics(runtime),
    Fun(),
    {After, _} = statistics(runtime),
    Mem0 = ets:info(?ETSTAB, memory)*erlang:system_info(wordsize),
    Mem = lists:flatten(io_lib:format("~.1f kB", [Mem0/1024])),
    io:fwrite(" ~-20w- ~11.2f s ~11s\n", [Name, (After-Before)/1000, Mem]).

%% -----------------------------------------------------------------------------
%% Scanning and parsing the rules

parse_rules() ->
    store_rules(read_rules(get_data(inport))).

store_rules({error,ErrorLine,Error}) ->
    add_error(ErrorLine, Error);
store_rules({Directives, Rules}) ->
    add_datas(directives, Directives),
    add_datas(rules, Rules).

read_rules(Inport) ->
    Tokens = rule_scanner:scan(Inport),
    case rule_parser:parse(Tokens) of
        {error, {ErrorLine, Mod, Message}} ->
            {error, ErrorLine, {error, Mod, Message}};
        {ok, {rules, Rules}} ->
            {rules, Rules};
        {ok, {Directives, Rules}} ->
            {Directives, Rules};
        {ok, Ss} ->
            Ss
    end.

%% -----------------------------------------------------------------------------
%% Writing the output file

write_file() ->
    Rules = get_datalist(rules),
    write_rules(Rules),
    nl().

write_rules([]) -> ok;
write_rules([R|Rs]) ->
    S = format_rule(R),
    S /= [] andalso begin fwrite(S), nl(),nl() end,
    write_rules(Rs).

format_rule(#rule{rewrite = [#rewrite{left = undefined}|_REST]}) ->
    []; %% TODO

format_rule(#rule{rewrite = [#rewrite{right = undefined}|_REST]}) ->
    []; %% TODO

format_rule(#rule{signature = {Name, _Ps}, rewrite = [#rewrite{left = L, right = R, conditions = Cs}|_REST]}) ->
    Var = fun({var, _, V}) -> {true, V};
                       (_) -> false
          end,
    V1 = lists:usort(lists:filtermap(Var, L)),
    V2 = lists:usort(lists:filtermap(Var, R)),
    X = lists:subtract(V2,V1),
    R2 = lists:map(fun({var, Ln, V}) ->
                           case lists:member(V, X) of
                               true -> {var, freshvar, V};
                               false -> {var, Ln, V}
                           end;
                      (Z) -> Z
                   end, R),
    set_data(freshvars, X),
    case get_data(outport) /= standard_io orelse element(3, Name) == get_data(out) of
        true ->
            "\\rrule{\\reduce{\n" ++ erltext(L) ++ "}{\n" ++ erltext(R2) ++ "}}\n{" ++ format({conditions, Cs}) ++ "}\n{" %% ++ format({signature, Name, Ps})
                ++ "}";
        false ->
            []
    end;

format_rule(_) ->
    []. %% TODO: rewrite this according to the new representation of rules

erltext(Ts) ->
    "\\begin{aligned}\n & " ++ string:join(erltex:tokens(Ts), ", \\\\\n & ") ++ "\n\\end{aligned}".

format({code, Ts}) -> string:join(erltex:tokens(Ts), ", ");
format({paren, X}) -> "(" ++ format(X) ++ ")";
format({conditions, []}) -> "";
format({conditions, Cs}) -> " & " ++ string:join(lists:map(fun format/1, Cs), ", \\\\\n & ");
format({predicate, P, As}) -> format(P) ++ "(" ++ string:join(lists:map(fun format/1, As), ", ") ++ ")";
format({app, {atom, _, intersect}, [A,B]}) -> format(A) ++ "\\cap" ++ format(B);
format({app, {atom, _, length}, [A]}) -> "|\\," ++ format(A) ++ "\\,|";
format({app, N, As}) -> format(N) ++ "(" ++ string:join(lists:map(fun format/1, As), ", ") ++ ")";
format({signature, Name, Ps}) -> format(Name) ++ "(" ++ string:join(lists:map(fun format/1, Ps), ", ") ++ ")";
format({'not', A}) -> "\\neg " ++ format(A);
format({'and', A, B}) -> format(A) ++ "\\land" ++ format(B);
format({'or', A, B}) -> format(A) ++ "\\lor" ++ format(B);
format({'implies', A, B}) -> format(A) ++ "\\implies" ++ format(B);
format({integer, L, I}) -> erltex:expr({integer, L, I});
format({atom, L, A}) -> erltex:expr({atom, L, A});
format({var, L, V}) ->
    case lists:member(V, get_data(freshvars)) of
        true -> erltex:expr({var, freshvar, V});
        false -> erltex:expr({var, L, V})
    end;
format({'=<', A, B}) -> format(A) ++ "\\leq" ++ format(B);
format({'>=', A, B}) -> format(A) ++ "\\geq" ++ format(B);
format({Op, A, B}) when Op == '+'; Op == '-'; Op == '*'; Op == '/';
                        Op == '='; Op == '<'; Op == '>'
                        -> format(A) ++ atom_to_list(Op) ++ format(B);
format(X) -> io:format("~n~n~n~n~n~p~n~n~n~n", [X]), "ALMA".


%%% ============================================================================
%%% Returning, errors

summary() ->
    report_errors(),
    report_warnings(),
    Es = pack_errors(get_datalist(errors)),
    Ws = pack_warnings(get_datalist(warnings)),
    if
        Es =:= [] ->
            case get_data(return_ast) of
                true ->
                    {get_datalist(directives), get_datalist(rules)};
                _ ->
                    case get_data(return_warnings) of
                        true  -> {ok, get_string(outfile), Ws};
                        false -> {ok, get_string(outfile)}
                    end
            end;
        true ->
            case get_data(return_errors) of
                true  -> {error, Es, Ws};
                false -> error
            end
    end.

pack_errors([{File,_} | _] = Es) ->
    [{File, flatmap(fun({_,E}) -> [E] end, sort(Es))}];
pack_errors([]) ->
    [].

pack_warnings([{File,_} | _] = Ws) ->
    [{File, flatmap(fun({_,W}) -> [W] end, sort(Ws))}];
pack_warnings([]) ->
    [].

report_errors() ->
    case get_data(report_errors) of
        true ->
            case get_datalist(errors) of
                [] -> ok;
                _ -> io:fwrite(<<"\n">>,[])
            end,
            foreach(fun({File,{none,Mod,E}}) ->
                            io:fwrite(<<"~s: ~s\n">>,
                                      [File,Mod:format_error(E)]);
                       ({File,{Line,Mod,E}}) ->
                            io:fwrite(<<"~s:~w: ~s\n">>,
                                      [File,Line,Mod:format_error(E)])
                    end, sort(get_datalist(errors)));
        false ->
            ok
    end.

report_warnings() ->
    case get_data(report_warnings) of
        true ->
            case get_datalist(warnings) of
                [] -> ok;
                _ -> io:fwrite(<<"\n">>,[])
            end,
            foreach(fun({File,{none,Mod,W}}) ->
                            io:fwrite(<<"~s: Warning: ~s\n">>,
                                      [File,Mod:format_error(W)]);
                       ({File,{Line,Mod,W}}) ->
                            io:fwrite(<<"~s:~w: Warning: ~s\n">>,
                                      [File,Line,Mod:format_error(W)])
                    end, sort(get_datalist(warnings)));
        false ->
            ok
    end.

add_error(E) ->
    add_error(none, E).

add_error(Line, E) ->
    add_error(get_string(infile), Line, E).

add_error(File, Line, E) ->
    add_data(errors, {File,{Line,?MODULE,E}}).

add_warning(Line, W) ->
    add_data(warnings, {get_string(infile),{Line,?MODULE,W}}).

%%% ============================================================================
%%% MISC

intersect(L1, L2) ->
    ordsets:to_list(ordsets:intersection(ordsets:from_list(L1),
                                         ordsets:from_list(L2))).

fwrite(Format) -> io:fwrite(get_data(outport), Format, []).
fwrite(Format, Args) -> io:fwrite(get_data(outport), Format, Args).
nl() -> xxx /= get_data(out) andalso io:nl(get_data(outport)).
format_filename(Filename) -> io_lib:write_string(filename:flatten(Filename)).

%% -----------------------------------------------------------------------------
%% ETS data storage

extract(X) -> case remove_label(X) of
                  {O} -> O;
                  O -> O
              end.
remove_label(Tuple) -> list_to_tuple(tl(tuple_to_list(Tuple))).
add_label(Label, Tuple) -> list_to_tuple([Label | tuple_to_list(Tuple)]).

get_datalist(T) -> [extract(E) || E <- ets:lookup(?ETSTAB, T)].
get_string(T) -> get_data(T).
get_data(T) ->
    case get_datalist(T) of
        []  -> none;
        [R] -> R;
        _   -> more
    end.

add_data(X) when is_tuple(X) orelse is_list(X) -> ets:insert(?ETSTAB, X).
add_data(T, X) when is_tuple(X) -> add_data(add_label(T, X));
add_data(T, X) -> add_data({T, X}).
add_datas(T, X) -> [add_data(T, E) || E <- X].

%% set_data(X) when is_tuple(X) -> ets:delete(?ETSTAB, element(1, X)), add_data(X).
set_data(T, X) -> ets:delete(?ETSTAB, T), add_data(T, X).
set_datalist(T, L) -> ets:delete(?ETSTAB, T), add_datas(T, L).
%% set_datalist(L = [E|_]) -> ets:delete(?ETSTAB, element(1, E)), add_data(L).

delete_data(X) when is_tuple(X) orelse is_list(X) -> ets:delete_object(?ETSTAB, X).
delete_data(T, X) when is_tuple(X) -> delete_data(add_label(T, X));
delete_data(T, X) -> delete_data({T, X}).
%% delete_datas(T, X) -> [delete_data(T, E) || E <- X].

%% =============================================================================
%% Misc

is_filename(T) ->
    try filename:flatten(T) of
        Filename -> Filename
    catch
        error:_ -> no
    end.

assure_extension(File, Ext) -> lists:concat([strip_extension(File, Ext), Ext]).

strip_extension(File, Ext) ->
    case filename:extension(File) of
        Ext    -> filename:rootname(File);
        _Other -> File
    end.
