%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

%%% @author Dniel Horpcsi <daniel-h@elte.hu>

-module(rule_interpreter).
-compile(export_all).

-include_lib("referl_core/src/core.hrl").
-include("rules.hrl").
-define(ETSTAB, kutyafule).

-define(BackupMetavars,
        Backup = ets:tab2list(metavars)
       ).
-define(ResetMetavars,
        ets:delete_all_objects(metavars),
        ets:insert(metavars, {make_ref(), []})
       ).
-define(RestoreMetavars,
        ets:delete_all_objects(metavars),
        [ets:insert(metavars, E) || E <- Backup]
       ).
-define(KeepOldMetavars,
        %% ?d({merge_matchings, Backup, ets:tab2list(metavars)}),
        Merged = merge_matchings(Backup, ets:tab2list(metavars)),
        ets:delete_all_objects(metavars),
        [ets:insert(metavars, E) || E <- Merged]
       ).

merge_matchings(X, []) -> X;
merge_matchings([], _) -> [];
merge_matchings([{_Id,Old}], [{Id2,New}]) ->
    Merged = lists:map(fun({Var,Val}) ->
                               case lists:keyfind(Var,1,New) of
                                   false -> {Var, Val};
                                   {Var, Val2} -> {Var, Val2}
                               end
                       end, Old),
    [{Id2, Merged}].

farm({'$gn', expr, _} = Expr) ->
    init({file, "farm"}),
    execute(introduce_farm, [{'Nw', 10}], Expr),
    finalize(),
    ok.

test(Name) -> test(Name, []).

test(Name, Args) ->
    init(),
    execute(Name, Args, guess(Name)),
    finalize(),
    ok.

guess(Name) ->
    case ?Graph:path(?Graph:root(), [module, {func, {name, '==', Name}}, {fundef, back}, funcl, body]) of
        []  -> undefined;
        [E] -> E;
        L   -> L
    end.

%% =============================================================================
%% INITIALIZATION, RULE STORE AND LOAD
%% =============================================================================

-define(RULEDIR, "./lib/rules/src/rules/").

init() ->
    put(already_copied, []),
    put(tobematched, []),
    put(tobechecked, []),
    ets:new(?ETSTAB, [public, set, named_table]).

init(D) ->
    init(),
    load_rules(D).

finalize() ->
    ets:delete(?ETSTAB),
    erase(),
    io:format("~n~n~nSuccessful test... file contents:~n~n"),
    io:format("~s~n", [?Syn:flat_text(hd(?Query:exec([file])))]).

%% -----------------------------------------------------------------------------

load_rules({file, F}) ->
    case rule_compiler:file(?RULEDIR++F, [return_ast, {out, {stdout, xxx}}]) of %% trick the rule compiler so that it doesn't produce tex output
        error -> ?Warning("Could not load and store the rules given in file ~p", [F]);
        D -> store_rules(D)
    end;
load_rules({string, S}) ->
    D = rule_compiler:string(S, [return_ast, {out, {stdout, xxx}}]), %% trick the rule compiler so that it doesn't produce tex output
    store_rules(D).

handle_directives([]) -> ok;
handle_directives([{uses, {atom,_,M}}|Rest]) ->
    load_rules({file, atom_to_list(M)}),
    handle_directives(Rest).

store_rules({Directives, Rules}) ->
    handle_directives(Directives),
    store_rules(Rules);
store_rules([]) -> ok;
store_rules([#rule{signature={{atom,_,Name}, Args}, rewrite=R}|Rest]) ->
    %io:format("Rule stored: ~p~n", [Name]),
    case ets:lookup(?ETSTAB, Name) of
        [] -> ets:insert(?ETSTAB, {Name, Args, prepare_rewrite(R)});
        _ -> ?Warning("Rule ~p is already defined, I'm omitting the new definition.", [Name])
    end,
    store_rules(Rest).

prepare_rewrite(#rewrite{left=undefined, right=R, next=N} = Rw) when R /= undefined ->
    Rw#rewrite{right=parse_erl(R), next=prepare_next(N)};
prepare_rewrite(#rewrite{left=L, right=undefined, next=N} = Rw) when L /= undefined ->
    Rw#rewrite{left =parse_erl(L), next=prepare_next(N)};
prepare_rewrite(#rewrite{left=L, right=R, next=N} = Rw) when L /= undefined, R /= undefined ->
    Rw#rewrite{left=parse_erl(L), right=parse_erl(R), next=prepare_next(N)};
prepare_rewrite(#rewrite{left=undefined, right=undefined, next=N} = Rw) ->
    Rw#rewrite{next=prepare_next(N)};
prepare_rewrite(R) -> R.


prepare_next({C, R}) -> {C, prepare_rewrite(R)};
prepare_next(undefined) -> undefined.

parse_erl(Ts) ->
    {ok, Trees} =
        case erl_parser:parse_exprs(Ts ++ [{dot, 0}]) of
            {error, _} ->
                case erl_parser:parse_clause(Ts ++ [{dot, 0}]) of
                    {error, _} ->
                        case erl_parser:parse_form(Ts) of
                            {error, _} ->
                                {ok, erl_parser:parse_forms(Ts)};
                            X -> X
                        end;
                    X -> X
                end;
            {ok, X} when is_list(X), length(X) > 1 -> {ok, [{block, -1, X}]};
            X -> X
        end,
    case Trees of
        [Tree] -> Tree;
        _      -> Trees
    end.

lookup_rule(Name) ->
    case ets:lookup(?ETSTAB, Name) of
        [] ->
            ?Warning("Rule ~p not found.", [Name]),
            error(rule_not_found);
        [{Name, Args, Rs}] ->
            {Args, Rs}
    end.

%% =============================================================================
%% EXECUTION OF REWRITE RULES
%% =============================================================================

%% Executing simple refactorings composed of rewrite rules

syntactic(undefined) -> undefined;
syntactic({'$gn',expr,_} = N) -> N;
syntactic({'$gn',form,_} = N) -> N;
syntactic({'$gn',func,_} = N) ->
    ?Warning("Handling a func target as its definition form", []),
    hd(?Graph:path(N, [{fundef, back}]));
syntactic(N) ->
    ?Warning("Couldn't handle target node ~p, I'm working with all the possible targets", [N]),
    undefined.
     

execute(RuleName, Params, TargetNode) when is_atom(RuleName) ->
    io:format("Executing rule ~p...~n", [RuleName]),
    {Args, R} = lookup_rule(RuleName),
    try
        %% Load the actual parameters of the refactoring
        ets:new(params,   [public, named_table]),
        [ets:insert(params, {Name, Val}) || {{var,_,Name}, Val} <- lists:zip(Args, Params)],

        %% Init the metavariable storage
        ets:new(metavars, [public, named_table]),
        ets:insert(metavars, {make_ref(), []}),

        %% Apply the rewrite rules specified by the refactoring
        begin_transaction(),
        rewrite(syntactic(TargetNode), R),
        end_transaction()
    after
        ets:delete(metavars),
        ets:delete(params)
    end.

begin_transaction() ->
    put(tobereplaced, []).
end_transaction() ->
    case get(tobereplaced) of
        undefined -> ok;
        L ->
            [replace(Old, New) || {Old, New} <- L],
            erase(tobereplaced),
            io:format("ESGFINALIZE1 ~p~n", [catch ?ESG:finalize()]),
            do_scheduled_matches()
    end.

commit() -> end_transaction(), begin_transaction().

candidates(#rewrite{left=undefined}) -> undefined;
candidates(#rewrite{left=L}) -> tree_conversion:nodes_alike(L);
candidates({_Comb, R1, _}) -> candidates(R1).

filter_solutions(Cs, Solutions) ->
    Filter = fun({Id, Vars} = M) ->
                     case valid_solution(M, Cs) of
                         true -> {true, M};
                         B when is_list(B) -> {true, {Id,Vars++B}};
                         _ -> false
                     end
             end,
    lists:filtermap(Filter, Solutions).

%% -----------------------------------------------------------------------------

rewrite_next(_, Y, undefined) ->
    Y;
rewrite_next(_Target, Result, {also, _})
  when Result /= successful ->
    ok;
rewrite_next(Target, Result, {then, #rewrite{modifiers=M}=RW}) ->
    case lists:member(onsuccess, M) of
        true when Result /= successful -> ok;
        _ -> rewrite_next(Target, {then, RW})
    end;
rewrite_next(Target, _, X) ->
    rewrite_next(Target, X).

rewrite_next(Target, {then, RW}) ->
    commit(),
    rewrite(Target, RW);
rewrite_next(Target, {also, RW}) ->
    rewrite(Target, RW);
rewrite_next(_Target, undefined) ->
%    ?d(vege),
    ok.

combine(#rewrite{next=undefined} = R, Next) ->
    R#rewrite{next=Next};
combine(#rewrite{next={C, R2}} = R, Next) ->
    R#rewrite{next={C, combine(R2, Next)}}.


determine_target(_, #rewrite{on = {var,_,On}}) ->
    case lookup_var(On) of
        undefined -> throw(rewrite_on_target_not_found);
        Ns -> Ns
    end;
determine_target(_, #rewrite{in = {var,_,In}}) ->
    ?d(In),
    Res = case lookup_var(In) of
              undefined ->
                  throw(rewrite_in_target_not_found);
              [{'$gn', func, _}|_] = Ns ->
                  ?Query:exec(Ns, [{fundef, back}, funcl]);
              Ns ->
                  ?Query:exec(Ns, ?Expr:deep_sub())
                  %% Here we could affect the order of the nodes (top-down or bottom-up)
          end,
    {only_this_one, Res};
determine_target(undefined, RW) ->
    candidates(RW);
determine_target(Target, #rewrite{modifiers=Ms}=RW) ->
    case lists:member(globally, Ms) of
        true -> determine_target(undefined, RW);
        false -> Target
    end.
    

clean(#rewrite{modifiers=Ms}=RW) ->
    Ms2 = lists:filter(fun(X) -> not lists:member(X, [globally]) end, Ms),
    RW#rewrite{in=undefined,on=undefined,modifiers=Ms2}.


rewrite(Target, #rewrite{apply=[], next=N}) ->
    rewrite_next(Target, failed, N);

rewrite(Target, #rewrite{apply=[R|T], next=N}) ->
    case rewrite(Target, R) of
        successful -> rewrite_next(Target, successful, N);
        _          -> rewrite(Target, #rewrite{apply=T, next = N})
    end;
rewrite(Target, #rewrite{type=insert}=RW) ->
    do_rewrite(Target, RW);
rewrite(undefined, #rewrite{apply={app, _, _}, in=undefined, on=undefined}=RW) ->
    do_rewrite(undefined, RW);
rewrite(Target, #rewrite{next=Next}=RW) ->
    case determine_target(Target, RW) of
        [] ->
            rewrite_next(Target, nothing_done, Next);
        {'$gn',_,_} = N ->
            do_rewrite(N, clean(RW));
        undefined ->
            do_rewrite(undefined, clean(RW));
        {only_this_one, Ns} when is_list(Ns) ->
            #rewrite{modifiers=Ms}=RW,
            Separated = not lists:member(simultanously, Ms),
            R = rewrite_each(Ns, (clean(RW))#rewrite{next=undefined}, Separated),
            rewrite_next(Target, R, Next);
        Ns ->
            rewrite_each(Ns, clean(RW), _Separated=true)
    end.


do_rewrite(Target, #rewrite{right=R, conditions=Cs, type=insert, next=Next}) ->
    Solutions = ets:tab2list(metavars),
    Solutions2 = filter_solutions(Cs, Solutions),
    case Solutions2 of
        [{_Id, Matching}] ->
            New = construct(R, Matching ++ ets:tab2list(params)),
            propose_replace(insert, New),
            case get(tobereplaced) of
                undefined -> io:format("ESGFINALIZE3 ~p~n", [catch ?ESG:finalize()]);
                _ -> ok
            end;
        [] ->
            ?Warning("No solutions!", []); %% error("no solutions");
        _ ->
            error("multiple solutions")
    end,
    rewrite_next(Target, Next);

do_rewrite(Target, #rewrite{apply={app, {atom,_,Name}, AArgs}, next=N}) ->
    {FArgs, R} = lookup_rule(Name),

    AArgs2 = lists:map(fun eval_expr/1, AArgs),

    ?BackupMetavars,
    ?ResetMetavars,
    [ensure_var(Nam, Val) || {{var,_,Nam}, Val} <- lists:zip(FArgs, AArgs2)],
    %% TODO these should be params rather than metavars so that lots
    %% of complications induced by ensure_var could be omitted
    try
        Result = rewrite(Target, combine(R, undefined)),
        ?RestoreMetavars, %% TODO reconsider
        rewrite_next(Target, Result, N)
    catch
        C:R -> ?d({rewrite, {apply, Name}, C, R}),failed
    after
        ?RestoreMetavars
    end;

do_rewrite(Node, #rewrite{left=L, right=R, conditions=Cs, type=Type, next=Next}=RW) ->
    catch put(current_file, hd(?Syn:get_file(Node))),
    log_rewrite(start, Node, RW),
    io:format("~p~n", [ets:tab2list(metavars)]),
    Result =
        try 
            match(Node, L),
            Solutions = ets:tab2list(metavars),
            io:format("~p~n", [Solutions]),
            Solutions2 = filter_solutions(Cs, Solutions),
            io:format("~p~n", [Solutions2]),
            
            ets:delete_all_objects(metavars),
            [ets:insert(metavars, {Id, M}) || {Id, M} <- Solutions2],
            
            case Solutions2 of
                [{_Id, Matching}] ->
                    if Type == replace ->
                            print_old_node(Node),
                            New = construct(R, Matching ++ ets:tab2list(params)),
                            postpone_resultcheck(New);
                       true ->
                            New = undefined
                    end,
                    case Type of
                        replace -> propose_replace(Node, New);
                        remove ->  propose_replace(Node, remove);
                        match_only -> nothing_to_do
                    end,
                    if Type == replace -> postpone_match(New, R, Cs, metalistlengths(Matching));
                       true -> nothing_to_do
                    end,
                    case get(tobereplaced) of
                        undefined when Type /= remove ->
                            io:format("ESGFINALIZE2 ~p~n", [catch ?ESG:finalize()]),
                            do_scheduled_matches();
                        _ -> ok
                    end,
                    successful;
                [] ->
                    ?Warning("No solutions!", []),
                    nothing_done;
                _ ->
                    error("multiple solutions"),
                    failed
            end
        catch
            Cl:Re ->
                ?d({rewrite_error, Cl, Re}),
                failed
        end,
    case Next of
        undefined -> Result;
        _ -> rewrite_next(Node, Result, Next)
    end.

lookup_final_result(Nodes) when is_list(Nodes) -> [lookup_final_result(N) || N <- Nodes];
lookup_final_result({'$gn', lex, _}) -> ok;
lookup_final_result({'$gn', _, _} = Node) ->
    Children = [X || {_, X} <- ?Syn:children(Node)],
    check_result_node(Node, ?ESG:data(Node), Children),
    lookup_final_result(Children).

check_result_node(Node, #expr{type=send_expr}, [L, _, R]) ->
    case ?ESG:data(L) of
        #expr{type=atom, value=final_result_marker} ->
            R2 = tree_conversion:copy(R),
            replace(Node, R2),
            put(result, R2);
        _ ->
            ok
    end;
check_result_node(_, _, _) ->
    ok.

metalistlengths(Matching) ->
    [{Var, length(Val)} || {Var, Val} <- Matching,
                           is_metalist(Var)].

rewrite_each(Xs, RW, Sep) ->
    S = rewrite_each(Xs, RW, Sep, 0),
    if S > 0 -> successful;
       true  -> failed
    end.

rewrite_each([], _, _, S) -> S;
rewrite_each([N|Rest], RW, Separated, S) ->
    io:format("----REWRITE_EACH NEW ELEM ~p-------------------------------~n", [N]),
    ?BackupMetavars,
    S2 =
        try
            case do_rewrite(N, RW) of
                successful ->
                    Separated andalso commit(),
                    S+1;
                _ -> S
            end
        catch
            %% error:bad_node -> ?d("bad node error");
            %% throw:{cannot_match,_,_} -> ?d("match exception")
            C:rR -> 
                ?d({rewrite_each, N, C, rR}),
                S
        after
            Rest == [] andalso (not Separated) andalso commit(),
            ?KeepOldMetavars
        end,
    rewrite_each(Rest, RW, Separated, S2).

%% -----------------------------------------------------------------------------

do_scheduled_matches() ->
    L = get(tobematched),
    [begin
         [delete_var(V) || V <- tree_conversion:vars(Pattern)],
         match(attach_data(Node), Pattern),
         Solutions = ets:tab2list(metavars),
         Solutions2 = filter_solutions(Conditions, Solutions),
         Solutions3 = filter_metalistlengths(MetalistLengths, Solutions2),
         ets:delete_all_objects(metavars),
         [ets:insert(metavars, {Id, M}) || {Id, M} <- Solutions3]
     end || {Node, Pattern, Conditions, MetalistLengths} <- L],
    put(tobematched, []),
    L2 = get(tobechecked),
    [lookup_final_result(N) || N <- L2],
    put(tobechecked, []).

filter_metalistlengths(Ls, Solutions) ->
    Filter = fun(M) ->
                     case valid_solution_metalistlength(M, Ls) of
                         true -> {true, M};
                         _ -> false
                     end
             end,
    lists:filtermap(Filter, Solutions).

valid_solution_metalistlength({_Id, Matching}, Ls) ->
    lists:all(fun({Var,Val}) ->
                      case is_metalist(Var) andalso lists:keymember(Var,1,Ls) of
                          false ->
                              true;
                          true ->
                              {Var, Length} = lists:keyfind(Var, 1, Ls),
                              length(Val) == Length
                      end
              end, Matching).

print_old_node(Nodes) when is_list(Nodes) ->
    lists:foreach(fun print_old_node/1, Nodes);
print_old_node(Node) ->
    io:format("------------------------------------------------------------~n"),
    io:format("ORIGINAL TEXT: ~s~n", [?Syn:flat_text(Node)]).

log_rewrite(start, Node, #rewrite{left=L, right=R, conditions=Cs}) ->
    io:format("========================================================================~n"),
    io:format("~nRunning rewrite on ~p...~n", [Node]),
    catch io:format("~s~n", [?Syn:flat_text(Node)]),
    io:format("left ~p, right ~p, conditions ~p~n~n", [L, R, Cs]).
    
    

%% =============================================================================
%% MATCHING
%% =============================================================================

match({X, Y}) -> match(X, Y).

match(L1, L2=[L3|_]) when is_list(L1), is_list(L3), length(L1) == length(L2) ->
    [match(X, Y) || {X, Y} <- lists:zip(L1, L2)];
match(L1, L2) when is_list(L1), is_list(L2) -> 
    case lists:filter(fun is_metalist/1, L2) of
        [] when length(L1) == length(L2) ->
            [match(X, Y) || {X, Y} <- lists:zip(L1, L2)];
        _ ->
            Comb = possible_combinations(L1, L2),
            Comb == [] andalso ?Warning("No possible combinations of patterns and nodes", []),
            NewMatchings = lists:append([isolated_matching(L) || L <- Comb]),
            ets:delete_all_objects(metavars),
            [ets:insert(metavars, {make_ref(), M}) || {_Id, M} <- NewMatchings],
            ok
    end;
match(?node, {var, _, '_'}) ->
    N, ok;
match(?node, {var, _, VarName}) ->
    ensure_var(VarName, N);
match(Ns0, {var, _, VarName}) when is_list(Ns0) ->
    case is_metalist(VarName) of
        true  ->
            {Ns, _Data} = lists:unzip(Ns0),
            ensure_var(VarName, Ns);
        false -> throw("simple metavars cannot match multiple nodes")
    end;
match(X, Y) -> tree_conversion:match(X, Y).

valid_solution({_Id, Matching}, Cs) ->
    Context = ets:new(context, [public, named_table]),
    try
        [ets:insert(context, M) || M <- Matching],
        %% [?d(catch eval_formula(C)) || C <- Cs],
        %% lists:foldl(fun(false, _)    -> false;
        %%                (_, false)    -> false;
        %%                (true, V )    -> V; %% V can be true or a list of bindings
        %%                (Tuple, true) -> [Tuple];
        %%                (Tuple, L)    -> [Tuple|L]
        %%             end,
        %%             [], lists:map(fun eval_formula/1, Cs)
        try lists:foldl(fun(C, L) ->
                                try eval_formula(C) of
                                    true -> L;
                                    B when is_tuple(B) -> ets:insert(context, B), [B|L];
                                    %% TODO: reconsider whether we
                                    %% still need this list when we
                                    %% extend the context
                                    %% anyway... this way the
                                    %% condition evaluation becomes
                                    %% sequential rather than
                                    %% simultanous
                                    _ -> throw(false)
                                catch
                                    %% TODO FIXME: when re-matching
                                    %% the replacement pattern to the
                                    %% replacement node, conditions
                                    %% referring to metavars only
                                    %% present in the matching pattern
                                    %% will throw bad_node errors;
                                    %% this should not be hidden, but
                                    %% for now this is the easiest
                                    %% solution...
                                    error:bad_node ->
                                        L
                                end
                        end,
                        [], Cs)
        catch throw:false -> false
        end
    after
        ets:delete(Context)
    end.

possible_combinations([], []) ->
    [[]];
possible_combinations(L1, [T|Ts]) ->
    case T of
        {var, _, VarName} ->
            case is_metalist(VarName) of
                false ->
                    [N|Ns] = L1,
                    [[{T, N} | C] || C <- possible_combinations(Ns, Ts)];
                true ->
                    NotMetaList = length(lists:filter(fun(Z) -> not is_metalist(Z) end, Ts)),
                    LLLL = [begin
                                {XX, YY} = lists:split(I, L1),
                                [[{T, XX} | C] || C <- possible_combinations(YY, Ts)]
                            end || I <- lists:seq(0, length(L1) - NotMetaList)],
                    lists:append(LLLL)
            end;
        _ ->
            [N|Ns] = L1,
            [[{T, N} | C] || C <- possible_combinations(Ns, Ts)]
    end;
possible_combinations([], [V={var, _, VarName}]) ->
    case is_metalist(VarName) of
        false -> [[]];
        true  -> {V, []}
    end;
possible_combinations(_, _) ->
    %% [] instead of [[]] perfectly filters out invalid cases
    [].

%% Checks if the matching is valid. Empty list indicates match
%% failure.
isolated_matching(L) ->
    ?BackupMetavars,
    try
        [match(attach_data(Ns), T) || {T, Ns} <- L],
        ets:tab2list(metavars)
    catch
        throw:_R -> []
    after
        ?RestoreMetavars
    end.

solution({_Id, L}) ->
    [{meta(N), if is_list(V) -> lists:map(fun ?Syn:flat_text/1, V);
                  true -> ?Syn:flat_text(V)
               end} || {N, V} <- L].


%% =============================================================================
%% EVALUATION OF CONDITION PREDICATES AND FUNCTIONS
%% =============================================================================

func(Name, Args0) ->
    Args = attach_data(Args0),
    try
        apply(rule_interpreter_lib, Name, Args)
    catch
        error:undef ->
            ?Warning("I did not find function/predicate ~p in rule_interpreter_lib, looking for it refpp_predicate_lib.", [Name]),
            try
                apply(refpp_predicate_lib, Name, detach_data(Args0))
            catch
                error:undef ->
                    ?Warning("I did not find function/predicate ~p in refpp_predicate_lib either, looking for it in the erlang module.", [Name]),
                    try
                        apply(erlang, Name, lists:map(fun extract_value/1, Args))
                    catch
                        error:undef ->
                            ?Warning("Could not execute function/predicate ~p: it does not exist.", [Name]),
                            false;
                        _:_ ->
                            ?Warning("Could not execute function/predicate ~p: some error happened.", [Name]),
                            false
                    end;
                _:_ ->
                    ?Warning("Could not execute function/predicate ~p: some error happened.", [Name]),
                    false
            end;
        error:function_clause ->
            ?Warning("Execution of ~p resulted in function_clause error.", [Name]),
            false
    end.

equivalent(A, B) -> A == B orelse eq(A, B) orelse eq(B, A).

eq(L1, L2) when is_list(L1), is_list(L2), length(L1) == length(L2) ->
    all(lists:zipwith(fun eq/2, L1, L2));
eq({'$gn',_,_}=N1, {{'$gn',_,_}=N2,_D2}) ->
    equivalent_nodes(N1, N2);
eq({'$gn',_,_}=N1, {'$gn',_,_}=N2) ->
    equivalent_nodes(N1, N2);
eq({{'$gn',_,_}=N1,_D1}, {{'$gn',_,_}=N2,_D2}) ->
    equivalent_nodes(N1, N2);
eq({'$gn',expr,_}=N, V) ->
    ?Expr:value(N) == V;
eq({'$gn',module,_}=N, V) ->
    ?Mod:name(N) == V;
eq({{'$gn',module,_},#module{name=V}}, V) ->
    true;
eq(?exprnode(_, V), V) ->
    ?true;
eq(_, _) ->
    false.

extract_value(?exprnode(_, V)) -> N, V;
extract_value(V) -> V.

eval_formula({'<', A, B})   -> eval_expr(A) < eval_expr(B);
eval_formula({'>', A, B})   -> eval_expr(A) > eval_expr(B);
eval_formula({'=', {var,_,X}=A, B}) ->
    case unbound_var(X) of
        false -> equivalent(eval_expr(A), eval_expr(B));
        true  -> {X, detach_data(eval_expr(B))}
    end;
eval_formula({'=', A, B})   -> equivalent(eval_expr(A), eval_expr(B));
eval_formula({'/=', A, B})   -> not equivalent(eval_expr(A), eval_expr(B));
eval_formula({'=<', A, B})  -> eval_expr(A) =< eval_expr(B);
eval_formula({'>=', A, B})  -> eval_expr(A) >= eval_expr(B);
eval_formula({'not', A})    -> not eval_formula(A);
eval_formula({'and', A, B}) -> eval_formula(A) andalso eval_formula(B);
eval_formula({'or', A, B})  -> eval_formula(A) orelse eval_formula(B);
eval_formula({'implies', A, B}) -> not eval_formula(A) orelse eval_formula(B);
eval_formula({predicate, {_,_,P}, As}) ->
%    io:format("Evaluating predicate ~p, ~p: ~p~n", [P, lists:map(fun eval_expr/1, As), [func(P, lists:map(fun eval_expr/1, As))]]),
    func(P, lists:map(fun eval_expr/1, As));
eval_formula({data,{var,_,_}=Var1,{var,_,_}=Var2}) ->
%    io:format("Evaluating dataflow predicate ~p ~p ~p ~p~n", [Var1, Var2, eval_expr(Var1), eval_expr(Var2)]),
%    ?d(?Dataflow:reach_1st([{'$gn', expr, 10}], [back], true)),
    case eval_expr(Var1) of
        {Node1, _} ->
            {Node2, _} = eval_expr(Var2),
            lists:any(fun(?node) -> N == Node1;
                         (_) -> false
                      end,
%                      attach_data(?Dataflow:reach_1st([Node2], [back], true)));
                      attach_data(?Dataflow:reach_1st([Node2], [back])));
        Value when is_atom(Value) ->
            eval_formula({data, {atom,0,Value}, Var2})
    end;
eval_formula({data,{atom,_,Value},{var,_,_}=Var}) ->
    {Node, _Data} = eval_expr(Var),
    lists:any(fun(?exprnode(atom, Value2)) -> N, Value2 == Value;
                 (_) -> false
              end,
              attach_data(?Dataflow:reach_1st([Node], [back], true))).

eval_expr({atom, _, V}) -> V;
eval_expr({string, _, V}) -> V;
eval_expr({integer, _, V}) -> V;
eval_expr({var, _, Name}) ->
    Val = lookup_var(Name),
    attach_data(Val);
eval_expr({'+', A, B}) -> eval_expr(A) + eval_expr(B);
eval_expr({'-', A, B}) -> eval_expr(A) - eval_expr(B);
eval_expr({'*', A, B}) -> eval_expr(A) * eval_expr(B);
eval_expr({'/', A, B}) -> eval_expr(A) / eval_expr(B);
eval_expr({paren, A}) -> eval_expr(A);
eval_expr({app, {_,_,F}, As}) ->
%    io:format("Evaluating application ~p, ~p: ~p~n", [F, lists:map(fun eval_expr/1, As), func(F, lists:map(fun eval_expr/1, As))]),
    func(F, lists:map(fun eval_expr/1, As)).


%% =============================================================================
%% REPLACEMENT CONSTRUCTION
%% =============================================================================

%% TODO: handle newvar predicates as metavar definitions (unbound metavars are newvars I think)

construct(Tree, Metavars) ->
    Context = ets:new(context, [public, named_table]),
    try
        [ets:insert(context, M) || M <- Metavars],
        if is_list(Tree) ->
                lists:flatmap(
                  fun(T) ->
                          [format_node(?Syn:construct(Node)) ||
                              Node <- lists:flatten([tree_conversion:tree2construct(T)])]
                  end, Tree);
           true ->
                C = tree_conversion:tree2construct(Tree),
                if is_list(C) ->
                        [format_node(?Syn:construct(C0)) || C0 <- C];
                   true ->
                        format_node(?Syn:construct(C))
                end
        end
    after
        ets:delete(Context)
    end.

postpone_resultcheck(Node) ->
    L = get(tobechecked),
    put(tobechecked, [Node | L]).

postpone_match(Node, Pattern, Conditions, MetalistLengths) ->
    L = get(tobematched),
    put(tobematched, [{Node, Pattern, Conditions, MetalistLengths} | L]).

propose_replace(Old, New) ->
    case get(tobereplaced) of
        undefined ->
            replace(Old, New);
        L ->
            ?d({propose,replace,[{Old, New} | L]}),
            put(tobereplaced, [{Old, New} | L])
    end.

%% =============================================================================
%% HELPERS
%% =============================================================================

all([])       -> true;
all([true|L]) -> all(L);
all(_)        -> false.

subst(Old, New, List) -> subst(Old, New, List, []).
subst(_, _, [], Acc) -> lists:reverse(Acc);
subst(Old,  New, [Old|List], Acc) ->
    case already_copied(Old) of
        true ->
            subst(Old, New, List, [Old|Acc]);
        false ->
            register_copied(New),
            subst(Old, New, List, [New|Acc])
    end;
subst(Old, New, [Other|List], Acc) ->
    subst(Old, New, List, [Other|Acc]).

format_node(Node) ->
    io:format("------------------------------------------------------------~n"),
    io:format("RESULT TEXT: ~s~n", [?Syn:flat_text(Node)]),
    io:format("------------------------------------------------------------~n"),
    Node.

meta(Name) ->
    S = atom_to_list(Name),
    case is_metalist(Name) of
        true -> string:sub_string(S, 9); 
        false -> S
    end.

is_metalist({var, _, Name}) ->
    is_metalist(Name);
is_metalist(Name) when is_atom(Name) ->
    string:str(atom_to_list(Name), "METALIST") == 1;
is_metalist(_) ->
    false.

attach_data(Ns) when is_list(Ns) -> [attach_data(N) || N <- Ns];
attach_data({'$gn', _, _} = N)   -> {N, ?Graph:data(N)};
attach_data(X) -> X.

detach_data(Ps) when is_list(Ps)        -> [detach_data(P) || P <- Ps];
detach_data({{'$gn', _, _} = N, _Data}) -> N;
detach_data(X) -> X.

%replace(Old, {construct, X, Y}) ->
%    replace(Old, construct(X, Y));
replace(Node, remove) ->
    [{_, Parent}|_] = ?Syn:parent(Node),
    ?Syn:replace(Parent, {node, Node}, []);
%    ?ESG:remove(Parent, Link, Node);
replace(insert, Node) ->
    ?ESG:insert(get(current_file), form, Node); %% TODO: refine
replace(Old, New) when not is_list(New) ->
    replace(Old, [New]);
replace([Old|_] = OldNodes, New) ->
    [{Link, Parent}|_] = ?Syn:parent(Old),
    case {Parent, New} of
        {{'$gn', file, _}, NewNodes} ->
            [?ESG:remove(Parent, Link, OldNode) || OldNode <- OldNodes],
            [?ESG:insert(Parent, Link, NewNode) || NewNode <- NewNodes];
        _ ->
            ?Syn:replace(Parent, {node, Old}, New)
    end,
    refresh_metavars(Old, hd(New));
replace(Old, New) ->
    [{Link, Parent}|_] = ?Syn:parent(Old),
    case {Parent, New} of
        {{'$gn', file, _}, NewNodes} ->
            %% TODO: replace does not support replacement of forms
            %% (refcore_erl_nodes does not know the syntactic
            %% structure of file nodes)
            ?ESG:remove(Parent, Link, Old),
            [?ESG:insert(Parent, Link, NewNode) || NewNode <- NewNodes];
        _ ->
            ?Syn:replace(Parent, {node, Old}, New)
    end,
    refresh_metavars(Old, hd(New)). %% TODO: and all the copied parts? SOLVED BY REMATCHING PERHAPS
%% it is not solved without this; if we replace something in E.. and then refer to E..'s elements, they are outdated; rematching does not help at this...

%% -----------------------------------------------------------------------------
%% Checks if two trees given in the standard abstract form are equivalent

equivalent_nodes(L1, L2) when is_list(L1), is_list(L2), length(L1) == length(L2) ->
    all(lists:zipwith(fun equivalent_nodes/2, L1, L2));

equivalent_nodes(#form{type=T, tag=V},
                 #form{type=T, tag=V}) -> true;
equivalent_nodes(#clause{type=T},
                 #clause{type=T}) -> true;
equivalent_nodes(#expr{type=T, value=V},
                 #expr{type=T, value=V}) -> true;
equivalent_nodes(#lex{type=T, data=D1},
                 #lex{type=T, data=D2}) -> equivalent_nodes(D1, D2);
equivalent_nodes(#token{type=T, text=V},
                 #token{type=T, text=V}) -> true;
equivalent_nodes(#expr{value=V}, V) -> true;
equivalent_nodes({{'$gn', _, _} = N1, Data}, {'$gn', _, _} = N2) ->
    Children = fun(N) -> [X||{_,X}<-?Syn:children(N)] end,
    equivalent_nodes(Data, ?Graph:data(N2)) andalso
        equivalent_nodes(Children(N1), Children(N2));
equivalent_nodes({'$gn', _, _} = N1, {{'$gn', _, _} = N2, Data}) ->
    Children = fun(N) -> [X||{_,X}<-?Syn:children(N)] end,
    equivalent_nodes(?Graph:data(N1), Data) andalso
        equivalent_nodes(Children(N1), Children(N2));
equivalent_nodes({'$gn', _, _} = N1, {'$gn', _, _} = N2) ->
    Children = fun(N) -> [X||{_,X}<-?Syn:children(N)] end,
    equivalent_nodes(?Graph:data(N1), ?Graph:data(N2)) andalso
        equivalent_nodes(Children(N1), Children(N2));
equivalent_nodes({'$gn', _, _} = N, V) ->
    equivalent_nodes(?Graph:data(N), V);
equivalent_nodes(_X, _Y) ->
    false.




%% =============================================================================
%% CONFIGURATION HANDLING
%% =============================================================================

unbound_var(VarName) ->
    lookup_var(VarName) == undefined.

lookup_var('PREVIOUS_RESULT') ->
    [get(result)];

lookup_var(VarName) ->
    case lookup_metavar(VarName) of
        undefined -> lookup_param(VarName);
        Value -> Value
    end.

lookup_metavar(VarName) ->
    L = case ets:info(context) of
            undefined ->
                lists:append(element(2, lists:unzip(ets:tab2list(metavars))));
            _ ->
                ets:tab2list(context)
        end,
    case lists:keyfind(VarName, 1, L) of
        false -> undefined;
        {VarName, Value} -> Value
    end.

lookup_param(VarName) ->
    case ets:lookup(params, VarName) of
        [] -> undefined;
        [{VarName, Value}] -> Value
    end.

delete_var(VarName) ->
    Ps = ets:tab2list(metavars),
    Ps2 = lists:filtermap(
            fun({Id, L}) ->
                    case lists:keymember(VarName, 1, L) of
                        true -> {true, {Id, lists:keydelete(VarName, 1, L)}};
                        false -> false
                    end
            end, Ps),
    [ets:delete(metavars, Id)         || {Id, _   } <- Ps2],
    [ets:insert(metavars, {Id, Data}) || {Id, Data} <- Ps2, is_list(Data)].

ensure_var(VarName, Value) ->
    Ps = ets:tab2list(metavars),
    Ps2 = lists:filtermap(
            fun({Id, L}) ->
                    case lists:keyfind(VarName, 1, L) of
                        false ->
                            case ets:lookup(params, VarName) of
                                [] ->
                                    {true, {Id, L ++ [{VarName, Value}]}};
                                [{VarName, Value2}] ->
                                    case equivalent(Value, Value2) of
                                        true  -> false;
                                        false -> {true, {Id, delete}}
                                    end
                            end;
                        {VarName, Value2} ->
                            case equivalent(Value, Value2) of
                                true  -> false;
                                false -> {true, {Id, delete}}
                            end
                    end
            end, Ps),
    [ets:delete(metavars, Id)   || {Id, _   } <- Ps2],
    [ets:insert(metavars, {Id, Data}) || {Id, Data} <- Ps2, is_list(Data)].

refresh_metavars(Old, New) ->
    Ps = ets:tab2list(metavars),
    Ps2 = lists:map(
            fun({Id, L}) ->
                    {Id,
                     lists:map(fun({VarName, Value}) when is_list(Value) ->
                                       {VarName, subst(Old, New, Value)};
                                  ({VarName, Value}) ->
                                       [R] = subst(Old, New, [Value]),
                                       {VarName, R}
                               end,
                               L)}
            end, Ps),

%    ?d(refresh_metavars),
%    ?d(Ps),
%    ?d(Ps2),

    [ets:delete(metavars, Id)   || {Id, _   } <- Ps2],
    [ets:insert(metavars, {Id, Data}) || {Id, Data} <- Ps2, is_list(Data)].


overwrite_var(VarName, Value) ->
    Ps = ets:tab2list(metavars),
    Ps2 = lists:filtermap(
            fun({Id, L}) ->
                    case lists:keyfind(VarName, 1, L) of
                        false ->
                            false;
                        {VarName, _Value2} ->
                            case already_copied(VarName) of
                                true ->
                                    false;
                                false ->
                                    register_copied(VarName),
                                    {true, {Id, lists:keyreplace(VarName, 1, L, {VarName, Value})}}
                            end
                    end
            end, Ps),
    [ets:delete(metavars, Id)   || {Id, _   } <- Ps2],
    [ets:insert(metavars, {Id, Data}) || {Id, Data} <- Ps2, is_list(Data)].

already_copied(V) -> lists:member(V, get(already_copied)).
register_copied(V) -> put(already_copied, [V|get(already_copied)]).
