%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

%%% @author Dniel Horpcsi <daniel-h@elte.hu>

-module(rule_scanner).
-export([scan/1]).

%% These standard Erlang symbols are allowed to be used in the rule
%% definitions. They are part of the description and not Erlang
%% symbols unless they are put in the Erlang-code section.
-define(Spec, ['->','(', ')', '{', '}', '[', ']',
               '-', '+', '*', '/', '>', '<', '=<', '>=', '=', '/=',
               ',', '--', ':', '|',
               '~', '\\', 'not', 'and', 'or']).

-define(KEYWORDS,
        %% scanned as variables by the standard scanner but returned as keyword tokens by this version
        ['USES',
         'RULE', 'REFACTORING', 'APPLY',
         %% Combinators:
         'THEN', 'ALSO', 'ORTRY',
         %% Modifiers:
         'IN', 'ON', 'WHERE', 'GLOBALLY', 'ONSUCCESS','SIMULTANOUSLY',
         %% Data-flow and control-flow predicates:
         'EXISTS', 'FORALL', 'PATH', 'INCLUDING', 'EXCLUDING', 'SATISFIES',
         %% Logic operations for combining predicates:
         'NOT', 'AND', 'OR', 'IMPLIES']).

scan(String) when is_list(String) ->
    String2 = re:replace(String, "\n\s*-+\s*\n", " \n~\n ", [global, {return, list}]),
    {ok, Tokens, _} = erl_scan:string(String2),
    preprocess(Tokens);
scan(In) ->
    String = read(In),
    scan(String).
    
read(File) ->
    case file:read_line(File) of
        {ok, Data} -> Data ++ read(File);
        eof        -> []
    end.

preprocess(Tokens) -> lists:map(fun process/1, replaces(Tokens)).

process(V = {var, Line, Name}) ->
    case lists:member(Name, ?KEYWORDS) of
        true  -> {list_to_atom(string:to_lower(atom_to_list(Name))), Line};
        false -> V
    end;
process(V = {Atom, _}) ->
    case lists:member(Atom, ?Spec) of
        true  -> V;
        false -> reserved(V)
    end;
process({Category, Line, Symbol}) -> {Category, Line, Symbol}.

reserved({Atom, Line}) ->
    Cat = case erl_scan:reserved_word(Atom) of
              true -> reserved_word;
              false -> reserved_symbol
          end,
    {Cat, Line, Atom}.

%%% ============================================================================
%%% Manipulating the token stream in order to make it parseable or
%%% change legal Erlang code fragments into special rule operators

replaces([{'\\', L1}, {'\\', _} | T]) ->
    [{atom,L1,final_result_marker}, {'!',L1} | replaces(T)];

replaces([{'\\', L1}, {'(', _} | T]) ->
    [{'\\', L1}, {'{', L1} | replaces(T)];
replaces([{'\\', L1}, {')', _} | T]) ->
    [{'\\', L1}, {'}', L1} | replaces(T)];

replaces([{'=', L1}, {'>', _} | T]) -> % prior to R17
    [{'~', L1} | replaces(T)];
replaces([{'=>', L1} | T]) -> % from R17
    [{'~', L1} | replaces(T)];

replaces([{var, Line, V}, {'..', _} | T]) ->
    [{var, Line, list_to_atom("METALIST" ++ atom_to_list(V))} | replaces(T)];

replaces([H|T]) -> [H | replaces(T)];
replaces([]) -> [].
