-module(tree_conversion).

-include_lib("referl_core/src/core.hrl").
-include("rules.hrl").


-export([tree2construct/1]).
-export([match/2]).
-export([nodes_alike/1]).
-export([vars/1]).
-export([copy/1]).

%-------------------------------------------------------------------------------
% ==============================================================================
% Convert standard Erlang abstract form to RefactorErl abstract form
% processed by the construct function
% ==============================================================================
%-------------------------------------------------------------------------------

%% NOTE TO SELF: in the modified parser, funclauses contain an
%% additional field: the function name; matching, lookfor as well as
%% construction are adjusted to this modification


%% - listcomp headclause may only have a single headexpr...
%% - listcomp generator only has a single pattern and body expr...
%% - args are within arglist in referl...

%% still doesn't support every syntactic construct...
tree2construct({attribute, _, module, Name}) ->
    {{form, module, Name}, []};
tree2construct({attribute, _, export, List}) ->
    L = [{funref, get_type(Name), get_type(Arity)} || {Name, Arity} <- List],
    {{form, export}, [{funlist, L}]};
tree2construct({attribute, _, import, {Module, List}}) ->
    L = [{funref, get_type(Name), get_type(Arity)} || {Name, Arity} <- List],
    {{form, import}, [get_type(Module), {funlist, L}]};

tree2construct({clause, L, Name, P, G, B}) ->
    tree2construct({fun_clause, get_type(Name),  {clause, L, P, G, B}});

tree2construct({function, _, _, _, [{var,_,_}]=Clauses}) ->
    {func, tree2construct(Clauses)}; %% TODO still just a nice workaround, the idea should be generalized

tree2construct({function, _, Name, _, Clauses}) ->
    {func, [tree2construct({fun_clause, get_type(Name), C}) || C <- Clauses]};

tree2construct({Kind, Header, {clause, _, _Name, Pattern, Guard, Body}}) -> %% compatibility issues
    {Kind, Header,
     lists:flatten([tree2construct(P) || P <- Pattern]),
     guard_disj(Guard),
     lists:flatten([tree2construct(B) || B <- Body])};

tree2construct({Kind, Header, {clause, _, Pattern, Guard, Body}}) ->
    {Kind, Header,
     lists:flatten([tree2construct(P) || P <- Pattern]),
     guard_disj(Guard),
     lists:flatten([tree2construct(B) || B <- Body])};
tree2construct({'catch', {clause, _, Pattern, Guard, Body}}) ->
    {pattern, [tuple_to_infix(Pattern)],
     guard_disj(Guard),
     lists:flatten([tree2construct(B) || B <- Body])};
tree2construct({guard, {clause, _, [], Guard, Body}}) ->
    {guard, guard_disj(Guard), lists:flatten([tree2construct(B) || B <- Body])};
tree2construct({_Kind, {var, _, _}= V}) ->
    tree2construct(V); %% TODO this is a workaround, should be generelised
tree2construct({Kind, {clause, L, _, Pattern, Guard, Body}}) ->
    tree2construct({Kind, {clause, L, Pattern, Guard, Body}});
tree2construct({Kind, {clause, _, Pattern, Guard, Body}}) ->
    {Kind, lists:flatten([tree2construct(P) || P <- Pattern]),
     guard_disj(Guard),
     lists:flatten([tree2construct(B) || B <- Body])};

tree2construct({record, _, Name, Attr}) ->
    {{record_expr, Name}, [tree2construct(A) || A <- Attr]};
tree2construct({record, _, Var, Name, Fields}) ->
    {{record_update, Name}, tree2construct(Var), [tree2construct(F) || F <- Fields]};
tree2construct({record_field, _, {atom, _, Field}, Value}) ->
    {{record_field, Field}, tree2construct(Value)};
tree2construct({record_field, _, Access, Field, Value}) ->
    {{record_access, Field}, tree2construct(Access), tree2construct(Value)};
tree2construct({record_index, _, Name, Field}) ->
    {{record_index, Name}, tree2construct(Field)};

tree2construct({'fun', _, Attr}) ->
    tree2construct(Attr);
tree2construct({clauses, Clauses}) ->
    {'fun', lists:flatten([tree2construct({fun_scope, C}) || C <- Clauses])};
tree2construct({function, Module, Name, Arity}) ->
    {implicit_fun, {{infix_expr, ':'}, get_type(Module), get_type(Name)}, get_type(Arity)};
tree2construct({function, Name, Arity}) ->    
    {implicit_fun, get_type(Name), get_type(Arity)};

tree2construct({block, _, Clauses}) ->
    {block_expr, [tree2construct(C) || C <- Clauses]};
tree2construct({'case', _, HeadCl, Branches}) ->
    {'case', tree2construct(HeadCl),
     lists:flatten([tree2construct({pattern, B}) || B <- Branches])};
tree2construct({'catch', _, Expr}) ->
    {'catch', tree2construct(Expr)};
tree2construct({'if', _, Clauses}) ->
    {'if', lists:flatten([tree2construct({guard, C}) || C <- Clauses])};
tree2construct({'receive', _, Clauses}) ->
    {'receive', [tree2construct({pattern, C}) || C <- Clauses], []};
tree2construct({'receive', _, Clauses, TimeOut, TBody}) ->
    {'receive', [tree2construct({pattern, C}) || C <- Clauses],
     [tree2construct(TimeOut), [tree2construct(B) || B <- TBody]]};
tree2construct({'try', _, HeadCl, ExprCl, CatchCl, AfterCl}) ->
    {'try', [tree2construct(H) || H <- HeadCl],
     [tree2construct({pattern, E}) || E <- ExprCl],
     [tree2construct({'catch', C}) || C <- CatchCl],
     [tree2construct(A) || A <- AfterCl]};

tree2construct(List) when (element(1, List) == nil) or (element(1, List) == cons) ->
    abslist_to_list(List);
tree2construct({lc, _, HeadExpr, Generator}) ->
    {list_comp, tree2construct(HeadExpr),
     [tree2construct(G) || G <- Generator]};
tree2construct({generate, _, Pattern, Clause}) ->
    {list_gen, tree2construct(Pattern), tree2construct(Clause)};

tree2construct({tuple, _, Attr}) ->
    {tuple, [tree2construct(A) || A <- Attr]};

tree2construct({match, _, LeftOp, RightOp}) ->
    {match_expr, tree2construct(LeftOp), tree2construct(RightOp)};
tree2construct({op, _, '!', LeftOp, RightOp}) ->
    {send_expr, tree2construct(LeftOp), tree2construct(RightOp)};
tree2construct({op, _, Operator, LeftOp, RightOp}) when (Operator == 'and') or 
                                                        (Operator == 'or') or
                                                        (Operator == 'andalso') or 
                                                        (Operator == 'orelse') ->
    {{infix_expr, Operator}, {paren, tree2construct(LeftOp)},
     {paren, tree2construct(RightOp)}};
tree2construct({op, _, Operator, LeftOp, RightOp}) ->
    {{infix_expr, Operator}, tree2construct(LeftOp),
     tree2construct(RightOp)};
tree2construct({op, _, Operator, Operand}) ->
    {{prefix_expr, Operator}, tree2construct(Operand)};

tree2construct({bin, _, Parameters}) ->
    {bin, [tree2construct(P) || P <- Parameters]};
tree2construct({bc, _, HeadExpr, Generator}) ->
    {bin_comp, tree2construct(HeadExpr), [tree2construct(G) || G <- Generator]};
tree2construct({b_generate, _, Pattern, Clause}) ->
    {bin_gen, tree2construct(Pattern), tree2construct(Clause)};
tree2construct({bin_element, _, Value, Size, TSL}) ->
    {binary_field, [case Size of
                        default -> tree2construct(Value);
                        _       -> {size_qualifier, tree2construct(Value), tree2construct(Size)}
                    end,
                    get_tsl(TSL)]};

tree2construct({call, _, Name, Parameters}) ->
    {app, tree2construct(Name), [tree2construct(P) || P <- Parameters]};
tree2construct({remote, _, Module, Fun}) ->
    {{infix_expr, ':'}, tree2construct(Module), tree2construct(Fun)};

tree2construct({var, _, '_'}) ->
    {joker, []};

tree2construct({var, _, Name}) ->
    %%?d({var, construction, context, ets:tab2list(context)}),
    case ets:lookup(context, Name) of
        [] ->
            GName = Name,
            %GName = list_to_atom(atom_to_list(Name) ++ integer_to_list(round(random:uniform() * math:pow(10, 2)))),
            %ets:insert(context, {Name, {var, 0, GName}}),
            {var, GName};
        %% TODO: Generate fresh variable names here, or do the generation in the interpreter?
        [{Name, Value}] ->
            C = create_or_copy(Value),
            %%?d("HOPPA_tree2construct"),
            %%?d(Name),
%            ?d(attach_data(lookup_var(Name))),
%%%%            overwrite_var(Name, C),
%            ?d(attach_data(lookup_var(Name))),
            C
    end;

tree2construct({Kind, _, Value}) ->
    {Kind, Value};

tree2construct(L) when is_list(L) ->
    lists:map(fun tree2construct/1, L);

tree2construct(_) ->
    [].

create_or_copy({'$gn',_,_} = Node) ->
    C = copy(Node),
%    refresh_metavars(Node, C),
    C;
create_or_copy(Values) when is_list(Values) ->
    lists:map(fun create_or_copy/1, Values);
create_or_copy(Value) ->
    get_type(Value).
    
abslist_to_list(List) -> abslist_to_list(List, []).

abslist_to_list({nil, _}, Elems) ->
    case Elems of
        [] -> {cons, []};
        _  -> {cons, {list, lists:flatten(Elems)}}
    end;
abslist_to_list({cons, _, Head, Tail}, Elems) ->
    abslist_to_list(Tail, Elems ++ [tree2construct(Head)]);
abslist_to_list(ImproperTail, Elems) ->
    {cons, {list, lists:flatten(Elems)}, tree2construct(ImproperTail)}.

%% @spec get_tsl(Value) -> list()
%%       Value = atom() | list()
%% @doc
get_tsl(default) ->
    [];
get_tsl([]) ->
    [];
get_tsl([Type | Rest]) ->
    [case Type of
	{T, Size} ->
	    {{bit_size_expr, Size}, get_type(T)};
	_ ->
	    get_type(Type)
    end | get_tsl(Rest)];
get_tsl(Type) ->
    get_type(Type).

guard_disj([]) -> [];
guard_disj([Head|Rest]) ->
    case Rest of
        [] -> guard_conj(Head);
	_  -> {{infix_expr, ';'}, guard_conj(Head), guard_disj(Rest)}
    end.

guard_conj([Head|Rest]) ->
    case Rest of
	[] -> tree2construct(Head);
	_  -> {{infix_expr, ','}, tree2construct(Head), guard_conj(Rest)}
    end.

%% @spec tuple_to_infix(tuple()) -> tuple()
%% @doc
tuple_to_infix([{tuple, _, [Class, Pattern, _]}]) ->
    case Class of
        'throw' ->
            tree2construct(Pattern);
        Class ->
            {{infix_expr, ':'}, tree2construct(Class), tree2construct(Pattern)}
    end.

get_type({var, _, Name}) ->
    case ets:lookup(context, Name) of
        [] -> {var, Name};
        [{Name, Value}] -> create_or_copy(Value)
    end;
get_type({T, _, V}) -> {T, V};
get_type(Value) when is_atom(Value)    -> {atom, Value};
get_type(Value) when is_integer(Value) -> {integer, Value};
get_type(Value) when is_float(Value)   -> {float, Value};
get_type(Value) when is_tuple(Value)   ->
    {tuple, [get_type(element(I, Value)) || I <- lists:seq(1, size(Value))]};
get_type(Value) when ((hd(Value) >= $A) and (hd(Value) =< $Z)) orelse
		     ((hd(Value) >= $a) and (hd(Value) =< $z))->
    {string, Value};
get_type(Value) ->
    {cons, {list, [get_type(Element) || Element <- Value]}}.

copy(Nodes) when is_list(Nodes) -> [copy(Node) || Node <- Nodes];
copy(Node)                      -> proplists:get_value(Node, ?Syn:copy(Node, fun(_) -> false end, true)).



%-------------------------------------------------------------------------------
% ==============================================================================
% Not converting, but matching the standard abstract syntax with the
% RefactorErl tree (not the RefactorErl abstract form, but the actual
% tree nodes). Warning: the abstract form may contain metavariables!
%
% Throws exception if the trees are not equivalent and cannot be matched.
% ==============================================================================
% -------------------------------------------------------------------------------

-define(match(X, Y), rule_interpreter:match(X, Y)).

match({'$gn', _, _} = Node, Tree) -> ?match({Node, ?Graph:data(Node)}, Tree);

%% -------------------- FORMS AND CLAUSES --------------------

match(?formnode(export),
      {attribute,_,export,LT})
->
    [FunList] = ?Graph:path(N, [eattr]),
    ?match(FunList, LT);

match(?formnode(func),
      {function,_,_N,_A,ClT})
->
    ?match(?Graph:path(N, [funcl]), ClT);

match(?clausenode(fundef),
      {clause,_,Name,HT,GT,BT})
->
    ?match(?Graph:path(N, [name]), [Name]),
    ?match(?Graph:path(N, [pattern]), HT),
    ?match(?Graph:path(N, [guard]), GT),
    ?match(?Graph:path(N, [body]), BT);

%% -------------------- TERMS --------------------

match(?exprnode(T, Value),
      {T, _, Value}) when T == integer;
                          T == float;
                          T == atom;
                          T == char;
                          T == string ->
    N, ok;

match(?exprnode(T, Value),
      Value) when T == integer;
                  T == float;
                  T == atom;
                  T == char;
                  T == string ->
    N, ok;

match(?exprnode(tuple),
      {tuple, _, SubsT})
->
    ?match(?Graph:path(N, [esub]), SubsT);

match(?exprnode(cons),
      {nil, _})
->
    [] = ?Graph:path(N, [esub]);

match(?exprnode(cons),
      {cons, _, HT, {nil, _}})
->
    case {?Graph:path(N, [esub]), HT} of
        {[], [{var,_,_}]} -> ?match([], hd(HT));
        _ -> ?match(?Graph:path(N, [esub]), [HT])
    end;

match(?exprnode(cons),
      {cons, _, HT, TT})
->
    ?match(?Graph:path(N, [esub]) , [HT, TT]);
            
match(?exprnode(list),
      EsT)
->
    ?match(?Graph:path(N, [esub]), EsT);

%% -------------------- COMPOUND EXPRESSIONS --------------------

match(?exprnode(prefix_expr, Op),
      {op, _, Op, RT})
->
    ?match(?Graph:path(N, [esub]), [RT]);

%match(?exprnode(send_expr),
%      {op, _, '!', {atom, _, final_result_marker}, RT})
%->
%    ?match(N, RT);

match(?exprnode(send_expr),
      {op, _, '!', LT, RT})
->
    ?match(?Graph:path(N, [esub]), [LT, RT]);

match(?exprnode(infix_expr, Op),
      {op, _, Op, LT, RT})
->
    ?match(?Graph:path(N, [esub]), [LT, RT]);

match(?exprnode(infix_expr, ':'),
      {remote, _, MT, FT})
->
    ?match(?Graph:path(N, [esub]), [MT, FT]);

%% -------------------- LISTCOMP EXPRESSIONS --------------------

match(?exprnode(list_comp),
      {lc, _, HT, GFT})
->
    [H, GF] = ?Graph:path(N, [exprcl]),
    ?match(?Graph:path(H, [body]), [HT]),
    ?match(?Graph:path(GF, [body]), GFT);

match(?exprnode(list_gen),
      {generate, _, PT, LT})
->
    [P, L] = ?Graph:path(N, [exprcl]),
    ?match(?Graph:path(P, [pattern]), [PT]),
    ?match(?Graph:path(L, [body]), [LT]);

%% -------------------- BLOCK, MATCH, APPLICATION EXPRESSIONS --------------------

match(?exprnode(block_expr),
      {block, _, EsT})
->
    ?match(?Graph:path(N, [exprcl, body]), EsT);

match(?exprnode(match_expr),
      {match, _, LT, RT})
->
    ?match(?Graph:path(N, [esub]), [LT, RT]);

match(?exprnode(application),
      {call, _, FT, ArgsT})
->
    [F, A0] = ?Graph:path(N, [esub]),
    Args = ?Graph:path(A0, [esub]),
    ?match([F, Args], [FT, ArgsT]);

%% -------------------- BRANCHING EXPRESSIONS --------------------

match(?exprnode(if_expr),
      {'if', _, CsT})
->
    ?match(?Graph:path(N, [exprcl]), CsT);

match(?clausenode(guard),
      {clause,_,[],GT,BT})
->
    ?match(?Graph:path(N, [guard]), GT),
    ?match(?Graph:path(N, [body]), BT);

match(?exprnode(case_expr),
      {'case', _, HT, CsT})
->
    ?match(?Graph:path(N, [headcl,body]), [HT]),
    ?match(?Graph:path(N, [exprcl]), CsT);

match(?clausenode(pattern),
      {clause,_,HT,GT,BT})
->
    ?match(?Graph:path(N, [pattern]), HT),
    ?match(?Graph:path(N, [guard]), GT),
    ?match(?Graph:path(N, [body]), BT);

%% -------------------- FUN EXPRESSIONS --------------------

match(?exprnode(fun_expr),
      {'fun',_,{clauses,CsT}}) ->
    ?match(?Graph:path(N, [exprcl]), CsT);

match(?clausenode(funexpr),
      {clause,_,_,HT,GT,BT}) %% TODO this is the new form of the clause, not the standard... (see the modified parser)
->
    ?match(?Graph:path(N, [pattern]), HT),
    ?match(?Graph:path(N, [guard]), GT),
    ?match(?Graph:path(N, [body]), BT);

%% match(?clausenode(funexpr),
%%       {clause,_,HT,GT,BT})
%% ->
%%     ?match(?Graph:path(N, [pattern]), HT),
%%     ?match(?Graph:path(N, [guard]), GT),
%%     ?match(?Graph:path(N, [body]), BT);

match(?exprnode(implicit_fun),
      {'fun',_,{function,F,A}})
->
    [Name, Arity] = ?Graph:path(N, [esub]),
    ?match([Name, Arity], [F, A]);

match(?exprnode(implicit_fun),
      {'fun',_,{function,M,F,A}})
->
    [Name, Arity] = ?Graph:path(N, [esub]),
    [ModName, FunName] = ?Graph:path(Name, [esub]),
    ?match([ModName, FunName, Arity], [M, F, A]);

match(?exprnode(funlist),
      L)
->
    ?match(?Graph:path(N, [esub]), L);

match(?exprnode(funref),
      {F, A})
->
    ?match(?Graph:path(N, [esub]), [F, A]);

%% -------------------- OTHERWISE --------------------
  
match(X, Y) ->
    throw({cannot_match, X, Y}).


%-------------------------------------------------------------------------------
% ==============================================================================
% Vars in the standard abstract term
% ==============================================================================
%-------------------------------------------------------------------------------

vars(L) when is_list(L) -> lists:flatmap(fun vars/1, L);
vars({function,_,_N,_A,ClT}) -> vars(ClT);
vars({clause,_,Name,HT,GT,BT}) -> vars([Name,HT,GT,BT]);
vars({tuple, _, SubsT}) -> vars(SubsT);
vars({cons, _, HT, {nil, _}}) -> vars(HT);
vars({cons, _, HT, TT}) -> vars([HT,TT]);
vars({op, _, _Op, RT}) -> vars(RT);
vars({op, _, _Op, LT, RT}) -> vars([LT,RT]);
vars({remote, _, MT, FT}) -> vars([MT,FT]);
vars({lc, _, HT, GFT}) -> vars([HT,GFT]);
vars({generate, _, PT, LT}) -> vars([PT,LT]);
vars({block, _, EsT}) -> vars(EsT);
vars({match, _, LT, RT}) -> vars([LT,RT]);
vars({call, _, FT, ArgsT}) -> vars([FT,ArgsT]);
vars({clause,_,HT,GT,BT}) -> vars([HT,GT,BT]);
vars({'if', _, CsT}) -> vars(CsT);
vars({'case', _, HT, CsT}) -> vars([HT,CsT]);
vars({'fun',_,{clauses,CsT}}) -> vars(CsT);
vars({'fun',_,{function,F,A}}) -> vars([F,A]);
vars({'fun',_,{function,M,F,A}}) -> vars([M,F,A]);
vars({var,_,Var}) -> [Var];
vars(_) -> [].


%-------------------------------------------------------------------------------
% ==============================================================================
% Looking for nodes in the RefactorErl tree that might be equivalent
% to the tree given in standard abstract form.
% ==============================================================================
%-------------------------------------------------------------------------------


nodes_alike({function, _, {var, _, _}, _Arity, _Clauses}) ->
    functions();
nodes_alike({function, _, {atom, _, Name}, _Arity, _Clauses}) ->
    case functions(Name) of
        [] -> functions(); %% we might look for a new function which has not been analysed by the semantics analyser yet
        L -> L
    end;
nodes_alike({T, _, Value}) when T == integer; T == float;
                                T == atom;    T == char;
                                T == string;  T == tuple ->
    lookfor(expr, T, Value);
nodes_alike({nil, _}) ->
    lists:filter(has_n_children(0), lookfor(expr, cons));
nodes_alike({cons, _, _, {nil, _}}) ->
    lists:filter(has_n_children(1), lookfor(expr, cons));
nodes_alike({cons, _, _, _}) ->
    lists:filter(has_n_children(2), lookfor(expr, cons));
nodes_alike({op, _, Op, _})      -> lookfor(expr, prefix_expr, Op);
nodes_alike({op, _, Op, _, _})   -> lookfor(expr, infix_expr, Op);
nodes_alike({remote, _, _, _})   -> lookfor(expr, infix_expr, ':');
nodes_alike({lc, _, _, _})       -> lookfor(expr, list_comp);
nodes_alike({generate, _, _, _}) -> lookfor(expr, list_gen);
nodes_alike({block, _, _})       -> lookfor(expr, block_expr);
nodes_alike({match, _, _, _})    -> lookfor(expr, match_expr);
nodes_alike({call, _, _, _})     -> lookfor(expr, application);
nodes_alike({'if', _, _})        -> lookfor(expr, if_expr);
nodes_alike({clause,_,[],_,_})   -> lookfor(clause, guard);
nodes_alike({'case', _, _})      -> lookfor(expr, case_expr);
nodes_alike({clause,_,_,_,_})    -> lookfor(clause, [pattern, funexpr]);
nodes_alike({clause,_,_,_,_,_})    -> lookfor(clause, [fundef]);
nodes_alike({'fun',_,{clauses,_}})       -> lookfor(expr, fun_expr);
nodes_alike({'fun',_,{function,_,_}})    -> lookfor(expr, implicit_fun);
nodes_alike({'fun',_,{function,_,_, _}}) -> lookfor(expr, implicit_fun);
nodes_alike([H|_]) when is_tuple(H),
                        element(1,H)==attribute orelse
                        element(1,H)==function ->
    lists:append([combinations(forms(File)) || File <- files()]).
%%    lists:append([sublists(length(L), forms(File)) || File <- files()]).

files() -> ?Query:exec([file]).
forms(File) -> ?Query:exec(File, [form]).

combinations([]) -> [];
combinations([H | T]) ->
    CT = combinations(T),
    [[H]] ++ [[H | L] || L <- CT] ++ CT.

sublists(N, L) when N =< length(L) ->
    [lists:sublist(lists:nthtail(U,L),N) || U <- lists:seq(0, length(L) - N)].

has_n_children(N) -> fun(Node) -> length(?Graph:path(Node, [esub])) == N end.

lookfor(expr, Type) ->
    exprs(fun(Expr) -> ?Expr:type(Expr) == Type end);
lookfor(clause, Types) when is_list(Types) ->
    clauses(fun(Clause) -> lists:member(?Clause:type(Clause), Types) end);
lookfor(clause, Type) ->
    lookfor(clause, [Type]).

lookfor(expr, Type, Value) ->
    exprs(fun(Expr) -> ?Expr:type(Expr) == Type andalso ?Expr:value(Expr) == Value end).

exprs(Filter) ->
    All = ?Query:exec(?Query:seq([?Mod:all(), ?Mod:locals(), ?Fun:definition(), ?Form:clauses(), ?Clause:body(), ?Expr:deep_sub()])),
    lists:filter(Filter, All).

clauses(Filter) ->
    FunClauses  = ?Query:exec(?Query:seq([?Mod:all(), ?Mod:locals(), ?Fun:definition(), ?Form:clauses()])),
    Exprs       = ?Query:exec(FunClauses, ?Query:seq(?Clause:body(), ?Expr:deep_sub())),
    Blocks      = lists:filter(fun(E) -> lists:member(?Expr:type(E), [fun_expr, block_expr]) end, Exprs),
    ExprClauses = ?Query:exec(Blocks, [exprcl]),
    All         = FunClauses ++ ExprClauses,
    lists:filter(Filter, All).

functions(Name) when is_atom(Name) ->
    ?Query:exec(?Query:seq([?Mod:all(), [{func, {name, '==', Name}}], ?Fun:definition()])).
functions()  ->
    ?Query:exec([file, {form, {type, '==', func}}]).

