-module(primes).
-behaviour(gen_server).

-export([primes/1]).
-export([precalculate_primes/0]).
-export([front/0, push/1]).
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).
-export([is_prime/1]).

-define(debug(S, X), io:format(S, X)).
-define(wait, timer:sleep(1000)).
-define(buffersize, 10).

primes(N) -> lists:filter(fun is_prime/1, lists:seq(1,N)).

precalculate_primes() ->
    gen_server:start_link({local, primes}, primes, [{?MODULE,is_prime}], []).

front() ->
    gen_server:call(primes, front).

push(X) ->
    gen_server:call(primes, {push, X}).

init({Pred, ChunkSize}) ->
    spawn(fun() -> work(Pred, 1, ChunkSize) end),
    {ok, []};

init(Pred) -> init({Pred, 100}).

handle_call(front, _From, []) ->
    {reply, buffer_empty, []};

handle_call(front, _From, [X|Xs]) ->
    {reply, X, Xs};
handle_call({push, X}, _From,  Xs) when length(Xs) < ?buffersize+1 ->
    ?debug("~p added.~n", [X]),
    {reply, ok, Xs++[X]};
handle_call({push, X}, _From, Xs) ->
    ?debug("~p cannot be added, the buffer is full.~n", [X]),
    {reply, buffer_full, Xs}.

handle_cast(_, St) ->
    {noreply, St}.

handle_info(_Info, State) ->
    {noreply, State}.

terminate(_Reason, _State) ->
    ok.

code_change(_Old, State, _Extra) ->
    {ok, State}.

work(Pred=[{M, F}], From, HowMany) ->
    [try2push(X) || X <- lists:seq(From, From+HowMany), M:F(X)],
    work(Pred, From+HowMany+1, HowMany).

try2push(X) ->
    ?debug("Trying to push ~p...~n", [X]),
    case push(X) of
        ok ->
            ok;
        buffer_full ->
            ?wait,
            try2push(X)
    end.

prime_factors(N) -> prime_factors(N, 2).
prime_factors(N, F) when F*F > N -> [N];
prime_factors(N, F) when N rem F == 0 -> [F | prime_factors(N div F, F)];
prime_factors(N, F) -> prime_factors(N, F + 1).

prime_factors2(N) -> [F || F <- lists:seq(2,N), N >= F*F orelse F == N, N rem F == 0].

is_prime(N) -> prime_factors2(N) == [N].
