%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Eötvös Loránd University.
%%% Portions created  by Eötvös Loránd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Eötvös Loránd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

%%% @doc Transforms a lists:zipwith/3 function application into a
%%% block expression containing a list comprehension.
%%%
%%% Example:
%%% f(L1, L2) -> lists:zipwith(fun(X, Y) -> X+Y end, L1, L2).
%%%
%%% Result of transformation:
%%% f(L1, L2) ->
%%%     begin
%%%         NV1 = lists:zip(L1, L2),
%%%         NV2 = fun({NV3,NV4}) -> fun(X, Y) -> X+Y end(NV3, NV4) end
%%%         [ NV2(NV5) || NV5 <- NV1 ]
%%%     end.
%%%
%%% @author Tokodi Gergely <demolong@caesar.elte.hu>

-module(refac_lists_zipwith).
-behaviour(gen_refac).

%% export of callback function.
-export([input_par_prompts/0, select_focus/1,
     check_pre_cond/1, selective/0,
     transform/1]).

-include("../../include/wrangler.hrl").

% we need no interaction, so it returns []
-spec (input_par_prompts/0::() -> [string()]).               
input_par_prompts() ->
    [].

-spec (select_focus/1::(#args{}) -> {ok, syntaxTree()}|{ok, none}).  
select_focus(_Args=#args{current_file_name=File,
             highlight_range={Start,End}}) ->
    {Exp,Fun} = pp_api_refac:select_exp(File, Start, End),
    {ok, {Exp, Fun}}.
		
    
%% we need no pre-condition checking, so it returns ok
-spec (check_pre_cond/1::(#args{}) -> ok).  
check_pre_cond(_Args) ->
    ok.

% always false
selective() ->
    false.

%% Do the actual program transformation here.
-spec (transform/1::(args()) -> 
    {ok, [{{filename(), filename()}, syntaxTree()}]}).
transform(_Args=#args{current_file_name=File,
              focus_sel={Exp,Fun}}) -> 
    % SelectedEntity is the entity that returned by the select_focus/1.
    ?STOP_TD_TP([rule_replace_zipwith(Exp,Fun),
        rule_replace_zipwith(Exp,Fun),
        rule_replace_zipwith(Exp,Fun)], [File]).

rule_replace_zipwith(Exp,Fun) -> 
    ?RULE(?FUN_APPLY(lists, zipwith, 3), 
    begin
        Args = api_refac:get_app_args(_This@),
        F = pp_api_refac:arg_nth(1, Args),
        L1 = pp_api_refac:arg_nth(2, Args),
        L2 = pp_api_refac:arg_nth(3, Args),

        Vars = api_refac:bound_var_names(Fun),
        [NV1,NV2,NV3,NV4,NV5] = pp_api_refac:make_new_names("NV",Vars,5),

        ?TO_AST(lists:append([
            "begin ", 
            NV1, " = lists:zip(", L1, ",", L2, "), ",
            NV2, " = fun({", NV3, ",", NV4, "}) ->",
            F, "(", NV3, ",", NV4, ") end, ",
            "[", NV2, "(", NV5, ") || ", NV5, " <- ", NV1, "] ",
            "end" 
        ]))
    end,
    api_refac:start_end_loc(_This@) == api_refac:start_end_loc(Exp)
    ).
