%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Eötvös Loránd University.
%%% Portions created  by Eötvös Loránd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Eötvös Loránd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

%%% @doc TODO
%%%
%%% @author Tokodi Gergely <demolong@caesar.elte.hu>

-module(pp_api_refac).

-export([make_new_name/2,make_new_names/3,
        select_exp/3,select_lc/3,arg_nth/2,
        insert_new_fun/4]).

-include("../include/wrangler.hrl").

%%-spec make_new_name(string(), [atom()]) ->string().		   
make_new_name(BaseName, UsedNames) ->
    case lists:member(list_to_atom(BaseName), UsedNames) of
    true ->
        NewName = string:concat(BaseName,  "_1"),
        make_new_name(NewName, UsedNames);
    _ -> 
        BaseName
    end.

%%-spec make_new_names(string(), [atom()]) -> [string()].
make_new_names(BaseName, UsedNames, N) ->
    [make_new_name(string:concat(BaseName,integer_to_list(X)), UsedNames) || X <- lists:seq(1,N)].

%%-spec select_exp(File::filename(), Start::pos(), End::pos()) -> {syntaxtree(), syntaxtree()} | {error, string()}.
select_exp(File, Start, End) ->
    {ok, {AnnAST, _Info}} = wrangler_ast_server:parse_annotate_file(File,true),
    Exp = case api_interface:pos_to_expr(AnnAST, Start, End) of
    {ok, Expr} ->
        Expr;
    {error, _} ->
        throw({error, "You have not selected an expression, "
          "or the function containing the expression does not parse."})
    end,

    case api_interface:expr_to_fun(AnnAST, Exp) of
    {ok, Fun} ->
        {Exp,Fun};
    {error, _} ->
        throw({error, "You have not selected an expression within a function."})
    end.

%%-spec select_lc(File::filename(), Start::pos(). End::pos()) -> {syntaxtree(), syntaxtree()} | {error, string()}.
select_lc(File, Start, End) ->
    {ok, {AnnAST, _Info}} = wrangler_ast_server:parse_annotate_file(File, true),
    Exp = case api_interface:range_to_node(AnnAST, {Start,End}, fun(_Expr) -> true end) of
    {ok, Expr} ->
        Expr;
    {error, _} -> 
        throw({error, "You have not selected an expression, "
          "or the function containing the expression does not parse."})
    end,

    case api_interface:expr_to_fun(AnnAST, Exp) of
    {ok, Fun} ->
        {Exp,Fun};
    {error, _} ->
        throw({error, "You have not selected an expression within a function."})
    end.

%%-spec arg_nth(integer(),[syntaxtree()]) -> string()
arg_nth(N, Args) ->
    lists:append([], ?PP(lists:nth(N, Args))).


%%-spec insert_new_fun(file(), syntaxtree(), string(), integer()) -> syntaxtree()
insert_new_fun(FileName,Expr,NewFunName, TabWidth) ->
    {_Start, _End} = api_refac:start_end_loc(Expr),
    {ok, {AnnAST, _Info}} = wrangler_ast_server:parse_annotate_file(FileName, true),
    {ok, Fun} = api_interface:expr_to_fun(AnnAST, Expr),
    %% Check if NewFunName is valid function name
    case api_refac:is_fun_name(NewFunName) of
        true -> ok;
        false -> throw({error, "Invalid function name!"})
    end,

    % Can i use this function in the given module?
    %InScopeFuns = [{F, A} || {_M, F, A} <- api_refac:inscope_funs(Info)],

    FunName = wrangler_syntax:atom_value(wrangler_syntax:function_name(Fun)),
    FunArity = wrangler_syntax:function_arity(Fun),
    FunPos = wrangler_syntax:get_pos(Fun),
    
    AnnAST1 = random(FileName, AnnAST, Expr, NewFunName,{FunName, FunArity, FunPos}),
    %[{{FileName, FileName}, AnnAST1}].
    Res = [{{FileName, FileName}, AnnAST1}],
    wrangler_write_file:write_refactored_files_for_preview(Res, TabWidth, ""),
      {ok, [FileName]}.





random(FileName, AnnAST, Expr, NewFunName, {EncFunName, EncFunArity, EncFunPos}) ->
    NewFunName1 = wrangler_syntax:atom(NewFunName),

    Clause = wrangler_syntax:clause([], [], [Expr]),
    NewFun = wrangler_syntax:function(NewFunName1, [Clause]),
    Forms = wrangler_syntax:form_list_elements(AnnAST),
    Fun = fun (Form) ->
          case wrangler_syntax:type(Form) of
              function ->
              Name = wrangler_syntax:atom_value(wrangler_syntax:function_name(Form)),
              Arity = wrangler_syntax:function_arity(Form),
              Pos = wrangler_syntax:get_pos(Form),
              case {Name, Arity, Pos} == {EncFunName, EncFunArity, EncFunPos} of
                  true ->
                  Toks = get_node_toks(FileName, wrangler_misc:get_toks(Form), [Expr]),
                                  NewFun1 = wrangler_misc:update_ann(NewFun, {toks, Toks}),
                  [Form, NewFun1];
                  _ -> [Form]
              end;
              _ -> [Form]
          end
      end,
    wrangler_syntax:form_list([F || Form <- Forms, F <- Fun(Form)]).


get_node_toks(FileName,Toks, Node) ->
    {Start, End} = wrangler_misc:start_end_loc(Node),
    case Start =={0,0} orelse End=={0,0} of
    true -> [];
    false ->
        Toks1 = lists:dropwhile(
              fun(T) ->
                  element(2, T)<Start
              end, Toks),
        Toks2=lists:takewhile(fun(T)->
                                          element(2, T)=<End
                                  end, Toks1),
            get_delimitor(FileName)++Toks2
    end.

get_delimitor(FileName) ->        
    case wrangler_misc:file_format(FileName) of
        dos ->
            [{whitespace, {0,0}, '\r'},
             {whitespace, {0,0}, '\n'},
             {whitespace, {0,0}, '\r'},
             {whitespace, {0,0}, '\n'}];
        mac -> 
            [{whitespace, {0,0}, '\r'},
            {whitespace, {0,0}, '\r'}];
        _ -> 
            [{whitespace, {0,0}, '\n'},
             {whitespace, {0,0}, '\n'}]
    end.